﻿/*
 *    File: _game.js
 * Version: 1.7.5
 * Project: Brad's Erotic Week
 * Authors: kexter, Wolfschadowe
 *    Date: 2018-06-04
 */

//#region #              ESLint configuration               #

/* global terminal */
/* exported bewCompareVersion, utChoose */

//#endregion

"use strict";

//#region #                 Game Variables                  #

var bewVersion = "1.7.5";  // Important: Only use #.#.# format!
var bewIsRelease = true;   // Set to true for general public releases.
var bewIsInternal = false;   // Set to true for developer versions.

// TODO: Readd support for early-access and tester-access.
//if (typeof accessEarly  === "undefined") { var accessEarly  = true; }
//if (typeof accessTester === "undefined") { var accessTester = true; }

var bewOnlineVersion = window.location.protocol !== "file:";
var bewCheckForUpdates = !bewOnlineVersion && (bewIsRelease || !bewIsInternal);
var bewDumpSession = false;

// Set this to false for release versions.
var bewDebugMenuEnabled = !bewIsRelease;

// Retro mode.
var bewRetroMode = false;

// Special save storage for the credits page.
var bewSpecialSave = "";

// Lookup table for the names of the individual story paths.
var bewStoryPaths = [
  "Deadbeat Main", "Deadbeat Club", "Brad Gentleman", "Blames", "Banes",
  "Beams", "Beng", "Bens", "Blam", "BJ", "Emily Main", "Emily Elite",
  "Emily Corrupt", "Emily Only", "Emily CS Main", "Emily CS Elite",
  "EF Two Time", "EM Elite", "Elf", "Emily Easy", "Faith Main", "Faith Elite",
  "Faith Corrupt", "FE Main", "FE Elite", "Flamens Elite", "FN Main", "Flames",
  "Fem", "Faith Easy", "Natalie Main", "Natalie Elite", "Natalie Corrupt",
  "Nefs", "NAF Elite", "NAM Elite", "NEF Elite", "NAE", "NL Elite",
  "Natalie Easy", "Serenthia Main", "Serenthia Elite", "Serenthia Corrupt",
  "SM Elite", "SF Elite", "SLM Elite", "SLAM Elite", "Same Elite", "SAN",
  "Serenthia Easy", "Lorien Main", "Lorien Elite", "Lorien Corrupt", "Lean",
  "LF Main", "LAN", "LNFE Elite", "Lame", "Leaf Elite", "Lorien Easy",
  "Azumi Main", "Azumi Elite", "Azumi Corrupt", "AME", "AF Main", "AFS Elite",
  "ANLM Elite", "AN Main", "AN Elite", "Azumi Easy", "Jasmine Main",
  "Jasmine Elite", "Jasmine Corrupt", "Jeans", "JA Elite", "JEFN Elite",
  "JF Elite", "JE Elite", "James", "Jasmine Easy", "Muirgheal Main",
  "Muirgheal Elite", "Muirgheal Corrupt", "MLN Elite", "MF Main", "Man",
  "ML Elite", "MN Elite", "Mean", "Muirgheal Easy"
];

// Polyfill requestAnimationFrame if necessary.
window.requestAnimationFrame = window.requestAnimationFrame ||
  window.webkitRequestAnimationFrame ||
  window.mozRequestAnimationFrame ||
  window.msRequestAnimationFrame ||
  window.oRequestAnimationFrame ||
  (function () {
    var frameTime = 1000 / 60;
    var callbackList = [];
    var tickTimer = null;
    function processCallbacks() {
      var i;
      var currentTime = Date.now();
      var callbacks = callbackList;
      callbackList = [];
      tickTimer = null;
      for (i = 0; i < callbacks.length; i += 1) {
        callbacks[i](currentTime);
      }
    }
    function addCallback(callback) {
      callbackList.push(callback);
      if (tickTimer === null) {
        tickTimer = setTimeout(processCallbacks, frameTime);
      }
    }
    return addCallback;
  }());

// Polyfill scrollIntoViewIfNeeded if necessary.
if (!Element.prototype.scrollIntoViewIfNeeded) {
  Element.prototype.scrollIntoViewIfNeeded = function (center) {
    center = typeof center === "undefined" ? true : !!center;
    var self = this;
    var parent = self.parentNode;
    var parentComputedStyle = window.getComputedStyle(parent, null);
    var parentBorderTopWidth = parseInt(
      parentComputedStyle.getPropertyValue("border-top-width"), 10);
    var parentBorderLeftWidth = parseInt(
      parentComputedStyle.getPropertyValue("border-left-width"), 10);
    var top = self.offsetTop - parent.offsetTop;
    var left = self.offsetLeft - parent.offsetLeft;
    var bottom = top + self.clientHeight - parentBorderTopWidth;
    var right = left + self.clientWidth - parentBorderLeftWidth;
    var clipTop = top < parent.scrollTop;
    var clipLeft = left < parent.scrollLeft;
    var clipBottom = bottom > parent.scrollTop + parent.clientHeight;
    var clipRight = right > parent.scrollLeft + parent.clientWidth;
    if (center) {
      if (clipTop || clipBottom) {
        parent.scrollTop = top - parentBorderTopWidth -
          (parent.clientHeight - self.clientHeight) / 2;
      }
      if (clipLeft || clipRight) {
        parent.scrollLeft = left - parentBorderLeftWidth -
          (parent.clientWidth - self.clientWidth) / 2;
      }
    } else if (clipTop || clipLeft || clipBottom || clipRight) {
      // TODO: Polyfill scrollIntoView()?
      if (typeof self.scrollIntoView === "function") {
        self.scrollIntoView(clipTop && !clipBottom);
      }
    }
  };
}

/**
 * Generates the version string to be displayed in the menu.
 * @returns {string} The formatted version string.
 */
function bewVersionString() {
  var str = "BEW " + bewVersion;
  //str += !!readStorageString("Patch") ? "+" : "";
  str += bewIsInternal ? " WIP" : "";
  str += bewIsRelease ? "" : "<br>DO NOT DISTRIBUTE";
  return str;
}

//#region Re-routing functionality.
(function () {
  var page, target;
  var currentUrl = location.href;
  var assetLocation = "/Files/";
  var indexPage = "/Start.htm";
  if (currentUrl.indexOf(indexPage) < 0 &&
      currentUrl.lastIndexOf(assetLocation) > -1) {
    page = currentUrl.replace(/.*(\/|\\)(.*)\.(.*)/, "$2").toLowerCase();
    target = currentUrl.substr(0, currentUrl.lastIndexOf(assetLocation));
    target += indexPage + "#" + page;
    window.location = target;
  }
}());
//#endregion

//#endregion

//#region #            Custom Utility Functions             #

//#region ------- Asset manager functions --------

var assetManager = (function () {
  var defaultImage = "data:image/gif;base64,R0lGODlhAQABAIAAAP///wAAACH5BAEAA" +
    "AAALAAAAAABAAEAAAICRAEAOw==";
  var microHeader = base64ToUint8Array(
    "/9j/4AAQSkZJRgABAQAAAQABAAD/2wBDACgcHiMeGSgjISMtKygwPGRBPDc3PHtYXUlkkYCZ" +
    "lo+AjIqgtObDoKrarYqMyP/L2u71////m8H////6/+b9//j/2wBDASstLTw1PHZBQXb4pYyl" +
    "+Pj4+Pj4+Pj4+Pj4+Pj4+Pj4+Pj4+Pj4+Pj4+Pj4+Pj4+Pj4+Pj4+Pj4+Pj4+Pj4+Pj/wAAR" +
    "CAAYADADASIAAhEBAxEB/8QAHwAAAQUBAQEBAQEAAAAAAAAAAAECAwQFBgcICQoL/8QAtRAA" +
    "AgEDAwIEAwUFBAQAAAF9AQIDAAQRBRIhMUEGE1FhByJxFDKBkaEII0KxwRVS0fAkM2JyggkK" +
    "FhcYGRolJicoKSo0NTY3ODk6Q0RFRkdISUpTVFVWV1hZWmNkZWZnaGlqc3R1dnd4eXqDhIWG" +
    "h4iJipKTlJWWl5iZmqKjpKWmp6ipqrKztLW2t7i5usLDxMXGx8jJytLT1NXW19jZ2uHi4+Tl" +
    "5ufo6erx8vP09fb3+Pn6/8QAHwEAAwEBAQEBAQEBAQAAAAAAAAECAwQFBgcICQoL/8QAtREA" +
    "AgECBAQDBAcFBAQAAQJ3AAECAxEEBSExBhJBUQdhcRMiMoEIFEKRobHBCSMzUvAVYnLRChYk" +
    "NOEl8RcYGRomJygpKjU2Nzg5OkNERUZHSElKU1RVVldYWVpjZGVmZ2hpanN0dXZ3eHl6goOE" +
    "hYaHiImKkpOUlZaXmJmaoqOkpaanqKmqsrO0tba3uLm6wsPExcbHyMnK0tPU1dbX2Nna4uPk" +
    "5ebn6Onq8vP09fb3+Pn6/9o="
  );
  //var markerSOS = new Uint8Array([0xff, 0xda]);
  var markerEOI = new Uint8Array([0xff, 0xd9]);
  var images = {
    microData: {},
    mimeType: {},
    data: {},
    url: {},
    lod: [
      {
        luma: {},
        chroma: {}
      },
      {},
      {}
    ]
  };
  function base64ToUint8Array(data) {
    return new Uint8Array(atob(data).split("").map(function (c) {
      return c.charCodeAt(0);
    }));
  }
  function uint8ArrayToBase64(data) {
    var result = [];
    var i;
    for (i = 0; i < data.length; i += 1) {
      result.push(String.fromCharCode(data[i]));
    }
    result = result.join("");
    return btoa(result);
  }
  function getMicroImage(id) {
    var result = "";
    if (typeof images.microData[id] !== "undefined") {
      result = "data:image/jpeg;base64," +
        uint8ArrayToBase64(images.microData[id]);
    }
    return result;
  }
  function getImageLod(id, level) {
    var result = "";
    level = level || "lod2";
    if (level === "lod0-gray") {
      // TODO: Implement lod0-gray support.
    } else if (level === "lod0") {
      // TODO: Implement lod0 support.
    } else if (level === "lod1-gray") {
      // TODO: Implement lod1-gray support.
    } else if (level === "lod1") {
      // TODO: Implement lod1 support.
    } else if (level === "lod2") {
      // TODO: Implement lod2 support.
    }
    return result;
  }
  function getImage(id, level) {
    var result = "";
    level = level || "";
    if (!level && typeof images.data[id] !== "undefined") {
      result = images.data[id];
    } else if (!level && typeof images.url[id] !== "undefined") {
      result = images.url[id];
    } else {
      result = getImageLod(id, level);
    }
    return result;
  }
  function defineImage(id, microData, mimeType) {
    var data;
    images.mimeType[id] = mimeType || "image/jpeg";
    // If specified, store the microData.
    if (typeof microData === "string") {
      microData = base64ToUint8Array(microData);
      data = new Uint8Array(microHeader.length + microData.length + 2);
      data.set(microHeader);
      data.set(microData, microHeader.length);
      data.set(markerEOI, data.length - 2);
      images.microData[id] = data;
    }
  }
  function registerLocale(lang, friendlyName) {
    if (!game.locales[lang]) {
      game.locales[lang] = {
        loaded: false,
        friendlyName: friendlyName,
        strings: {}
      };
    }
  }
  function defineLocale(lang, friendlyName) {
    registerLocale(lang, friendlyName);
    game.locales[lang].loaded = true;
    return game.locales[lang].strings;
  }
  function loadLocale(lang, callback) {
    var src = "Files/locale." + lang + ".js?v=" + bewVersion;
    // Make sure the default locale is always loaded first.
    if (lang !== game.defaultLocale && (!game.locales[game.defaultLocale] ||
        !game.locales[game.defaultLocale].loaded)) {
      loadLocale(game.defaultLocale, function () {
        terminal.addScript(src, callback);
      });
    } else {
      terminal.addScript(src, callback);
    }
  }
  function doApplyLocale(lang, callback) {
    var list = document.querySelectorAll("[brew-string]");
    var i, el, template;
    game.activeLocale = lang;
    for (i = 0; i < list.length; i += 1) {
      el = list[i];
      template = el.getAttribute("brew-string");
      el.innerHTML = $.evaluateText(template);
    }
    list = document.querySelectorAll("[brew-title]");
    for (i = 0; i < list.length; i += 1) {
      el = list[i];
      template = el.getAttribute("brew-title");
      el.title = $.evaluateText(template);
    }
    if (typeof callback === "function") {
      callback();
    }
  }
  function applyLocale(lang, callback) {
    var locale;
    if (lang !== game.activeLocale) {
      locale = game.locales[lang];
      if (locale) {
        if (!locale.loaded) {
          loadLocale(lang, function () {
            doApplyLocale(lang, callback);
          });
        } else {
          doApplyLocale(lang, callback);
        }
      }
    } else {
      if (typeof callback === "function") {
        callback();
      }
    }
  }
  function defineScript(url) {
    if (url) {
      game.scripts.push(url);
    }
  }
  function defineSources(id, lod0pack, lod1pack, lod2pack) {
    // Verify that id and lod0pack is specified.
    if (typeof id === "string" && typeof lod0pack === "string") {
      // TODO: Add support for only specifying lod0pack - for PNGs, etc.
      if (typeof lod1pack === "undefined") {
        // If only id and lod0pack is defined then assume that
        // either an url or a data uri is specified.
        if (lod0pack.indexOf("data:") === 0) {
          // A data uri was provided, store it as data.
          images.data[id] = lod0pack;
        } else {
          // Assume that an url was specified.
          images.url[id] = lod0pack;
        }
      }
      // TODO: Implement support for lod-packs.
      void lod2pack;
    }
  }
  function requestImage(id, level) {
    var result = "";
    if (level === "micro") {
      result = getMicroImage(id);
    } else {
      result = getImage(id, level);
    }
    return result || defaultImage;
  }
  /**
   * Returns a resource string from the current locale.
   * @param {string} id The unique identifier of the string.
   * @param {string} [defaultValue]
   *   The value to return if the string was not found. Defaults to `id´.
   * @returns {string} The requested string.
   */
  function requestString(id, defaultValue) {
    var result = game.locales[game.activeLocale].strings[id];
    if (typeof result === "undefined") {
      result = game.locales[game.defaultLocale].strings[id];
    }
    if (typeof result === "undefined") {
      if (typeof defaultValue !== "undefined") {
        result = defaultValue;
      } else {
        result = id;
      }
    }
    return result;
  }
  function storeLod0Luma(id, data) {
    // TODO: Implement assetsManager.storeLod0Luma().
    void data;
  }
  function storeLod0Chroma(id, data) {
    // TODO: Implement assetsManager.storeLod0Chroma().
    void data;
  }
  function storeLod1(id, data) {
    // TODO: Implement assetsManager.storeLod1().
    void data;
  }
  function storeLod2(id, data) {
    // TODO: Implement assetsManager.storeLod2().
    void data;
  }
  return {
    applyLocale: applyLocale,
    defineImage: defineImage,
    defineLocale: defineLocale,
    defineScript: defineScript,
    defineSources: defineSources,
    registerLocale: registerLocale,
    requestImage: requestImage,
    requestString: requestString,
    storeLod0Luma: storeLod0Luma,
    storeLod0Chroma: storeLod0Chroma,
    storeLod1: storeLod1,
    storeLod2: storeLod2
  };
}());

//#endregion

//#region ----------- Engine functions -----------

// The global game object.
var game = {
  uid: "BEW",
  initialized: false,
  running: false,
  scripts: [],
  stylesheets: [],
  variables: {},
  functions: {},
  pages: {},
  activeLocale: "",
  defaultLocale: "en-US",
  locales: {}
};

// The global template object.
var template = (function () {
  function getCanvas() {
    return document.getElementById("MainImage");
  }
  function getHotspotContainer() {
    return document.getElementById("HotspotContainer");
  }
  function getResponseContainer() {
    return document.getElementById("ResponseContainer");
  }
  function getTextContainer() {
    return document.getElementById("MainText");
  }
  function getDebugTooltip() {
    var el = document.getElementById("BiskTooltip");
    if (!el) {
      el = document.createElement("div");
      el.id = "BiskTooltip";
      el.className = "bisk-source";
      document.getElementById("ContentWrapper").appendChild(el);
    }
    return el;
  }
  function clearDebugTooltip() {
    if (debugMode.isEnabled()) {
      getDebugTooltip().style.display = "none";
    }
  }
  function updateDebugTooltip(e, div) {
    var el, x, y, r;
    var value = div.getAttribute("data-bisk") || "";
    if (debugMode.isEnabled() && value) {
      el = getDebugTooltip();
      if (el.style.display !== "inline-block") {
        el.style.display = "inline-block";
        el.innerHTML = bewUtilAnnotateBisk(value);
      }
      r = document.getElementById("ContentWrapper").getBoundingClientRect();
      x = e.pageX + 16 - r.left - (window.pageXOffset || 0);
      y = e.pageY + 24 - r.top - (window.pageYOffset || 0);
      el.style.transform = "translate(" + x + "px," + y + "px)";
    }
  }
  function addHotspot(polygon, textTemplate, callback) {
    function updateTooltipPosition(e) {
      var r = container.getBoundingClientRect();
      var k = p.getBoundingClientRect();
      var width = r.right - r.left;
      var height = r.bottom - r.top;
      var w = Math.ceil((k.right - k.left) / 2);
      var h = k.bottom - k.top;
      var x = e.pageX - r.left - (window.pageXOffset || 0);
      var y = e.pageY - r.top - (window.pageYOffset || 0);
      x = x + w > width ? width - w : x;
      x = x - w < 0 ? 0 : x - w;
      y = y > height ? height : y;
      y = y - h < 0 ? 0 : y - h;
      p.style.transform = "translate(" + x + "px," + y + "px)";
      updateDebugTooltip(e, div);
    }
    var doSpellCheck = debugMode.isEnabled() && checkFlag("spellChecking");
    var text = $.evaluateText(textTemplate);
    var container = getHotspotContainer();
    var result = null;
    var div, a, b, p, x, y, w, h;
    // Currently we only supports rects.
    if (text && polygon.length === 4) {
      x = polygon[0];
      y = polygon[1];
      w = polygon[2];
      h = polygon[3];
      // Generate the hotspot.
      div = document.createElement("div");
      div.className = "hotspot";
      div.style.left = x + "%";
      div.style.top = y + "%";
      div.style.width = w + "%";
      div.style.height = h + "%";
      // Create the main element.
      a = document.createElement("i");
      a.role = "button";
      // TODO: Readd tabIndex for hotspots.
      //tabIndex += 1;
      //a.tabIndex = tabIndex;
      if (typeof callback === "function") {
        a.addEventListener("click", callback);
        a.addEventListener("keydown", bewKeyDownOnClick);
      }
      // Create the "hover-on-tooltip".
      p = document.createElement("p");
      p.spellcheck = doSpellCheck;
      p.contentEditable = doSpellCheck;
      $.bindText(p, textTemplate);
      // Create the "cover" element.
      b = document.createElement("b");
      div.addEventListener("mousemove", updateTooltipPosition);
      div.addEventListener("mouseover", updateTooltipPosition);
      div.addEventListener("mouseout", clearDebugTooltip);
      // Put it all together and add it to the tree.
      div.appendChild(a);
      div.appendChild(b);
      container.insertBefore(p, container.firstChild);
      container.insertBefore(div, container.firstChild);
      result = div;
    }
    return result;
  }
  function addResponse(textTemplate, callback) {
    function updateTooltipPosition(e) {
      updateDebugTooltip(e, el);
    }
    var doSpellCheck = debugMode.isEnabled() && checkFlag("spellChecking");
    var text = $.evaluateText(textTemplate);
    var container = getResponseContainer();
    var result = null;
    var el, p;
    // Do not add empty responses.
    if (text) {
      // Create the response element.
      el = document.createElement("div");
      el.className = "response";
      el.role = "button";
      // TODO: Readd tabIndex for responses.
      //tabIndex += 1;
      //el.tabIndex = tabIndex;
      if (typeof callback === "function") {
        el.addEventListener("click", callback);
        el.addEventListener("keydown", bewKeyDownOnClick);
      }
      p = document.createElement("p");
      p.spellcheck = doSpellCheck;
      p.contentEditable = doSpellCheck;
      $.bindText(p, textTemplate);
      el.appendChild(p);
      el.addEventListener("mousemove", updateTooltipPosition);
      el.addEventListener("mouseover", updateTooltipPosition);
      el.addEventListener("mouseout", clearDebugTooltip);
      container.appendChild(el);
      result = el;
    }
    return result;
  }
  function clearText() {
    getTextContainer().innerHTML = "";
  }
  function clearHotspots() {
    var container = template.getHotspotContainer();
    container.classList.remove("easter-egg");
    container.innerHTML = "";
    clearDebugTooltip();
  }
  function clearResponses() {
    getResponseContainer().innerHTML = "";
  }
  return {
    getCanvas: getCanvas,
    getHotspotContainer: getHotspotContainer,
    getResponseContainer: getResponseContainer,
    getTextContainer: getTextContainer,
    addHotspot: addHotspot,
    addResponse: addResponse,
    clearText: clearText,
    clearHotspots: clearHotspots,
    clearResponses: clearResponses
  };
}());

// The engine constructor.
function BrewEngine() {
  var drawingLayers = false;
  var fx = new EffectsHandler();
  var history = new HistoryHandler();
  var retroBuffer = document.createElement("canvas");
  var retroCtx = retroBuffer.getContext("2d");
  var buffer = document.createElement("canvas");
  var bufferCtx = buffer.getContext("2d");
  var canvas = template.getCanvas();
  var ctx = canvas.getContext("2d");
  var topText = "";
  var layers = [];
  var hotspots = [];
  var responses = [];
  var pageLoadEvents = [];
  function BrewFrame() {
    return {
      canvas: document.createElement("canvas"),
      loading: false,
      ready: false,
      microReady: false,
      id: "",
      delay: 0,
      rect: null
    };
  }
  function BrewLayer() {
    return {
      x: 0,
      y: 0,
      w: 100,
      h: 100,
      frameStart: null,
      currentFrame: 0,
      frames: []
    };
  }
  function BrewHotspot(polygon, text, callback) {
    return {
      polygon: polygon || [0, 0, 100, 100],
      text: text || "",
      callback: callback || null
    };
  }
  function BrewResponse(text, callback) {
    return {
      text: text || "",
      callback: callback || null
    };
  }
  function EffectsHandler() {
    var el = document.getElementById("FXFade");
    function enableDiv(background) {
      background = background || "rgba(0,0,0,0)";
      el.style.background = background;
      el.classList.add("fade");
    }
    function disableDiv() {
      el.classList.remove("fade");
    }
    function fadeIn() {
      enableDiv("#000");
      el.classList.remove("out");
      el.classList.add("in");
      setTimeout(function () {
        el.style.background = "rgba(0,0,0,0)";
        setTimeout(function () {
          disableDiv();
        }, 400);
      }, 100);
    }
    function fadeOut() {
      enableDiv("rgba(0,0,0,0)");
      el.classList.remove("in");
      el.classList.add("out");
      setTimeout(function () {
        el.style.background = "#000";
      }, 100);
    }
    function zoomFromRect(x, y, w, h, callback) {
      var startTime, progress;
      var transitionTime = 200;
      var rect = {
        x: x,
        y: y,
        w: w,
        h: h
      };
      function stepAnim(timestamp) {
        var scale, x, y, w, h;
        if (!startTime) {
          startTime = timestamp;
        }
        progress = timestamp - startTime;
        if (progress < transitionTime) {
          scale = 1 - (progress / transitionTime);
          x = rect.x * canvas.width / 100;
          y = rect.y * canvas.height / 100;
          w = rect.w * canvas.width / 100;
          h = rect.h * canvas.height / 100;
          x *= scale;
          y *= scale;
          w = (canvas.width * (1 - scale)) + (w * scale);
          h = (canvas.height * (1 - scale)) + (h * scale);
          ctx.drawImage(buffer, x, y, w, h, 0, 0, canvas.width, canvas.height);
          requestAnimationFrame(stepAnim);
        } else {
          if (typeof callback === "function") {
            callback();
          }
          disableDiv();
        }
      }
      if (checkFlag("noEffects")) {
        if (typeof callback === "function") {
          callback();
        }
      } else {
        enableDiv();
        requestAnimationFrame(stepAnim);
      }
    }
    function zoomToRect(x, y, w, h, callback) {
      var startTime, progress;
      var transitionTime = 200;
      var rect = {
        x: x,
        y: y,
        w: w,
        h: h
      };
      function stepAnim(timestamp) {
        var scale, x, y, w, h;
        if (!startTime) {
          startTime = timestamp;
        }
        progress = timestamp - startTime;
        if (progress < transitionTime) {
          scale = progress / transitionTime;
          x = rect.x * canvas.width / 100;
          y = rect.y * canvas.height / 100;
          w = rect.w * canvas.width / 100;
          h = rect.h * canvas.height / 100;
          x *= scale;
          y *= scale;
          w = (canvas.width * (1 - scale)) + (w * scale);
          h = (canvas.height * (1 - scale)) + (h * scale);
          ctx.drawImage(buffer, x, y, w, h, 0, 0, canvas.width, canvas.height);
          requestAnimationFrame(stepAnim);
        } else {
          if (typeof callback === "function") {
            callback();
          }
          disableDiv();
        }
      }
      if (checkFlag("noEffects")) {
        if (typeof callback === "function") {
          callback();
        }
      } else {
        enableDiv();
        clear();
        requestAnimationFrame(stepAnim);
      }
    }
    return {
      zoomFromRectData: null,
      enableDiv: enableDiv,
      disableDiv: disableDiv,
      fadeIn: fadeIn,
      fadeOut: fadeOut,
      zoomFromRect: zoomFromRect,
      zoomToRect: zoomToRect
    };
  }
  function HistoryHandler() {
    var backupKey = (game.uid || "X") + ":History";
    var pageStartTime = Date.now();
    var usersChoice = "";
    var cursor = -1;
    var states = [];
    var canRecord = true;
    var canBackup = _checkSessionStorage();
    // TODO: Move to terminal.* in the future.
    function _checkSessionStorage() {
      var hasStorage = false;
      // First, see if sessionStorage is available.
      try {
        hasStorage = !!sessionStorage;
      } catch (e) {
        // Nop.
      }
      // Then check if we can actually save data.
      if (hasStorage) {
        try {
          sessionStorage.setItem("-_-", "1");
          sessionStorage.removeItem("-_-");
        } catch (e) {
          hasStorage = false;
        }
      }
      return hasStorage;
    }
    function _backup() {
      if (canBackup) {
        sessionStorage.setItem(backupKey, JSON.stringify({
          version: 1,
          cursor: cursor,
          states: JSON.stringify(states.slice(0, cursor))
        }));
      }
    }
    function _restore() {
      var data;
      if (canBackup) {
        data = sessionStorage.getItem(backupKey) || "{}";
        try {
          data = JSON.parse(data);
        } catch (e) {
          data = {};
        }
        if (data.version && data.cursor && data.states && data.version === 1) {
          cursor = data.cursor;
          states = JSON.parse(data.states);
        }
        updateButtons();
      }
    }
    function clear() {
      states = [];
      cursor = -1;
      pageStartTime = Date.now();
      usersChoice = "";
      updateButtons();
    }
    function back() {
      cursor -= 1;
      cursor = cursor < 0 ? 0 : cursor;
      restoreState();
    }
    function forward() {
      cursor += 1;
      cursor = cursor > states.length - 1 ? states.length - 1 : cursor;
      restoreState();
    }
    function getState() {
      return states[cursor] || {};
    }
    function record(page, state) {
      var prev;
      if (canRecord) {
        states.splice(cursor + 1);
        cursor = states.length;
        if (states.length > 0) {
          prev = states[states.length - 1];
          prev.choice = usersChoice;
          prev.time = Date.now() - pageStartTime;
        }
        states.push({
          choice: "",
          page: page,
          state: state,
          time: 0
        });
        pageStartTime = Date.now();
        usersChoice = "";
        updateButtons();
      }
    }
    function recordChoice(type, value) {
      if (canRecord) {
        usersChoice = type + ":" + value;
      }
    }
    /** Restores the current state. */
    function restoreState() {
      var state = states[cursor] || null;
      if (state) {
        setState(state.state);
        canRecord = false;
        $.jmp(state.page);
        canRecord = true;
      }
      updateButtons();
    }
    function serialize() {
      return JSON.stringify(states);
    }
    /** Hides and shows the forward button as necessary. */
    function updateButtons() {
      var backButton = document.getElementById("DebugBackButton");
      var forwardButton = document.getElementById("DebugForwardButton");
      if (backButton) {
        if (states.length < 2) {
          backButton.classList.add("hidden");
        } else {
          backButton.classList.remove("hidden");
        }
      }
      if (forwardButton) {
        if (cursor > states.length - 2) {
          forwardButton.classList.add("hidden");
        } else {
          forwardButton.classList.remove("hidden");
        }
      }
      _backup();
    }
    // Restore history data when a page is reloaded.
    _restore();
    return {
      back: back,
      clear: clear,
      forward: forward,
      getState: getState,
      record: record,
      recordChoice: recordChoice,
      serialize: serialize
    };
  }
  function addHotspot(descriptor, text, callback) {
    var polygon = descriptor.split(",").map(Number);
    callback = typeof callback === "string"
      ? game.functions[callback]
      : callback;
    hotspots.push(new BrewHotspot(polygon, text, callback));
  }
  function addResponse(text, callback) {
    callback = typeof callback === "string"
      ? game.functions[callback]
      : callback;
    responses.push(new BrewResponse(text, callback));
  }
  function blit(descriptor, layerId) {
    layerId = typeof layerId === "undefined" ? layers.length : layerId;
    var layer = new BrewLayer();
    var i, coords, frame, data;
    if (layerId >= layers.length) {
      layerId = layers.length;
    }
    // Turn the descriptor into an animated spite layer.
    // "n1-bar17b-h::500|n1-bar17b-h1::500|n1-bar17b-h2::500"
    // "sample@25,25,50,50"
    descriptor = descriptor.split("@");
    if (descriptor.length > 1) {
      coords = descriptor[1].split(",").map(Number);
      layer.x = coords[0];
      layer.y = coords[1];
      layer.w = coords[2];
      layer.h = coords[3];
    }
    descriptor = descriptor[0].split("|");
    for (i = 0; i < descriptor.length; i += 1) {
      frame = new BrewFrame();
      data = descriptor[i].split(":");
      frame.id = data[0];
      if (data.length > 1) {
        if (data[1].length > 0) {
          frame.rect = data[1].split(",").map(Number);
        }
        if (data.length > 2) {
          frame.delay = parseInt(data[2], 10);
        }
      }
      layer.frames.push(frame);
    }
    layers[layerId] = layer;
    return layerId;
  }
  function clearPage(keepText, keepHotspots, keepResponses) {
    keepText = keepText || false;
    keepHotspots = keepHotspots || false;
    keepResponses = keepResponses || false;
    // Clear the page. The image is intentionally not cleared to avoid flicker.
    if (!keepText) {
      template.clearText();
    }
    if (!keepHotspots) {
      template.clearHotspots();
    }
    if (!keepResponses) {
      template.clearResponses();
    }
  }
  function clear(keepText, keepImages, keepHotspots, keepResponses) {
    keepText = keepText || false;
    keepImages = keepImages || false;
    keepHotspots = keepHotspots || false;
    keepResponses = keepResponses || false;
    if (!keepText) {
      topText = "";
    }
    if (!keepImages) {
      layers = [];
    }
    if (!keepHotspots) {
      hotspots = [];
    }
    if (!keepResponses) {
      responses = [];
    }
    // Clear the page.
    clearPage(keepText, keepHotspots, keepResponses);
  }
  function drawLayers() {
    var redrawRequested = false;
    var microsDrawn = 0;
    function doRedraw() {
      redrawRequested = false;
      doDrawLayers();
    }
    function doDrawLayers() {
      var i, j, layer, frame, srcRect, dstRect;
      var microsReady = 0;
      var blurRequired = false;
      var updateRequired = false;
      var layersComplete = true;
      for (i = 0; i < layers.length; i += 1) {
        layer = layers[i];
        frame = layer.frames[layer.currentFrame];
        microsReady += frame.microReady ? 1 : 0;
        microsReady -= frame.ready ? 1 : 0;
        updateRequired = updateRequired || (!frame.ready && !frame.microReady);
        layersComplete = layersComplete && frame.ready;
      }
      if (updateRequired || !layersComplete) {
        if (!redrawRequested) {
          redrawRequested = true;
          requestAnimationFrame(doRedraw);
        }
        if (updateRequired || microsReady === microsDrawn) {
          return;
        }
      }
      microsDrawn = microsReady;
      for (i = 0; i < layers.length; i += 1) {
        layer = layers[i];
        frame = layer.frames[layer.currentFrame];
        if (!frame.ready) {
          // Try to find a frame that is ready if the current one is not.
          for (j = layer.currentFrame; j >= 0; j -= 1) {
            if (layer.frames[j].ready) {
              frame = layer.frames[j];
              layer.currentFrame = j;
              break;
            }
          }
        }
        blurRequired = blurRequired || !frame.ready;
        srcRect = [0, 0, frame.canvas.width, frame.canvas.height];
        dstRect = [
          layer.x * canvas.width / 100,
          layer.y * canvas.height / 100,
          layer.w * canvas.width / 100,
          layer.h * canvas.height / 100
        ];
        if (frame.ready || frame.microReady) {
          bufferCtx.drawImage(frame.canvas,
            srcRect[0], srcRect[1], srcRect[2], srcRect[3],
            dstRect[0], dstRect[1], dstRect[2], dstRect[3]);
        }
      }
      // Blur the canvas by applying a CSS blur filter.
      if (!layersComplete && blurRequired) {
        canvas.classList.add("blurred");
      } else if (layers.length > 0) {
        canvas.classList.remove("blurred");
      }
      if (fx.zoomFromRectData !== null) {
        fx.zoomFromRect(fx.zoomFromRectData.x, fx.zoomFromRectData.y,
          fx.zoomFromRectData.w, fx.zoomFromRectData.h, doDrawLayers);
        fx.zoomFromRectData = null;
      } else {
        // Draw the composite image to the main canvas.
        if (bewRetroMode) {
          retroCtx.drawImage(
            buffer, 0, 0, retroBuffer.width, retroBuffer.height);
          gfx.reduceColors(retroBuffer);
          if (typeof ctx.imageSmoothingEnabled !== "undefined") {
            ctx.imageSmoothingEnabled = false;
          } else {
            ctx.msImageSmoothingEnabled = false;
            ctx.mozImageSmoothingEnabled = false;
            ctx.webkitImageSmoothingEnabled = false;
          }
          if (!document.body.classList.contains("retro")) {
            document.body.classList.add("retro");
          }
          ctx.drawImage(retroBuffer, 0, 0, canvas.width, canvas.height);
        } else {
          if (document.body.classList.contains("retro")) {
            document.body.classList.remove("retro");
          }
          ctx.drawImage(buffer, 0, 0, canvas.width, canvas.height);
        }
      }
    }
    /** @param {BrewLayer} layer - The layer to load. */
    function loadLayer(layer) {
      var microData;
      var img = document.createElement("img");
      var microImg = document.createElement("img");
      var frame = layer.frames[layer.currentFrame];
      function drawFrame(image) {
        var width = image.naturalWidth;
        var height = image.naturalHeight;
        var ctx = frame.canvas.getContext("2d");
        var rect = frame.rect || [0, 0, 100, 100];
        rect = [
          rect[0] * width / 100,
          rect[1] * height / 100,
          rect[2] * width / 100,
          rect[3] * height / 100
        ];
        frame.canvas.width = rect[2];
        frame.canvas.height = rect[3];
        ctx.drawImage(image, rect[0], rect[1], rect[2], rect[3],
          0, 0, rect[2], rect[3]);
      }
      function microImageOnLoad() {
        // Delay the drawing of the micro-image by 200ms, this way the
        // artificial pop-in/flicker is reduced if the full image is
        // loaded in the meantime.
        setTimeout(function () {
          frame.microReady = true;
          // Only draw the micro-image if the frame is not yet ready.
          if (!frame.ready) {
            drawFrame(microImg);
          }
        }, 200);
      }
      if (!frame.ready && !frame.loading) {
        // Try to load the micro-image.
        microData = assetManager.requestImage(frame.id, "micro");
        if (microData) {
          microImg.addEventListener("load", microImageOnLoad);
          microImg.src = microData;
        }
        // Load the full image.
        img.addEventListener("load", function () {
          drawFrame(img);
          frame.loading = false;
          frame.ready = true;
        });
        frame.loading = true;
        img.src = assetManager.requestImage(frame.id);
      }
    }
    function loadLayers() {
      var i;
      for (i = 0; i < layers.length; i += 1) {
        loadLayer(layers[i]);
      }
      doDrawLayers();
    }
    /** @param {DOMHighResTimeStamp} timestamp - The current timestamp. */
    function updateAnimations(timestamp) {
      var i, delta, layer;
      var updateRequired = false;
      var drawRequired = layers.length === 1 && layers[0].frames.length === 1;
      for (i = 0; i < layers.length; i += 1) {
        layer = layers[i];
        if (!layer.frames[layer.currentFrame].ready) {
          // Do not progress the animation if the current frame is not ready.
          // This ensures that non-looping animations play their single cycle,
          // albeit probably slower than intended.
          if (!layer.frameStart) {
            drawRequired = true;
          }
          layer.frameStart = timestamp;
          updateRequired = true;
        } else if (layer.frames[layer.currentFrame].delay > 0 &&
            layer.frames.length > 1) {
          updateRequired = true;
          if (!layer.frameStart) {
            layer.frameStart = timestamp;
            drawRequired = true;
          }
          delta = timestamp - layer.frameStart;
          if (delta > layer.frames[layer.currentFrame].delay) {
            layer.frameStart = timestamp;
            layer.currentFrame = (layer.currentFrame + 1) % layer.frames.length;
            drawRequired = true;
          }
        }
      }
      if (drawRequired) {
        loadLayers();
      }
      if (updateRequired) {
        requestAnimationFrame(updateAnimations);
      }
      drawingLayers = false;
    }
    // Bail, if we are in testing mode.
    if (utMode) {
      return;
    }
    if (!drawingLayers) {
      drawingLayers = true;
      requestAnimationFrame(updateAnimations);
    }
  }
  /**
   * Evaluates {{condition;text}{...}} blocks in the specified text.
   * @param {string} text The input text.
   * @param {string} [context]
   *   The context to use. Defaults to the current page.
   * @param {boolean} [keepConditionalBlocks]
   *   Set to keep conditional blocks.Defaults to false.
   * @returns {string} The evaluated string.
   */
  function evaluateText(text, context, keepConditionalBlocks) {
    keepConditionalBlocks = keepConditionalBlocks || false;
    var result = text;
    context = context || getCurrentPage();
    /**
     * Resolves {{condition;text}{...}} blocks in the specified text.
     * @param {string} text The input text.
     * @returns {string} The text with resolved conditional blocks.
     */
    function resolveConditionalBlocks(text) {
      var s, result, outerMatch, innerMatch, cases, hadReturn;
      var rOuter = /\{\s*(\{(\w|\W)+?\})\s*\}/g;
      var rInner = /\{(((\w|\W)*?)\|)?((\w|\W)*?)\}/g;
      result = text;
      outerMatch = rOuter.exec(text);
      while (outerMatch !== null) {
        cases = [];
        hadReturn = false;
        innerMatch = rInner.exec(outerMatch[1]);
        while (innerMatch !== null) {
          s = innerMatch[4].replace(/"/g, "\\\"");
          if (!innerMatch[2]) {
            cases.push("return \"" + s + "\"");
            hadReturn = true;
          } else {
            cases.push(
              "if " + innerMatch[2] + " then return \"" + s + "\" end");
          }
          innerMatch = rInner.exec(outerMatch[1]);
        }
        if (!hadReturn) {
          cases.push("return \"\"");
        }
        cases = "function __inlineText as string " + cases.join(" ") + " end";
        $.compile(cases, "evaluateText()");
        s = $.call("__inlineText!", []);
        result = result.replace(outerMatch[0], s);
        outerMatch = rOuter.exec(text);
      }
      return result;
    }
    /**
     * Resolves all <@...> references in the specified text.
     * @param {string} text The input text.
     * @returns {string} The text with resolved references.
     */
    function resolveReferences(text) {
      var s, ref, match;
      var result = text;
      var rRef = /<@(.+?)>/g;
      match = rRef.exec(text);
      while (match !== null) {
        ref = "pages." + context + "." + match[1];
        s = $.requestString(ref);
        if (s === ref) {
          s = $.requestString(match[1]);
        }
        result = result.replace(match[0], s);
        match = rRef.exec(text);
      }
      return result;
    }
    /**
     * Prettifies the specified text. Converts BBCode, quotes, etc.
     * @param {string} text The input text.
     * @returns {string} The prettified text.
     */
    function prettify(text) {
      var result = text;
      /**
       * Converts quotes and apostrophes to a "prettier" representation.
       * @param {string} text The input text.
       * @returns {string} The text with prettified quotes.
       */
      function convertQuotes(text) {
        var i;
        var inTag = false;
        var inQuote = false;
        var result = "";
        for (i = 0; i < text.length; i += 1) {
          inTag = inTag || text[i] === "<";
          if (inTag) {
            result += text[i];
            inTag = text[i] === ">" ? false : inTag;
          } else {
            if (text[i] === "\"") {
              inQuote = !inQuote;
              result += inQuote
                ? $.requestString("formatting.quoteStart")
                : $.requestString("formatting.quoteEnd");
            } else if (text[i] === "'") {
              if (i > 0 && text[i - 1] === " ") {
                result += $.requestString("formatting.singleQuoteStart");
              } else if (i < text.length - 1 && text[i + 1] === " ") {
                result += $.requestString("formatting.singleQuoteEnd");
              } else {
                result += $.requestString("formatting.apostrophe");
              }
            } else {
              result += text[i];
            }
          }
        }
        return result;
      }
      /**
       * Converts BBCode to HTML+CSS.
       * @param {string} text The input text.
       * @param {boolean} noColor Set to exclude color tags from the conversion.
       * @returns {string} The converted text.
       */
      function convertBBCode(text, noColor) {
        var s, strBefore, strAfter;
        var result = typeof text === "undefined" ? "" : text;
        noColor = typeof noColor === "undefined" ? false : noColor;
        result = result.replace(/\[i\]/g, "<span class=\"italic\">");
        result = result.replace(/\[b\]/g, "<span class=\"bold\">");
        result = result.replace(/\[rip\]/g, "<span class=\"rip\">");
        result = result.replace(/\[\/(b|i|rip)\]/g, "</span>");
        result = result.replace(/\[\/color\]/g, noColor ? "" : "</span>");
        result = result.replace(/\[\/url\]/g, "</a>");
        // Replace all opening color tags.
        while (result.indexOf("[color=") > -1) {
          strBefore = result.slice(0, result.indexOf("[color="));
          s = result.slice(result.indexOf("[color=") + 7);
          strAfter = s.slice(s.indexOf("]") + 1);
          s = s.slice(0, s.indexOf("]"));
          result = strBefore +
            (noColor ? "" : "<span style=\"color: " + s + ";\">") + strAfter;
        }
        // Replace url tags.
        while (result.indexOf("[url=") > -1) {
          strBefore = result.slice(0, result.indexOf("[url="));
          s = result.slice(result.indexOf("[url=") + 5);
          strAfter = s.slice(s.indexOf("]") + 1);
          s = s.slice(0, s.indexOf("]"));
          if (strAfter.indexOf("</a>") === 0) {
            strAfter = s + strAfter;
          }
          result = strBefore + "<a href=\"" + s + "\">" + strAfter;
        }
        return result;
      }
      // Convert line-breaks.
      result = result.replace(/\n/g, "<br>");
      result = result.replace(/<br>/g, " <br>");
      // Prettify ellipses.
      result = result.replace(/\.{2,}/g, "... ");
      result = result.replace(/ {2,}/g, " ");
      result = result.replace("...", "…");
      // Convert quotes.
      result = convertQuotes(result);
      // Simple BBCode conversion.
      result = convertBBCode(result);
      return result;
    }
    // Resolve {{...}} constructs.
    if (!keepConditionalBlocks) {
      result = resolveConditionalBlocks(result);
    }
    // Resolve <@...> references.
    result = resolveReferences(result);
    // Prettify the string.
    result = prettify(result);
    return result;
  }
  function generate() {
    var currentPage = getCurrentPage();
    // Clear the previous page.
    clear();
    // Set up recording for the recently used variables.
    debugRelevantVariables = [];
    debugRecordTouches = true;
    // Set up the current page.
    if (typeof game.pages[currentPage] === "function") {
      game.pages[currentPage]();
    }
    update();
    // Disable the recording of recently used variables.
    debugRecordTouches = false;
  }
  function update(doPreloadEvents) {
    doPreloadEvents = typeof doPreloadEvents === "undefined"
      ? true
      : doPreloadEvents;
    var currentPage = getCurrentPage();
    var topTextEl = template.getTextContainer();
    var i;
    /** @param {BrewHotspot} index The index of the hotspot to generate. */
    function generateHotspot(index) {
      var el, textTemplate;
      var hotspot = hotspots[index];
      var text = evaluateText(hotspot.text);
      function onClickHandler() {
        // Record the choice.
        history.recordChoice("h", index + 1);
        // Call the callback function.
        if (typeof hotspot.callback === "function") {
          hotspot.callback();
        }
      }
      // Currently we only supports rects.
      if (text && hotspot.polygon.length === 4) {
        textTemplate = hotspot.text;
        if ($.predict(hotspot.callback).page === $.get("$endpage")) {
          textTemplate += "<br>⛔ <@sToBeContinuedFuture> ⛔";
        }
        el = template.addHotspot(hotspot.polygon, textTemplate, onClickHandler);
        if (el !== null && typeof hotspot.callback === "function") {
          if (debugMode.isEnabled() && checkFlag("showHints")) {
            el.setAttribute("data-bisk",
              bewUtilFormatFunctionHint(hotspot.callback));
          }
          el.addEventListener("mouseover", function () {
            bewUtilPredictNextImage(hotspot.callback);
          });
        }
      }
    }
    /** @param {BrewResponse} index The index of the response to generate. */
    function generateResponse(index) {
      var el, textTemplate;
      var response = responses[index];
      var text = evaluateText(response.text);
      function onClickHandler() {
        // Record the choice.
        history.recordChoice("r", index + 1);
        // Call the callback function.
        if (typeof response.callback === "function") {
          response.callback();
        }
      }
      // Do not add empty responses.
      if (text) {
        textTemplate = response.text;
        if ($.predict(response.callback).page === $.get("$endpage")) {
          textTemplate += "<br>⛔ <@sToBeContinuedFuture> ⛔";
        }
        el = template.addResponse(textTemplate, onClickHandler);
        if (el !== null && typeof response.callback === "function") {
          if (debugMode.isEnabled() && checkFlag("showHints")) {
            el.setAttribute("data-bisk",
              bewUtilFormatFunctionHint(response.callback));
          }
          el.addEventListener("mouseover", function () {
            bewUtilPredictNextImage(response.callback);
          });
        }
      }
    }
    clearPage();
    // Generate the page.
    $.bindText(topTextEl, topText);
    for (i = 0; i < hotspots.length; i += 1) {
      generateHotspot(i);
    }
    for (i = 0; i < responses.length; i += 1) {
      generateResponse(i);
    }
    drawLayers();
    if (typeof debugGenerateRelevantList === "function") {
      debugGenerateRelevantList();
    }
    if (doPreloadEvents) {
      for (i = 0; i < pageLoadEvents.length; i += 1) {
        if (pageLoadEvents[i].name === "*" ||
            pageLoadEvents[i].name === currentPage) {
          pageLoadEvents[i].callback();
        }
      }
    }
  }
  function getCurrentImage() {
    var result = "";
    if (layers.length > 0 && layers[0].frames.length > 0) {
      result = layers[0].frames[layers[0].currentFrame].id || "";
    }
    return result;
  }
  function getCurrentPage() {
    return history.getState().page || "";
  }
  function onPageLoad(name, callback) {
    if (typeof name === "string" && typeof callback === "function") {
      pageLoadEvents.push({
        name: name,
        callback: callback
      });
    }
  }
  function print(text) {
    topText = text;
  }
  function bindText(targetElement, textTemplate) {
    if (targetElement && textTemplate) {
      targetElement.setAttribute("brew-string", textTemplate);
      targetElement.innerHTML = evaluateText(textTemplate);
    }
  }
  function bindHint(targetElement, hintTemplate) {
    if (targetElement && hintTemplate) {
      targetElement.setAttribute("brew-title", hintTemplate);
      targetElement.title = evaluateText(hintTemplate).replace(/<br>/g, "\n");
    }
  }
  // Initialize the back buffer.
  buffer.width = canvas.width;
  buffer.height = canvas.height;
  // Initialize the buffer used by retro-mode.
  retroBuffer.width = 320;
  retroBuffer.height = 160;
  return {
    addHotspot: addHotspot,
    addResponse: addResponse,
    bindText: bindText,
    bindHint: bindHint,
    blit: blit,
    clear: clear,
    evaluateText: evaluateText,
    fx: fx,
    generate: generate,
    getCurrentImage: getCurrentImage,
    getCurrentPage: getCurrentPage,
    history: history,
    onPageLoad: onPageLoad,
    print: print,
    update: update
  };
}

// The global gfx object.
var gfx = (function () {
  var GfxModule = new Function("n", "t", "i",
    "'use asm';var r=new n.Uint8Array(i);function u(n,t){n=n|0;t=t|0;var u=0," +
    "o=0,i=0,f=0,s=0,h=0,e=0;for(f=n,s=(n|0)*2|0,o=0;(o|0)<=(t|0);o=o+4|0)for" +
    "(u=o|0;(u|0)<(o+3|0);u=u+1|0)h=(r[u]|0)>>5<<5|(r[u]|0)>>3,e=((r[u]|0)-(h" +
    "|0)|0)>>3,r[u]=h,i=(r[u+4|0]|0)+(e|0)|0,i=(i|0)<0?0:(i|0)>255?255:i|0,r[" +
    "u+4|0]=i,i=(r[u+8|0]|0)+(e|0)|0,i=(i|0)<0?0:(i|0)>255?255:i|0,r[u+8|0]=i" +
    ",i=(r[u+f-4|0]|0)+(e|0)|0,i=(i|0)<0?0:(i|0)>255?255:i|0,r[u+f-4|0]=i,i=(" +
    "r[u+f|0]|0)+(e|0)|0,i=(i|0)<0?0:(i|0)>255?255:i|0,r[u+f|0]=i,i=(r[u+f+4|" +
    "0]|0)+(e|0)|0,i=(i|0)<0?0:(i|0)>255?255:i|0,r[u+f+4|0]=i,i=(r[u+s|0]|0)+" +
    "(e|0)|0,i=(i|0)<0?0:(i|0)>255?255:i|0,r[u+s|0]=i}return{dither:u}"
  );
  var buffer = new ArrayBuffer(0x400000);
  var module = new GfxModule(window, {}, buffer);
  // Reduces the color-depth of the specified canvas.
  function reduceColors(canvas) {
    var i, imageData, arr;
    var ctx = canvas.getContext("2d");
    try {
      imageData = ctx.getImageData(0, 0, canvas.width, canvas.height);
      arr = new Uint8Array(buffer);
      arr.set(imageData.data);
      module.dither(imageData.width * 4, imageData.data.length);
      if (typeof Uint8ClampedArray !== "undefined" &&
          imageData.data instanceof Uint8ClampedArray) {
        imageData.data.set(arr.slice(0, imageData.data.length));
      } else {
        for (i = 0; i < imageData.data.length; i += 4) {
          imageData.data[i] = arr[i];
          imageData.data[i + 1] = arr[i + 1];
          imageData.data[i + 2] = arr[i + 2];
        }
      }
      ctx.putImageData(imageData, 0, 0);
    } catch (e) {
      // Do nothing.
    }
  }
  return {
    reduceColors: reduceColors
  };
}());

/**
 * Returns the variable object or null.
 * @param {string} name The name of the variable.
 * @returns {object} The requested variable-object.
 */
function bewGetVariable(name) {
  return game.variables[name] || null;
}

/**
 * Sets the specified `raw` data of the specified variable.
 * @param {string} name The name of the variable.
 * @param {(string|boolean|number)} value The raw value to be set.
 */
function bewSetVariable(name, value) {
  var variable = bewGetVariable(name);
  // Numbers and boolean values are stored as (floating point) numbers,
  // while strings and bitfields (bunches) are stored as strings.
  if (variable !== null) {
    if (variable.type === "number") {
      value = typeof value !== "undefined" ? value : 0;
      value = value === "" ? 0 : value;
      value = typeof value === "number" ? value : parseFloat(value);
      variable.value = value;
    } else if (variable.type === "boolean") {
      if (typeof value === "string") {
        variable.value = value === "true" ? 1 : 0;
      } else {
        variable.value = value ? 1 : 0;
      }
    } else {
      // Store all other types of variables as strings.
      value = typeof value !== "undefined" ? value : "";
      variable.value = value + "";
    }
  }
}

var bewDataSerializationPending = false;

// Serializes the game data.
function bewSerializeData() {
  function StorageEntry() {
    return {
      v: "3",  // Version number, stored as a hex-value.
      d: []
    };
  }
  function doSerializeData() {
    var i, k, saves, result, previous, variable, len;
    var names = Object.keys(game.variables);
    var sizeLimit = terminal.localStorageSupported ? 2 * 1024 * 1024 : 4000;
    var persistentData = new StorageEntry();
    var volatileData = new StorageEntry();
    for (i = 0; i < names.length; i += 1) {
      variable = bewGetVariable(names[i]);
      if (variable.value !== variable.defaultValue) {
        if (variable.persistent) {
          persistentData.d.push([names[i], variable.value]);
        } else {
          volatileData.d.push([names[i], variable.value]);
        }
      }
    }
    // First, backup the volatile data (the current state of the game).
    result = JSON.stringify(volatileData);
    if (sizeLimit > result.length) {
      sizeLimit -= result.length;
      // TODO: Cache the previous state of the storage variable.
      previous = readStorageString("V");
      if (result !== previous) {
        setStorageString("V", result);
      }
    } else {
      // If this is reached, it's alarming since volatile data should
      // not take up so much storage space that there would be no
      // additional space available for saves.
      throw new Error("Not enough storage space to backup volatile data.");
    }
    // Secondly, backup the persistent data (saved game states).
    result = JSON.stringify(persistentData);
    // Try and maintain the sizeLimit for locally stored data.
    len = result.length;
    if (len > sizeLimit) {
      saves = [];
      k = -1;
      for (i = 0; i < persistentData.d.length; i += 1) {
        if (persistentData.d[i][0] === "$saves") {
          saves = JSON.parse(persistentData.d[i][1]);
          k = i;
          break;
        }
      }
      if (k > -1) {
        while (len > sizeLimit) {
          // Remove the oldest save and subtract its size from the total size.
          // The +5 comes from the two escaped quotes and a comma (\"(.+)\",).
          len -= saves.shift().length + 5;
          if (saves.length < 1) {
            break;
          }
        }
        persistentData.d[k][1] = JSON.stringify(saves);
        result = JSON.stringify(persistentData);
      } else {
        // This really shouldn't happen but just in case it does, error out.
        throw new Error("Storage limit reached without any saves!");
      }
    }
    // TODO: Cache the previous state of the storage variable.
    previous = readStorageString("P");
    if (result !== previous) {
      setStorageString("P", result);
    }
    bewDataSerializationPending = false;
  }
  // Batch successive data-serialization requests together.
  if (!bewDataSerializationPending) {
    bewDataSerializationPending = true;
    requestAnimationFrame(doSerializeData);
  }
}

// Deserialize the game data.
function bewDeserializeData() {
  var i, version;
  var loggingState = bewDebugLoggingEnabled;
  var persistentData = readStorageString("P");
  var volatileData = readStorageString("V");
  function upgradeSavesToV3() {
    var i;
    var saves = bewGetSaves();
    for (i = 0; i < saves.length; i += 1) {
      saves[i] = bewUtilFixSave(saves[i]);
    }
    $.set("$saves", JSON.stringify(saves));
  }
  if (!game.initialized) {
    game.initialized = true;
    // Upgrade the variable-storage to the new system if necessary.
    if (!persistentData) {
      bewUpgradeStorage();
      persistentData = readStorageString("P");
      volatileData = readStorageString("V");
    }
    // If the data is not separated to :P and :V then load it from :Data.
    if (!persistentData) {
      persistentData = readStorageString("Data");
      setStorageString("Data", "");
    }
    if (persistentData) {
      persistentData = JSON.parse(persistentData);
      version = parseInt(persistentData.v, 16);
      persistentData = persistentData.d;
      if (version <= 3) {
        bewDebugLoggingEnabled = false;
        for (i = 0; i < persistentData.length; i += 1) {
          $.set(persistentData[i][0], persistentData[i][1]);
        }
        // Upgrade version 1 and version 2 saves.
        if (version <= 2) {
          // Trim unnecessary info from the saves.
          upgradeSavesToV3();
          // Serialize the game data and store it.
          bewSerializeData();
        }
        bewDebugLoggingEnabled = loggingState;
      }
    }
    if (volatileData) {
      volatileData = JSON.parse(volatileData);
      version = parseInt(volatileData.v, 16);
      volatileData = volatileData.d;
      if (version <= 3) {
        bewDebugLoggingEnabled = false;
        for (i = 0; i < volatileData.length; i += 1) {
          $.set(volatileData[i][0], volatileData[i][1]);
        }
        bewDebugLoggingEnabled = loggingState;
      }
    }
  }
}

// Sets the specified bitfiled's specified bit to the specified value.
function bewSetBitfieldValue(name, enumList, bit, value) {
  var i, k, c, data, result;
  var bitfield = $.get(name);
  var index = enumList.indexOf(bit);
  if (index < 0) {
    terminal.error("bewSetBitfieldValue() -> \"" + bit +
      "\" is not an element of " + name + ".");
    return;
  }
  data = Array.apply(null, new Array(((enumList.length + 3) >>> 2) * 4))
    .map(function () {
      return 0;
    });
  for (k = 0; k < bitfield.length; k += 1) {
    c = parseInt(bitfield.charAt(k), 16);
    for (i = 0; i < 4; i += 1) {
      data[k * 4 + i] = c & 1;
      c >>>= 1;
    }
  }
  data[index] = value ? 1 : 0;
  result = "";
  for (i = 0; i < data.length; i += 4) {
    c = data[i] | (data[i + 1] << 1) | (data[i + 2] << 2) | (data[i + 3] << 3);
    result += c.toString(16);
  }
  result = result.replace(/0+$/, "");
  $.set(name, result);
}

// Returns the specified bit of the specified bitfield.
function bewGetBitfieldValue(name, enumList, bit) {
  var bitfield = $.get(name);
  var index = enumList.indexOf(bit);
  if (index > bitfield.length * 4 - 1) {
    return 0;
  }
  bitfield = bitfield.charAt(index >> 2);
  index %= 4;
  return (parseInt(bitfield, 16) >> index) & 1;
}

// Convert the internal storage-system to the latest version.
function bewUpgradeStorage() {
  var i, item;
  var variables = readStorageString("Variables").split(":");
  var saves = [];
  var d = Date.now();
  // Returns the bitfield-based representation of the achievement-storage.
  function bewUpgradeAchievements() {
    var i, j, k, c, s;
    var achievements = getAchievementList();
    var arr = [];
    // This returns the value of the specified variable
    // from the legacy variable-storage.
    function readVariableStorage(storageName, name) {
      var i, str, value;
      var strings = readStorageString(storageName);
      if (strings !== "") {
        strings = strings.split(":");
        for (i = 0; i < strings.length; i += 1) {
          str = strings[i].split("=");
          if (str[0] === name) {
            value = parseInt(str[1], 10);
            value = isNaN(value) ? 0 : value;
            return value;
          }
        }
      }
      return 0;
    }
    // Returns the current state of the achievement.
    function getAchievementLegacy(name) {
      var index = achievements.indexOf(name);
      if (index > -1) {
        if (isActivity(name)) {
          return readVariableStorage("Variables", name);
        }
        return readVariableStorage("Achievements", name);
      }
      return 0;
    }
    for (i = 0; i < achievements.length; i += 1) {
      // First touch the achievement to cache its value.
      getAchievementLegacy(achievements[i]);
      arr.push(achievements[i]);
    }
    k = Math.ceil(arr.length / 4);
    while (arr.length < k * 4) {
      arr.push(0);
    }
    s = "";
    for (j = 0; j < k; j += 1) {
      c = 0;
      for (i = 0; i < 4; i += 1) {
        c += arr[j * 4 + i] << i;
      }
      s += c.toString(16);
    }
    s = s.replace(/0+$/, "");
    return s;
  }
  // Upgrades a single save to the new format.
  function bewUpgradeSave(save, date) {
    save = save.split(":");
    if (save[0].indexOf("#") > -1) {
      save[0] = save[0].split("#").join("|");
    } else {
      save[0] += "|";
    }
    save = date.toString(16) + "|" + save.join("|");
    save = save.split("|");
    // Date | page | label | <empty> | vars ...
    save[1] = bewCleanId(save[1]);
    save[3] = "";
    save = save.join("|");
    return save;
  }
  bewSetVariable("$achievements", bewUpgradeAchievements());
  for (i = 0; i < variables.length; i += 1) {
    item = variables[i].split("=");
    bewSetVariable(item[0], item[1]);
  }
  // Upgrade saves.
  item = readStorageString("Saveq");
  if (item) {
    bewSetVariable("$qsave", bewUpgradeSave(item, d));
    setStorageString("Saveq", "");
  }
  for (i = 1; i < 1000; i += 1) {
    item = readStorageString("Save" + i);
    if (item) {
      saves.push(bewUpgradeSave(item, d + i));
      // Clear the old save-data.
      setStorageString("Save" + i, "");
    }
  }
  bewSetVariable("$saves", JSON.stringify(saves));
  // Remove any still present old data.
  setStorageString("Achievements", "");
  setStorageString("Variables", "");
  // Serialize the game data and store it.
  bewSerializeData();
}

/**
 * Cleans up the specified identifier.
 * @param {string} id The identifier to clean.
 * @returns {string} The sanitized identifier.
 */
function bewCleanId(id) {
  return id.replace(/.*(\/|\\)(.*?)(\.(.*)|$)/, "$2").toLowerCase();
}

// Transforms JS VM code to Bisk.
function bewUtilJsToBisk(code) {
  code = code.replace(/! /g, "not ");
  code = code.replace(/ && /g, " and ");
  code = code.replace(/ \|\| /g, " or ");
  code = code.replace(/ === /g, " == ");
  code = code.replace(/ !== /g, " <> ");
  code = code.replace(/\( /g, "(");
  code = code.replace(/ \)/g, ")");
  code = code.replace(/return \$\.jmp\((.*?)\);?/g, "jmp $1");
  code = code.replace(/\$\.check\("(.*?)",\s*(.*?)\)/g, "$1 has $2");
  code = code.replace(/\$\.call\(\\?"(.*?)!.*?\\?",\s*\[(.*?)]\)/g, "!$1($2)");
  code = code.replace(
    /\$\.(set|add|sub|mul|div|mod)\("(.*?)",\s*(.+?)\);?/g, "$1 $2 $3");
  code = code.split(", ").join(" ");
  code = code.replace(/\(\$\._if\("(.*?)" (.*?)\)\)\s*\{/g, "$2 then");
  code = code.replace(/}\s*else\s+if/g, "elif");
  code = code.replace(/}\s*else\s*{/g, "else");
  code = code.replace(/\n}/g, "\nend");
  code = code.replace(/\$\.get\("(.*?)"\)/g, "$1");
  return code;
}

// Formats the source of the specified function.
function bewUtilFormatFunctionHint(func) {
  if (typeof func === "function") {
    func = func.toString();
  }
  func = func.replace(/^function[^(]*?\((\w|\W)*?\)\s*\{\s*/, "");
  func = func.replace(/\s*["']use strict["'];\s*/, "");
  func = func.replace(/\s*\}\s*$/, "");
  func = func.replace(/\s*;\s*/g, ";\n");
  func = func.replace(/;\s*$/, "");
  // JS -> Bisk transformation.
  func = bewUtilJsToBisk(func);
  return func;
}

// Annotates the provided input for highlighting.
function bewUtilAnnotateBisk(input) {
  // TODO: Implement proper parsing for Bisk highlighting.
  var result = (input + " ")
    .replace(/("(\w|\W)+?")/g, "<span class=\"string\">$1</span>")
    .replace(/(\s?)(true|false|as|if|then|else|elif|end)(\s)/g,
      "$1<span class=\"keyword\">$2</span>$3")
    .replace(/(\s?)(set|add|sub|mul|div|jmp)(\s)/g,
      "$1<span class=\"command\">$2</span>$3")
    .replace(/(\s)(not|and|or|has|<|<=|<>|>=|>|==)(\s)/g,
      "$1<span class=\"symbol\">$2</span>$3")
    .replace(/((!.+)\()/g, "<span class=\"function-call\">$2</span>(")
    .replace(/(\s)(\d+(.\d+)?)/g, "$1<span class=\"number\">$2</span>")
    .replace(/\s*$/, "");
  return result;
}

/**
 * Pre-caches the provided list of images.
 * @param {(string[]|string)} images The list of images to cache.
 */
function bewUtilPredictiveLoad(images) {
  var CACHE_LIMIT = 10;
  var container = document.getElementById("ImageCache");
  var i, src, el, list;
  /**
   * Returns a cached entry.
   * @param {string} id The identifier of the cache-entry.
   * @returns {Element} The requested cache-entry.
   */
  function getCacheEntry(id) {
    var j;
    for (j = 0; j < container.children.length; j += 1) {
      if (container.children[j].getAttribute("data-id") === id) {
        return container.children[j];
      }
    }
    return null;
  }
  images = typeof images === "string" ? [images] : images;
  list = images.map(bewCleanId);
  if (!container) {
    container = document.createElement("div");
    container.id = "ImageCache";
    container.style.display = "none";
    document.body.appendChild(container);
  }
  // Increase the age for all cached items.
  for (i = 0; i < container.children.length; i += 1) {
    el = container.children[i];
    el.setAttribute("data-age", Number(el.getAttribute("data-age")) + 1);
  }
  // Check if the image is cached already.
  for (i = list.length - 1; i >= 0; i -= 1) {
    el = getCacheEntry(list[i]);
    if (el) {
      // Reset the cache entry's age to 0 and its owner to the current page.
      el.setAttribute("data-age", 0);
      el.setAttribute("data-owner", $.getCurrentPage());
      // Remove the item from the list.
      list.splice(i, 1);
    }
  }
  // Cache all remaining items on the list.
  for (i = 0; i < list.length; i += 1) {
    src = assetManager.requestImage(list[i]);
    // TODO: Only prefetch new images if there are no images in-flight.
    if (src) {
      el = document.createElement("img");
      el.setAttribute("data-id", list[i]);
      el.setAttribute("data-age", 0);
      el.setAttribute("data-owner", $.getCurrentPage());
      el.src = src;
      container.insertBefore(el, container.firstChild);
    }
  }
  // Purge older entries from the cache.
  if (container.children.length > CACHE_LIMIT) {
    for (
      i = container.children.length - 1;
      i >= 0 && container.children.length > CACHE_LIMIT;
      i -= 1
    ) {
      el = container.children[i];
      // Only purge older items which are not owned by the current page.
      if (el.getAttribute("data-owner") !== $.getCurrentPage() &&
          Number(el.getAttribute("data-age")) > 0) {
        container.removeChild(el);
      }
    }
  }
}

// Predict the next image to load.
function bewUtilPredictNextImage(func) {
  bewUtilPredictiveLoad($.predict(func).images);
}

// Descrambles the specified image and draws it on the specified canvas.
function bewDescrambleImage(img, canvas) {
  function scRandom(range) {
    var m = 4294967296;
    var a = 1664525;
    var c = 1013904223;
    scRandSeed = (a * scRandSeed + c) % m;
    return Math.floor(range * (scRandSeed / m));
  }
  var x, y, k, temp, block, blocks, order;
  var ctx = canvas.getContext("2d");
  var scRandSeed = 5318008;
  var kBlockSize = 8;
  var kIterations = 256;
  var width = canvas.width;
  var height = canvas.height;
  var w = Math.ceil(width / kBlockSize) * kBlockSize;
  var h = Math.ceil(height / kBlockSize) * kBlockSize;
  for (y = 0; y < h / kBlockSize; y += 1) {
    blocks = [];
    order = [];
    for (x = 0; x < w / kBlockSize; x += 1) {
      block = {};
      block.x = x * kBlockSize;
      block.y = y * kBlockSize;
      blocks.push(block);
      order.push(x);
    }
    //if (y != h / kBlockSize) {
    for (x = 0; x < kIterations; x += 1) {
      k = Math.floor(1 + scRandom(Math.floor(width / kBlockSize) - 1));
      temp = order[0];
      order[0] = order[k];
      order[k] = temp;
    }
    //}
    for (x = 0; x < blocks.length; x += 1) {
      block = blocks[order[x]];
      ctx.drawImage(img, block.x, block.y, kBlockSize, kBlockSize,
        x * kBlockSize, block.y, kBlockSize, kBlockSize);
    }
  }
}

//#endregion

//#region --- _functions.js override functions ---

// Checks if the specified flag is enabled or not.
function checkFlag(name) {
  return $.check("$flags", name);
}

// Sets the specified flag to the specified value.
function setFlag(name, value) {
  if (value) {
    $.add("$flags", name);
  } else {
    $.sub("$flags", name);
  }
}

// Toggles the specified flag.
function toggleFlag(name) {
  setFlag(name, !checkFlag(name));
  return checkFlag(name);
}

/**
 * Initializes/resets all variables.
 * @param {boolean} [all]
 *   If true then it resets all persistent variables as well. Defaults to false.
 */
function initVars(all) {
  var i, variable;
  var names = Object.keys(game.variables);
  var loggingState = bewDebugLoggingEnabled;
  all = typeof all === "undefined" ? false : all;
  bewDebugLoggingEnabled = false;
  for (i = 0; i < names.length; i += 1) {
    variable = bewGetVariable(names[i]);
    if (!variable.persistent || all) {
      $.set(names[i], variable.defaultValue);
    }
  }
  bewDebugLoggingEnabled = loggingState;
}

// Returns an array containing all currently stored saves.
function bewGetSaves() {
  var saves = $.get("$saves");
  return saves === "" ? [] : JSON.parse(saves);
}

// Deletes the specified save.
function deleteSave(id) {
  var removedSave, saves;
  if (id === "q") {
    $.set("$qsave", "");
  } else if (id === "X") {
    $.set("$xsave", "");
  } else {
    saves = bewGetSaves();
    if (id >= 0 && id < saves.length) {
      removedSave = saves.splice(id, 1)[0];
      user.removeSave(removedSave.split("|")[0]);
    }
    $.set("$saves", JSON.stringify(saves));
  }
}

/**
 * Returns the current state (non-persistent variables) in save-string format.
 * @returns {string} The current state in save-string format.
 */
function getState() {
  var i, variable;
  var names = Object.keys(game.variables);
  var result = [];
  // Generate the save-string.
  for (i = 0; i < names.length; i += 1) {
    variable = bewGetVariable(names[i]);
    // Persistent variables don't need to be saved.
    if (variable.persistent) {
      continue;
    }
    // Only store the value if it's set.
    if (variable.value !== variable.defaultValue) {
      result.push(names[i] + "=" + variable.value);
    }
  }
  return result.join("|");
}

/**
 * Sets the current state to the provided one.
 * @param {string} state - The state in save-string format.
 */
function setState(state) {
  var i, item;
  var loggingState = bewDebugLoggingEnabled;
  bewDebugLoggingEnabled = false;
  // First, initialize the variables to their default state.
  initVars();
  // Then set the variables to their value in the provided state.
  state = state.split("|");
  for (i = 0; i < state.length; i += 1) {
    item = state[i].split("=");
    $.set(item[0], item.slice(1).join("="));
  }
  bewDebugLoggingEnabled = loggingState;
}

// Saves the current state with the specified id.
function saveState(id, page) {
  // Perfom the actual saving of the specified game-state.
  function doSaveState(data) {
    var saves;
    var d = Date.now().toString(16);
    data = d + "|" + data;
    if (id === "q") {
      $.set("$qsave", data);
    } else if (id === "X") {
      $.set("$xsave", data);
    } else {
      saves = bewGetSaves();
      if (id < saves.length) {
        saves.splice(id, 1);
      }
      saves.push(data);
      $.set("$saves", JSON.stringify(saves));
    }
  }
  // Date | Page | Description | <Empty> | Variables ...
  id = typeof id === "undefined" ? "" : id;
  page = typeof page === "undefined" ? $.getCurrentPage() : page;
  // Special case for the quicksave.
  page = page.indexOf("|") < 0 ? page + "|" : page;
  // Special case for the credits page.
  if (typeof bewSpecialSave === "object") {
    page = page.split("|");
    if (page.length > 1 && page[0].replace(/.*(\/|\\)/, "") !== page[1]) {
      bewSpecialSave[1] = page[1];
    }
    doSaveState(bewSpecialSave.join("|"));
  } else {
    // Use the state from $.history instead of the current state. This makes
    // $.add() and similar functions prossible not just on page transitions
    // but on page loads as well.
    doSaveState([page, "", $.history.getState().state].join("|"));
  }
  return false;
}

// Restores a previously saved state by the specified id.
function restoreState(id) {
  var page, save;
  id = typeof id === "undefined" ? "" : id;
  if (id === "q") {
    save = $.get("$qsave");
  } else {
    save = bewGetSaves()[id];
  }
  if (!save) {
    return;
  }
  // Date | page | label | <empty> | vars ...
  save = save.split("|");
  setState(save.slice(4).join("|"));
  page = save[1];
  if (page !== "") {
    setFlag("ingame", true);
    $.history.clear();
    $.jmp(page + "|transition=fade");
  }
}

// Fixes legacy saves.
function bewUtilFixSave(save) {
  var page, savedState;
  var loggingState = bewDebugLoggingEnabled;
  bewDebugLoggingEnabled = false;
  // Date | page | label | <empty> | vars ...
  save = save.split("|");
  // Remove the label if it's the same as the page.
  save[2] = save[2] === save[1] ? "" : save[2];
  // Clear the old image-idetifier field.
  save[3] = "";
  // Clean up the page-idetifier field.
  save[1] = bewCleanId(save[1]);
  page = save[1];
  if (page === "_credits") {
    // Fix the special case of Episode 2 saves on the credits screen.
    savedState = getState();
    setState(save.slice(4).join("|"));
    if ($.get("scene") === 9) {
      // Azumi Inclusion episode exit.
      page = "n1clubap013";
    } else if (($.shared.checkN(42) && $.get("ztemp1") > 4) ||
        ($.shared.checkN(26, 1) && $.get("ntempb") === 2)) {
      // EDD - Dinner Date episode exit.
      page = "n1club096";
    } else {
      // For all other cases, return to Brad's bedroom.
      page = "n1home007";
    }
    save[1] = page;
    setState(savedState);
  }
  bewDebugLoggingEnabled = loggingState;
  return save.join("|");
}

// Creates a cookie or localStorage entry and sets its value.
function setStorageString(name, value, legacy) {
  var delim, d, fullName;
  value = typeof value !== "undefined" ? value : "";
  legacy = typeof legacy !== "undefined" ? legacy : false;
  delim = typeof game.uid !== "undefined" ? game.uid : "X";
  fullName = legacy ? name + delim : delim + ":" + name;
  if (terminal.localStorageSupported) {
    if (value !== "") {
      localStorage.setItem(fullName, String(value));
    } else {
      localStorage.removeItem(fullName);
    }
  } else {
    d = new Date();
    if (value !== "") {
      d.setFullYear(d.getFullYear() + 10);
    } else {
      d.setFullYear(d.getFullYear() - 1);
    }
    document.cookie =
      fullName + "=" + value + "; expires=" + d.toUTCString() + "; path=/";
  }
}

// Reads the cookie or localStorage entry and returns its value as a string.
function readStorageString(name, legacy) {
  var i, val, delim, temp, fullName, strings, str;
  legacy = typeof legacy !== "undefined" ? legacy : false;
  delim = typeof game.uid !== "undefined" ? game.uid : "X";
  fullName = legacy ? name + delim : delim + ":" + name;
  val = "";
  if (terminal.localStorageSupported) {
    temp = localStorage.getItem(fullName);
    if (temp !== null) {
      val = String(temp);
    }
  } else {
    fullName += "=";
    strings = document.cookie.split(";");
    for (i = 0; i < strings.length; i += 1) {
      str = strings[i];
      while (str.charAt(0) === " ") {
        str = str.substring(1, str.length);
      }
      if (str.indexOf(fullName) === 0) {
        val = str.substring(fullName.length, str.length);
      }
    }
  }
  return val;
}

//#endregion

//#region ------ Dynamic Patching functions ------

// Returns the difference between the specified version and the current version.
// `version´ is a string and it must be in "#.#.#" format.
function bewCompareVersion(version) {
  var v = version.split(".");
  var c = bewVersion.split(".");
  if (v.length === 3 && c.length === 3) {
    c = c.map(Number);
    v = v.map(Number);
    if (c[0] === 0) {
      return 0;
    }
    v = (v[0] * 1000000) + (v[1] * 1000) + v[2];
    c = (c[0] * 1000000) + (c[1] * 1000) + c[2];
    return v - c;
  }
  return 0;
}

//#endregion

//#region -------- Localization functions --------

// Generates the diff between the default locale and all other locales.
function bewGenerateLocaleDiff() {
  var i, s, missing, unchanged, extra, wrapper, header, content, closeBtn;
  var el = document.getElementById("LocaleDiff");
  var lang = game.locales[game.activeLocale].strings;
  var keys = Object.keys(lang);
  var defLang = game.locales[game.defaultLocale].strings;
  var defKeys = Object.keys(defLang);
  var excludeList = [
    "sCreditsRoll"
  ];
  if (game.activeLocale !== game.defaultLocale) {
    missing = [];
    unchanged = [];
    for (i = 0; i < defKeys.length; i += 1) {
      if (excludeList.indexOf(defKeys[i]) < 0) {
        if (keys.indexOf(defKeys[i]) < 0) {
          missing.push(defKeys[i]);
        } else if (lang[defKeys[i]] === defLang[defKeys[i]]) {
          unchanged.push(defKeys[i]);
        }
      }
    }
    extra = [];
    for (i = 0; i < keys.length; i += 1) {
      if (defKeys.indexOf(keys[i]) < 0) {
        extra.push(keys[i]);
      }
    }
    s = [];
    if (missing.length > 0) {
      s.push("Missing keys from \"" + game.activeLocale + "\":");
      s.push("==========================\n");
      for (i = 0; i < missing.length; i += 1) {
        s.push(missing[i]);
      }
    }
    if (unchanged.length > 0) {
      if (s.length > 0) {
        s.push("\n");
      }
      s.push("Unchanged keys in \"" + game.activeLocale + "\":");
      s.push("==========================\n");
      for (i = 0; i < unchanged.length; i += 1) {
        s.push(unchanged[i]);
      }
    }
    if (extra.length > 0) {
      if (s.length > 0) {
        s.push("\n");
      }
      s.push("Extra keys in \"" + game.activeLocale + "\":");
      s.push("======================\n");
      for (i = 0; i < extra.length; i += 1) {
        s.push(extra[i]);
      }
    }
    s = s.join("\n");
    if (el === null) {
      // Create the wrapper div.
      wrapper = document.createElement("div");
      wrapper.id = "LocaleDiffWrapper";
      wrapper.className = "debug-diff-wrapper";
      // Create the header div.
      header = document.createElement("div");
      header.className = "debug-diff-header";
      header.innerHTML = "Diff for \"" + game.activeLocale + "\"";
      // Create the close button.
      closeBtn = document.createElement("div");
      closeBtn.className = "debug-diff-header-close-btn";
      closeBtn.innerHTML = "×";
      closeBtn.addEventListener("click", function () {
        var el = document.getElementById("LocaleDiffWrapper");
        el.parentNode.removeChild(el);
      });
      // Create the content div.
      content = document.createElement("div");
      content.className = "debug-diff-content";
      // Create the pre.
      el = document.createElement("pre");
      el.id = "LocaleDiff";
      // Add the elements to the tree.
      header.appendChild(closeBtn);
      content.appendChild(el);
      wrapper.appendChild(header);
      wrapper.appendChild(content);
      document.body.appendChild(wrapper);
    }
    el.innerHTML = s;
  }
}

//#endregion

//#region --------- UI upgrade functions ---------

function uiToggleMenu() {
  var menuButton = document.getElementById("MenuButton");
  var textTemplate;
  if (document.body.classList.contains("in-menu")) {
    document.body.classList.remove("in-menu");
    $.bindText(menuButton, "<@sMenu>");
  } else {
    textTemplate = checkFlag("ingame") ? "<@sResumeGame>" : "<@sMenu>";
    document.body.classList.add("in-menu");
    $.bindText(menuButton, textTemplate);
    uiGenerateMainMenu();
  }
}

// Clears the menu wrapper and sets it to the (optionally) specified class.
function uiClearMenu(className) {
  var menu = document.getElementById("Menu");
  var brDiv = document.getElementById("BottomMenu");
  if (brDiv) {
    brDiv.parentNode.removeChild(brDiv);
  }
  className = typeof className === "undefined"
    ? "menu-wrapper main-menu"
    : className;
  menu.innerHTML = "";
  menu.className = className;
}

function uiAddMenuItem(captionLabel, func) {
  var el = document.createElement("div");
  function defocusElement(e) {
    if (e.target) {
      e.target.blur();
    }
  }
  el.className = "menu-item";
  $.bindText(el, captionLabel);
  el.role = "button";
  el.tabIndex = 0;
  if (typeof func === "function") {
    el.addEventListener("click", defocusElement);
    el.addEventListener("click", func);
    el.addEventListener("keydown", bewKeyDownOnClick);
  }
  document.getElementById("Menu").appendChild(el);
}

function uiGenerateBottomMenu() {
  var a, div, textTemplate;
  var brDiv = document.getElementById("BottomMenu");
  if (brDiv) {
    brDiv.parentNode.removeChild(brDiv);
  }
  // Generate the bottom-right menu.
  brDiv = document.createElement("div");
  brDiv.className = "br-menu";
  brDiv.id = "BottomMenu";
  document.getElementById("Menu").parentNode.appendChild(brDiv);
  // Generate the bottom-right menu wrapper.
  div = document.createElement("div");
  div.className = "br-menu-wrapper";
  brDiv.appendChild(div);
  a = div;  // Store the wrapper-div in a.
  // Generate and add the version information.
  div = document.createElement("div");
  div.className = "version-info";
  div.innerHTML = bewVersionString();
  brDiv.appendChild(div);
  // Populate the bottom-right menu.
  brDiv = a;  // Make brDiv be the wrapper div.
  // Add the "Enable online features"-button (only for offline versions).
  if (bewCheckForUpdates && !bewOnlineVersion) {
    div = document.createElement("div");
    div.className = "br-menu-item" +
      (checkFlag("onlineEnabled") ? "" : " flicker");
    textTemplate = checkFlag("onlineEnabled")
      ? "<@sDisableOnlineFeatures>"
      : "<@sEnableOnlineFeatures>";
    $.bindText(div, textTemplate);
    div.addEventListener("click", function () {
      toggleFlag("onlineEnabled");
      uiGenerateMainMenu();
    });
    brDiv.appendChild(div);
  }
  // Add the "Join the conversation"-link.
  div = document.createElement("div");
  div.className = "br-menu-item";
  $.bindText(div, "<@sForum>");
  div.addEventListener("click", function () {
    var forumPage;
    var link = "http://bew.rocks/forum";
    if (typeof window.opener !== "undefined" && window.opener !== null) {
      forumPage = window.opener.open(link);
    } else {
      forumPage = window.open(link);
    }
    forumPage.focus();
  });
  brDiv.appendChild(div);
  // Add the "Support the development"-link.
  div = document.createElement("div");
  div.className = "br-menu-item";
  $.bindText(div, "<@sDonate>");
  div.addEventListener("click", function () {
    var donatePage;
    var link = "http://bew.rocks/donate";
    if (typeof window.opener !== "undefined" && window.opener !== null) {
      donatePage = window.opener.open(link);
    } else {
      donatePage = window.open(link);
    }
    donatePage.focus();
  });
  brDiv.appendChild(div);
}

// Generates the main menu.
function uiGenerateMainMenu() {
  var el;
  var patchInfoFuncName = "bewPatchInfo";
  function receivePatchInfo() {
    if (typeof window[patchInfoFuncName] === "function") {
      window[patchInfoFuncName]();
    }
  }
  uiClearMenu();
  if (checkFlag("ingame")) {
    uiAddMenuItem("<@sSaveGame>", generateSavePage);
  } else {
    uiAddMenuItem("<@sNewGame>", startNewGame);
  }
  uiAddMenuItem("<@sLoadGame>", generateRestorePage);
  uiAddMenuItem("<@sOptions>", uiGenerateOptionsMenu);
  uiAddMenuItem("<@sLanguage>", uiGenerateLanguageMenu);
  if (bewDebugMenuEnabled) {
    uiAddMenuItem("Debug Menu", uiGenerateDebugMenu);
  }
  if ($.shared.isAchievementSet("DM") || $.shared.isAchievementSet("d1")) {
    uiAddMenuItem("<@sBonusScenes>", uiGenerateBonusMenu);
  }
  uiAddMenuItem("<@sAchievements>", generateAchievementPage);
  uiAddMenuItem("<@sHelp>", uiGenerateHelpMenu);
  if (checkFlag("ingame")) {
    uiAddMenuItem("<@sQuitCurrentGame>", function () {
      $.jmp($.get("$startpage"));
      $.fx.fadeIn();
    });
  }
  uiGenerateBottomMenu();
  // Check for updates.
  if ((checkFlag("onlineEnabled") && bewCheckForUpdates) || bewOnlineVersion) {
    el = document.getElementById("patch-script");
    if (!el) {
      el = terminal.addScript(
        atob(checkFlag("betaFeatures")
          ? "aHR0cHM6Ly9icmFkc2Vyb3RpY3dlZWsucm9ja3MvdXBkYXRlL2JhY29uMi5qcw=="
          : "aHR0cHM6Ly9icmFkc2Vyb3RpY3dlZWsucm9ja3MvdXBkYXRlL2JhY29uLmpz") +
            "?r=" + Math.floor(Math.random() * 500).toString(),
        receivePatchInfo,
        false
      );
      el.id = "patch-script";
    } else {
      receivePatchInfo();
    }
  }
  // Get the info of the currently logged in user.
  user.updateUserInfo();
}

// Generates the bonus-menu.
function uiGenerateBonusMenu() {
  uiClearMenu();
  if ($.shared.isAchievementSet("DM")) {
    uiAddMenuItem("<@sBonusDemo>", function () {
      uiGenerateDemoBonusScene(0);
    });
  } else {
    uiAddMenuItem("<@sBonusLocked> - <@achievements.DM>");
  }
  if ($.shared.isAchievementSet("d1")) {
    uiAddMenuItem("<@sBonusDay1>", function () {
      uiGenerateDay1BonusScene(0);
    });
  } else {
    uiAddMenuItem("<@sBonusLocked> - <@achievements.d1>");
  }
  uiAddMenuItem("<@sBack>", uiGenerateMainMenu);
}

// Generates the options-menu.
function uiGenerateOptionsMenu() {
  uiClearMenu();
  if (typeof FileReader !== "undefined") {
    uiAddMenuItem("<@sExportSaves>", exportSaves);
    uiAddMenuItem("<@sImportSaves>", importSaves);
    uiAddMenuItem("<@sExportAchievements>", exportAchievements);
    uiAddMenuItem("<@sImportAchievements>", importAchievements);
  }
  uiAddMenuItem("<@sGenerateDebugFile>", replaySavePathData);
  uiAddMenuItem("<@sBack>", uiGenerateMainMenu);
}

// Generates the language-menu.
function uiGenerateLanguageMenu() {
  uiClearMenu();
  Object.keys(game.locales).sort().forEach(function (lang) {
    uiAddMenuItem(game.locales[lang].friendlyName, function () {
      assetManager.applyLocale(lang, function () {
        $.set("$lang", lang);
        document.getElementById("Content").lang = lang;
      });
    });
  });
  uiAddMenuItem("<@sBack>", uiGenerateMainMenu);
}

// Generates the help-menu.
function uiGenerateHelpMenu() {
  uiClearMenu();
  uiAddMenuItem("<@sInstructions>", function () {
    uiGenerateInstructions(0);
  });
  uiAddMenuItem("<@sReportABug>", uiGenerateBugReportPage);
  uiAddMenuItem("<@sBack>", uiGenerateMainMenu);
}

// Generates the debug-menu.
function uiGenerateDebugMenu() {
  uiClearMenu();
  uiAddMenuItem(
    (checkFlag("debugMode") ? "Disable" : "Enable") + " Debug Mode",
    function () {
      debugMode.toggle();
      uiGenerateDebugMenu();
    });
  uiAddMenuItem(
    (checkFlag("showHotspots") ? "Hide" : "Show") + " Hotspots",
    function () {
      if (toggleFlag("showHotspots")) {
        document.body.classList.add("show-hotspots");
      } else {
        document.body.classList.remove("show-hotspots");
      }
      uiGenerateDebugMenu();
    });
  uiAddMenuItem(
    (checkFlag("showHints") ? "Hide" : "Show") + " OnClick Hints",
    function () {
      toggleFlag("showHints");
      uiGenerateDebugMenu();
    });
  uiAddMenuItem(
    (checkFlag("spellChecking") ? "Disable" : "Enable") + " Spellchecking",
    function () {
      toggleFlag("spellChecking");
      if (debugMode.isEnabled()) {
        debugMode.enable();
      }
      uiGenerateDebugMenu();
    });
  uiAddMenuItem(
    (checkFlag("betaFeatures") ? "Disable" : "Enable") + " Beta Patches",
    function () {
      toggleFlag("betaFeatures");
      uiGenerateDebugMenu();
    });
  uiAddMenuItem("Diff locale", bewGenerateLocaleDiff);
  uiAddMenuItem("Reset Achievements", function () {
    // Reset all achievements.
    user.forceAchievements("");
  });
  uiAddMenuItem("Back", uiGenerateMainMenu);
}

// Calls the onclick event if the return or space key is pressed.
function bewKeyDownOnClick(e) {
  // 13 = Return, 32 = Space
  if ([13, 32].indexOf(e.which) > -1) {
    e.preventDefault();
    e.target.click();
  }
}

// Generates the header.
function bewGenerateHeader() {
  var loggingState = bewDebugLoggingEnabled;
  var pageHeader = document.getElementById("PageHeader");
  var menuContainer = document.createElement("div");
  var quickSave = document.createElement("div");
  var quickLoad = document.createElement("div");
  var menuBtn = document.createElement("div");
  function haveMet(id) {
    return (!!$.get(id + "happy") || !!$.get(id + "angry") ||
      !!$.get(id + "aroused") || !!$.get(id + "threat"));
  }
  function getActiveGirl() {
    var result = Math.floor($.get("path") / 10);
    // If the player hasn't yet met the active girl then default to Emily.
    if ((result === 2 && !haveMet("f")) ||
        (result === 3 && !haveMet("n")) ||
        (result === 4 && !haveMet("s")) ||
        (result === 5 && !haveMet("l")) ||
        (result === 6 && !haveMet("a")) ||
        (result === 7 && !haveMet("j")) ||
        (result === 8 && !haveMet("m"))) {
      result = 1;
    }
    return result;
  }
  /** @param {Element} wrapper - The wrapper element for a score bubble. */
  function updateScoreBubble(wrapper) {
    var name, index, value, met, icon, score, className, title;
    var activeGirl = getActiveGirl();
    if (wrapper) {
      name = wrapper.getAttribute("data-name") || "";
      index = wrapper.getAttribute("data-index") || "";
      if (name && index) {
        index = parseInt(index, 10);
        value = $.shared.calcPoints(name[0].toLowerCase());
        met = haveMet(name[0].toLowerCase());
        // TODO: Implement update(wrapper, "className", "...")!
        className = "hdr-score" +
          (activeGirl === index && met ? " active" : "");
        if (wrapper.className !== className) {
          wrapper.className = className;
        }
        title = met ? name : "???";
        if (wrapper.title !== title) {
          wrapper.title = title;
        }
        icon = wrapper.childNodes[0];
        score = wrapper.childNodes[1];
        className = "hdr-girl-icon icon-" + name + (met ? "" : " notmet");
        if (icon.className !== className) {
          icon.className = className;
        }
        if (score.textContent !== value) {
          score.textContent = value;
        }
      }
    }
  }
  var bubbles = null;
  function updateScoreBubbles() {
    var i;
    if (!bubbles) {
      bubbles = document.querySelectorAll(".hdr-menu-container .hdr-score");
    }
    if (bubbles) {
      for (i = 0; i < bubbles.length; i += 1) {
        updateScoreBubble(bubbles[i]);
      }
    }
  }
  function addScoreBubble(parent, name, index) {
    var wrapper = document.createElement("div");
    var icon = document.createElement("div");
    var score = document.createElement("div");
    index = index || 0;
    //wrapper.className = "hdr-score";
    wrapper.setAttribute("data-name", name);
    wrapper.setAttribute("data-index", index);
    icon.className = "hdr-girl-icon icon-" + name;
    score.className = "hdr-score-number";
    wrapper.appendChild(icon);
    wrapper.appendChild(score);
    parent.appendChild(wrapper);
    updateScoreBubble(wrapper);
  }
  bewDebugLoggingEnabled = false;
  pageHeader.innerHTML = "";
  menuContainer.className = "hdr-menu-container";
  quickSave.className = "hdr-quicksave";
  $.bindHint(quickSave, "<@sQuickSave>");
  quickSave.role = "button";
  quickSave.tabIndex = 0;
  quickSave.addEventListener("click", function () {
    if (checkFlag("ingame")) {
      saveState("q");
    }
  });
  quickSave.addEventListener("keydown", bewKeyDownOnClick);
  quickLoad.className = "hdr-quickload";
  $.bindHint(quickLoad, "<@sQuickLoad>");
  quickLoad.role = "button";
  quickLoad.tabIndex = 0;
  quickLoad.addEventListener("click", function () {
    // Reset the replay system and restore the saved state.
    $.history.clear();
    restoreState("q");
    // TODO: Handle hiding the menu more elegantly while loading the quicksave.
    $.fx.enableDiv("#000");
    document.body.classList.remove("in-menu");
    $.bindText(menuBtn, "<@sMenu>");
  });
  quickLoad.addEventListener("keydown", bewKeyDownOnClick);
  menuBtn.className = "hdr-menu-btn";
  menuBtn.id = "MenuButton";
  $.bindText(menuBtn, "<@sMenu>");
  menuBtn.role = "button";
  menuBtn.tabIndex = 0;
  menuBtn.addEventListener("click", function () {
    // TODO: Do not bind this behavior to the start-page after the menu rework.
    if ($.getCurrentPage() !== $.get("$startpage")) {
      uiToggleMenu();
    }
  });
  menuBtn.addEventListener("keydown", bewKeyDownOnClick);
  // ... left scores ...
  addScoreBubble(menuContainer, "Emily", 1);
  addScoreBubble(menuContainer, "Faith", 2);
  addScoreBubble(menuContainer, "Natalie", 3);
  addScoreBubble(menuContainer, "Serenthia", 4);
  menuContainer.appendChild(quickSave);
  menuContainer.appendChild(menuBtn);
  menuContainer.appendChild(quickLoad);
  // ... right scores ...
  addScoreBubble(menuContainer, "Lorien", 5);
  addScoreBubble(menuContainer, "Azumi", 6);
  addScoreBubble(menuContainer, "Jasmine", 7);
  addScoreBubble(menuContainer, "Muirgheal", 8);
  pageHeader.appendChild(menuContainer);
  // Register the custom callback for variable updates.
  $.onVariableChange("*", updateScoreBubbles);
  bewDebugLoggingEnabled = loggingState;
}

//#endregion

//#region --- Automatically executed functions ---

var bewScriptsLoaded = 0;

function bewScriptOnLoadEvent() {
  var target = null;
  var page = location.hash.replace("#", "");
  bewScriptsLoaded += 1;
  if (bewScriptsLoaded === game.scripts.length) {
    // Populate the values for all variables.
    bewDeserializeData();
    // Get the info of the currently logged in user.
    user.updateUserInfo();
    // Set debug vars based on config options.
    bewDebugMenuEnabled = bewDebugMenuEnabled || checkFlag("showDebugMenu");
    if (bewDebugMenuEnabled && checkFlag("debugMode")) {
      debugMode.enable();
    }
    assetManager.applyLocale($.get("$lang") || game.defaultLocale, function () {
      showScore();
      debug();
      bewDebugInspectorAnnotateAchievements();
      if (debugMode.isEnabled() && checkFlag("showHotspots")) {
        document.body.classList.add("show-hotspots");
      }
      // Generate all additional localization-dependent UI elements.
      bewGenerateHeader();
      // Set the game's state to running.
      game.running = true;
      document.body.classList.remove("loading");
      page = !game.pages[page] ? $.get("$startpage") : page;
      $.jmp(page);
      $.fx.fadeIn();
      // If we're in a frame, send a message that we are loaded.
      if (window.top !== window.self) {
        target = parent.postMessage
          ? parent
          : parent.document.postMessage
            ? parent.document
            : target;
        if (target) {
          target.postMessage("bew|loaded", "*");
        }
      }
    });
  }
}

function bewLoadGame() {
  function loadStyleSheet(url) {
    var el = document.createElement("link");
    el.rel = "stylesheet";
    el.addEventListener("load", function () {
      terminal.print("... " + url + " loaded.");
    });
    el.addEventListener("error", function () {
      terminal.error("Error loading file: " + url);
    });
    terminal.print("Loading " + url + "...");
    // TODO: Implement better resource versioning.
    el.href = url + "?v=" + bewVersion;
    document.head.appendChild(el);
  }
  function loadScript(src, callback) {
    // TODO: Implement better resource versioning.
    terminal.addScript(src + "?v=" + bewVersion, callback);
  }
  // Add the main style-sheet.
  loadStyleSheet("Files/_style.css");
  // Load the fonts.
  loadStyleSheet("assets/fonts.css");
  // Load the game-data.
  loadScript("Files/_gamedata.js", function () {
    var i;
    for (i = 0; i < game.scripts.length; i += 1) {
      loadScript(game.scripts[i], bewScriptOnLoadEvent);
    }
  });
}

function bewRegisterPageLoadHandlers() {
  $.onPageLoad("*", function () {
    var currentPage = $.getCurrentPage();
    var endPage = $.get("$endpage");
    if (currentPage === endPage) {
      // Dump the replay-data if enabled.
      replayDumpHistory();
    }
    template.getHotspotContainer().parentNode.classList.remove("dream-fx");
  });
  $.onPageLoad("_begin", function () {
    var menuBtn = document.getElementById("MenuButton");
    $.shared.updateAchievements();
    setFlag("ingame", false);
    if (!document.body.classList.contains("in-menu")) {
      uiToggleMenu();
    }
    $.bindText(menuBtn, "<@sMenu>");
  });
  $.onPageLoad("_credits", function () {
    var el = document.createElement("div");
    var pre = document.createElement("pre");
    $.bindText(pre, "<@sCreditsRoll>");
    el.className = "credits-wrapper";
    el.appendChild(pre);
    template.getHotspotContainer().appendChild(el);
  });
  $.onPageLoad("d1recap01", function () {
    template.getHotspotContainer().parentNode.classList.add("dream-fx");
  });
  $.onPageLoad("difficulty", function () {
    var tdList = document.getElementsByClassName("response");
    function diffSetImage(id) {
      if ($.getCurrentImage() !== id) {
        $.blit(id, 0);
        $.update();
      }
    }
    tdList[0].addEventListener("mouseover", function () {
      diffSetImage("diffeasy");
    });
    tdList[1].addEventListener("mouseover", function () {
      diffSetImage("diffmed");
    });
    tdList[2].addEventListener("mouseover", function () {
      diffSetImage("diffhard");
    });
  });
  $.onPageLoad("d1office004", function () {
    var ticCurrentSymbol = "X";
    var ticGameOver = false;
    function ticCreateReward() {
      var k = 100 / 18;
      $.blit(
        "d1o4-eeanim-strip:0," + (k * 0) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 1) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 2) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 3) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 4) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 5) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 6) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 7) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 8) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 9) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 10) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 11) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 12) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 13) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 14) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 15) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 16) + ",100," + k + ":100|" +
        "d1o4-eeanim-strip:0," + (k * 17) + ",100," + k + ":0" +
        "@0,0,34,11", 1);
      $.update(false);
    }
    function ticCheckWin() {
      var i;
      var fields = document.getElementsByClassName("tictactoe-field");
      function checkWin(a, b, c) {
        var won = (fields[a].innerHTML !== "") &&
          ((fields[a].innerHTML === fields[b].innerHTML) &&
          (fields[a].innerHTML === fields[c].innerHTML));
        if (won) {
          fields[a].className = "tictactoe-field won";
          fields[b].className = "tictactoe-field won";
          fields[c].className = "tictactoe-field won";
          return true;
        }
        return false;
      }
      for (i = 0; i < 3; i += 1) {
        // Check rows.
        if (checkWin(i * 3, i * 3 + 1, i * 3 + 2)) {
          return true;
        }
        // Check columns.
        if (checkWin(i, i + 3, i + 6)) {
          return true;
        }
      }
      // Check crosses.
      if (checkWin(0, 4, 8)) {
        return true;
      }
      if (checkWin(2, 4, 6)) {
        return true;
      }
      // Check if the game-field if full.
      for (i = 0; i < fields.length; i += 1) {
        if (fields[i].innerHTML === "") {
          return false;
        }
      }
      return true;
    }
    function ticCheckReward() {
      var fields = document.getElementsByClassName("tictactoe-field");
      var reward = (fields[0].innerHTML === "X") &&
        (fields[1].innerHTML === "O") &&
        (fields[2].innerHTML === "X") &&
        (fields[3].innerHTML === "X") &&
        (fields[4].innerHTML === "O") &&
        (fields[5].innerHTML === "X") &&
        (fields[6].innerHTML === "O") &&
        (fields[7].innerHTML === "X") &&
        (fields[8].innerHTML === "O");
      if (reward) {
        ticCreateReward();
      }
    }
    function ticResetField() {
      var i, div;
      var container = document.getElementById("TicTacToe");
      while (container.firstChild) {
        container.removeChild(container.firstChild);
      }
      for (i = 0; i < 9; i += 1) {
        div = document.createElement("div");
        div.className = "tictactoe-field";
        div.addEventListener("click", ticHandleStep);
        container.appendChild(div);
      }
    }
    function ticHandleStep(e) {
      var div = e.target;
      if (div.innerHTML === "" && !ticGameOver) {
        div.innerHTML = ticCurrentSymbol;
        ticCurrentSymbol = ticCurrentSymbol === "O" ? "X" : "O";
        ticGameOver = ticCheckWin();
        if (ticGameOver) {
          ticCheckReward();
        }
      } else if (ticGameOver) {
        ticGameOver = false;
        ticResetField();
      }
    }
    function ticStartTicTacToe(e) {
      e.target.removeEventListener("click", ticStartTicTacToe);
      e.target.title = "";
      ticResetField();
    }
    function generateTicTacToe() {
      var wrapper = document.createElement("div");
      var container = document.createElement("div");
      wrapper.id = "TicTacToeWrapper";
      wrapper.className = "tictactoe-wrapper";
      wrapper.appendChild(container);
      template.getHotspotContainer().appendChild(wrapper);
      template.getHotspotContainer().classList.add("easter-egg");
      container.id = "TicTacToe";
      container.className = "tictactoe";
      container.title = "Click!";
      container.addEventListener("click", ticStartTicTacToe);
    }
    generateTicTacToe();
  });
}

function bewRegisterVariableChangeCallbacks() {
  var regenRequested = false;
  // Updates the text of the specified variable's debug field.
  function debugUpdateVarField(name, value) {
    var td = document.getElementById("DebugVar-" + name);
    if (td !== null) {
      value = parseInt(value, 10);
      if (!value) {
        td.setAttribute("class", "value");
      } else {
        td.setAttribute("class", "value set");
      }
      td.innerHTML = name + " = " + value;
    }
  }
  // Updates the class of the specified achievement's debug field.
  // The value will be set to the value of "achieved".
  function debugUpdateAchievementField(name, achieved) {
    var td = document.getElementById("DebugAchievements-" + name);
    if (td !== null) {
      if (achieved === 0) {
        td.setAttribute("class", "value");
        td.innerHTML = name + " = 0";
      } else {
        td.setAttribute("class", "value set");
        td.innerHTML = name + " = 1";
      }
    }
  }
  // Updates the class of the specified events's debug field.
  // The value will be set to the value of "enabled".
  function debugUpdateEventField(id, enabled) {
    var td = document.getElementById("DebugEvents-" + id.replace(/[.\s]+/g, "-"));
    if (td !== null) {
      if (enabled) {
        td.setAttribute("class", "value set");
        td.innerHTML = id + " = 1";
      } else {
        td.setAttribute("class", "value");
        td.innerHTML = id + " = 0";
      }
    }
  }
  $.onVariableChange("*", function (name) {
    function regeneratePage() {
      $.generate();
      regenRequested = false;
    }
    var variable = bewGetVariable(name);
    debugUpdateVarField(name, variable.value);
    // Update the scores.
    showScore();
    // Schedule a page re-generation.
    if (!regenRequested) {
      regenRequested = true;
      requestAnimationFrame(regeneratePage);
    }
  });
  $.onVariableChange("_events", function (name) {
    var i, element;
    var variable = bewGetVariable(name);
    for (i = 0; i < variable.elements.length; i += 1) {
      element = variable.elements[i];
      debugUpdateEventField(element,
        bewGetBitfieldValue(name, variable.elements, element));
    }
  });
  $.onVariableChange("$achievements", function (name) {
    var i, element;
    var variable = bewGetVariable(name);
    for (i = 0; i < variable.elements.length; i += 1) {
      element = variable.elements[i];
      debugUpdateAchievementField(element,
        bewGetBitfieldValue(name, variable.elements, element));
    }
  });
  $.onVariableChange("$achievements", function (name, previousValue) {
    var i, current, previous, element;
    var variable = bewGetVariable(name);
    for (i = 0; i < variable.elements.length; i += 1) {
      element = variable.elements[i];
      previous = i > previousValue.length * 4 - 1
        ? false
        : ((parseInt(previousValue.charAt(i >> 2), 16) >> (i % 4)) & 1) === 1;
      current = i > variable.value.length * 4 - 1
        ? false
        : ((parseInt(variable.value.charAt(i >> 2), 16) >> (i % 4)) & 1) === 1;
      if (current && current !== previous && !isActivity(element)) {
        // Add the achievement pop-up.
        achievementAddPopup(element);
      }
    }
  });
}

// PlaybackTool automation.
function bewAutomation(e) {
  var page, state;
  var msg = typeof e.data === "string" ? e.data.split("|") : ["xXx"];
  if (msg[0] === "pt" && !!msg[1]) {
    if (msg[1] === "init") {
      //document.body.classList.remove("debug");
    } else if (msg[1] === "jump" && !!msg[2]) {
      page = msg[2];
      // Fix and re-route no longer existing pages.
      page = page === "_startmenu" ? "intro1" : page;
      // Restore state.
      state = msg.slice(3).join("|");
      setState(state);
      // Clear the history.
      $.history.clear();
      // Jump to the page.
      $.jmp(page);
    } else if (msg[1] === "highlight" && !!msg[2] && !!msg[3]) {
      // TODO: Mark the selected hotspot or response for highlighting.
      // HACK:
      //   The page is potentially regenerated on the next frame, so highlight
      //   the selected hotspot or response on the frame after that.
      requestAnimationFrame(function () {
        requestAnimationFrame(function () {
          var list;
          var el = null;
          var index = parseInt(msg[3], 10) - 1;
          if (msg[2] === "h") {
            // Select msg[3]th hotspot.
            list = document.getElementsByClassName("hotspot");
            index = list.length - index - 1;
            if (!!list && index < list.length) {
              el = list[index];
            }
          } else if (msg[2] === "r") {
            // Select msg[3]th response.
            list = document.getElementsByClassName("response");
            if (!!list && index < list.length) {
              el = list[index];
            }
          }
          if (el) {
            el.style.boxShadow = "0 0 0 3px #dd0 inset";
          }
        });
      });
    }
  }
}

var bewKeyCache = "";

// This is the keydown event handler for the game-window.
function bewOnKeyDown(e) {
  var key = e.which || e.keyCode;
  var c = String.fromCharCode(key).toUpperCase();
  bewKeyCache += c;
  if (bewKeyCache.length > 4) {
    bewKeyCache = bewKeyCache.substr(bewKeyCache.length - 5);
    if (bewKeyCache.indexOf("IDDQD") > -1) {
      bewDebugMenuEnabled = toggleFlag("showDebugMenu");
    }
    if (bewKeyCache.indexOf("RETRO") > -1) {
      bewRetroMode = !bewRetroMode;
      $.generate();
    }
  }
}

// This is the resize-event for the game-window.
function bewResizeEvent() {
  var w, h, fontSize;
  var el = document.getElementById("ContentWrapper") || null;
  var width = Math.max(window.innerWidth, window.innerHeight);
  var height = Math.min(window.innerWidth, window.innerHeight);
  h = height - Math.max(270, 100 + (200 * height / 1000));
  if (h > 425) {
    w = Math.min(h * 200 / width, 100);
  } else {
    w = 100;
    h = window.innerWidth / 2;
  }
  fontSize = Math.max(75, h * 100 / 625);
  // Set the width and font size for game.
  if (el) {
    el.style.fontSize = fontSize + "%";
    el.style.width = w + "%";
  }
}

// Execution of these functions is moved here from the html file.
(function () {
  var doResize = false;
  bewResizeEvent();
  window.addEventListener("resize", function () {
    // Throttle resize events.
    if (!doResize) {
      doResize = true;
      requestAnimationFrame(function () {
        bewResizeEvent();
        doResize = false;
      });
    }
  });
  window.addEventListener("keydown", bewOnKeyDown);
  window.addEventListener("message", bewAutomation);
  window.addEventListener("scroll", function () {
    var y = document.body.scrollY || window.pageYOffset;
    var footer = document.querySelector(".page-footer") || null;
    var brMenu = document.querySelector(".br-menu") || null;
    if (footer) {
      footer.style.display = y > 0 ? "none" : "";
    }
    if (brMenu) {
      brMenu.style.display = y > 0 ? "none" : "";
    }
  });
  window.addEventListener("load", function () {
    bewRegisterPageLoadHandlers();
    bewRegisterVariableChangeCallbacks();
    bewLoadGame();
  });
}());

//#endregion

//#region ------ Replay specific functions -------

// TODO: Rename replay* functions.

// Dump the replay history.
function replayDumpHistory() {
  var request;
  var data = $.history.serialize();
  if (bewDumpSession) {
    data = "session=" + encodeURI(data);
    if (data.length > 25 && typeof XMLHttpRequest !== "undefined") {
      request = new XMLHttpRequest();
      request.open(
        "POST",
        atob("aHR0cHM6Ly9icmFkc2Vyb3RpY3dlZWsucm9ja3MvZHVtcC9kdW1wLnBocA=="),
        true
      );
      request.setRequestHeader("Content-Type",
        "application/x-www-form-urlencoded; charset=UTF-8");
      request.send(data);
    }
  }
}

// Saves the path data from the current replay.
function replaySavePathData() {
  var i, d, fileName, variable;
  var names = Object.keys(game.variables);
  var output = $.history.serialize();
  var states = JSON.parse(output);
  function pad(n) {
    var s = n.toString();
    if (s.length < 2) {
      s = "0" + s;
    }
    return s;
  }
  if (states !== null) {
    // After the JSON data, append a human readable version.
    output += "\n\n--- Path Trace ---\n";
    for (i = 0; i < states.length; i += 1) {
      output += states[i].page + " --- " + states[i].state + "\n";
    }
    output += "\n--- Variable Dump ---\n";
    for (i = 0; i < names.length; i += 1) {
      variable = bewGetVariable(names[i]);
      output += names[i] + " = " + variable.value + "\n";
    }
    if (output !== "") {
      d = new Date();
      fileName = $.getCurrentPage() + "-" + d.getUTCFullYear() +
        pad(d.getUTCMonth() + 1) + pad(d.getUTCDate()) + "-" +
        pad(d.getUTCHours()) + pad(d.getUTCMinutes()) + ".txt";
      saveTextAsFile(output, fileName);
    }
  }
}

//#endregion

//#region ----- Debug Log specific functions -----

var bewDebugLoggingEnabled = false;
var bewDebugLog = [];

// Display the debug info for performed checks.
function debugGenerateLog() {
  var i, k, item, div, indent;
  var containerDiv = document.getElementById("DebugLog");
  var debugDiv = document.getElementById("DebugInspector");
  if (bewDebugLog.length > 0 && !!debugDiv) {
    if (!containerDiv) {
      containerDiv = document.createElement("div");
      containerDiv.id = "DebugLog";
      containerDiv.className = "debug-log";
      debugDiv.parentNode.insertBefore(containerDiv, debugDiv);
    } else {
      containerDiv.innerHTML = "";
    }
    for (i = 0; i < bewDebugLog.length; i += 1) {
      item = bewDebugLog[i];
      indent = "";
      for (k = 0; k < item.level; k += 1) {
        indent += "  ";
      }
      div = document.createElement("div");
      div.className = item.className;
      div.innerHTML = indent + item.text;
      containerDiv.appendChild(div);
    }
  } else if (containerDiv) {
    containerDiv.innerHTML = "";
  }
  bewDebugLog = [];
}

/**
 * Adds a new line to the debug-log (call-log).
 * @param {string} text The initial text to add.
 * @param {object} [options] Additinal options for the line.
 * @returns {object} The generated debug-log line-object.
 */
function debugLogAddLine(text, options) {
  var result = null;
  if (bewDebugLoggingEnabled) {
    result = options || {
      level: 0,
      className: ""
    };
    result.level = result.level || 0;
    result.className = result.className || "";
    result.text = text;
    bewDebugLog.push(result);
  }
  return result;
}

//#endregion

//#region ------- General helper functions -------

/**
 * Saves the specified text to the specified file.
 * @param {string} text - The text-data to be saved.
 * @param {string} fileName - The name of the output file.
 */
function saveTextAsFile(text, fileName) {
  var a, textBlob;
  if (typeof Blob === "undefined") {
    return;
  }
  textBlob = new Blob(
    [text],
    {
      type: "text/plain",
      endings: "native"
    }
  );
  // IE has a custom way of saving Blobs.
  if (typeof navigator.msSaveBlob !== "undefined") {
    navigator.msSaveBlob(textBlob, fileName);
  } else {
    a = document.createElement("a");
    a.download = fileName;
    a.href = URL.createObjectURL(textBlob);
    a.addEventListener("click", function (e) {
      document.body.removeChild(e.target);
    });
    a.style.display = "none";
    document.body.appendChild(a);
    a.click();
  }
}

/**
 * Prompts the user for a file then returns its contents as text.
 * @param {function} callback - A callback function to process the input data.
 * @remarks The callback function should have a single argument where the
 *   text data will be passed to it.
 */
function loadFileAsText(callback) {
  var input = document.getElementById("HiddenFileInput");
  if (input !== null) {
    input.parentNode.removeChild(input);
  }
  input = document.createElement("input");
  input.id = "HiddenFileInput";
  input.style.display = "none";
  input.type = "file";
  //input.setAttribute("accept", "text/plain");
  input.addEventListener("change", function (event) {
    var fileReader = new FileReader();
    fileReader.onload = function (e) {
      callback(e.target.result);
    };
    fileReader.readAsText(event.target.files[0], "UTF-8");
  });
  document.body.appendChild(input);
  input.click();
}

//#endregion

//#region ------ Backup specific functions -------

// Exports the current set of saves to a file.
function exportSaves() {
  var i, k, save, saveData;
  var saveList = [];
  var saves = bewGetSaves();
  for (i = 0; i < saves.length; i += 1) {
    saveData = saves[i].split("|");
    save = {};
    save.date = saveData[0];
    save.page = saveData[1];
    save.label = saveData[2];
    save.state = [];
    for (k = 4; k < saveData.length; k += 1) {
      save.state.push(saveData[k]);
    }
    saveList.push(save);
  }
  saveTextAsFile(JSON.stringify(saveList), "bew-saves.json");
}

// Imports a set of saves from a file and replaces the current set with it.
function importSaves() {
  loadFileAsText(function (str) {
    var i, data;
    var saves = [];
    var d = Date.now();
    var saveList = JSON.parse(str);
    // Clear the set of saves.
    $.set("$saves", "");
    // Add the new set of saves.
    for (i = 0; i < saveList.length; i += 1) {
      if (typeof saveList[i].date === "undefined") {
        // Convert version 1 to version 2.
        // { name[, imgSrc], state }
        saveList[i].date = (d + 1).toString(16);
        saveList[i].name = saveList[i].name.split("#");
        saveList[i].page = bewCleanId(saveList[i].name[0]);
        saveList[i].label = saveList[i].name[1];
        saveList[i].state = saveList[i].state.split(":");
      }
      saveList[i].state = saveList[i].state.join("|");
      data = [
        saveList[i].date,
        bewCleanId(saveList[i].page),
        saveList[i].label,
        "",
        saveList[i].state
      ].join("|");
      data = bewUtilFixSave(data);
      saves.push(data);
    }
    $.set("$saves", JSON.stringify(saves));
  });
}

// Exports all set achievements.
function exportAchievements() {
  var i, name;
  var data = {};
  var achievements = getAchievementList();
  data.activities = [];
  data.achievements = [];
  for (i = 0; i < achievements.length; i += 1) {
    name = achievements[i];
    if ($.shared.isAchievementSet(name)) {
      if (isActivity(name)) {
        data.activities.push(name);
      } else {
        data.achievements.push(name);
      }
    }
  }
  saveTextAsFile(JSON.stringify(data), "bew-achievements.json");
}

// Imports and sets a list of achievements.
// NOTE: It merges the current set of achievements with the imported ones.
function importAchievements() {
  loadFileAsText(function (str) {
    var i;
    var difficulty = $.get("difficulty");
    var data = JSON.parse(str);
    // Sanity check on the input data.
    if (typeof data.activities === "undefined" ||
        typeof data.achievements === "undefined") {
      return;
    }
    // Temporarily set the difficulty to hard so all achievements can be set.
    $.set("difficulty", 2);
    achievementCanPopUp = false;
    // First set the activities.
    for (i = 0; i < data.activities.length; i += 1) {
      $.shared.setAchievement(data.activities[i]);
    }
    // Then set the achievements.
    for (i = 0; i < data.achievements.length; i += 1) {
      if (getAchievementDependecies(data.achievements[i]).length < 1) {
        $.shared.setAchievement(data.achievements[i]);
      }
    }
    achievementCanPopUp = true;
    // Restore the original difficulty setting.
    $.set("difficulty", difficulty);
  });
}

//#endregion

//#region ------- Test specific functions --------

var utMode = false;

/**
 * Return the number of possible choices for the current page.
 * @returns {number} The number of possible choices.
 */
function utGetChoiceCount() {
  return document.getElementsByClassName("hotspot").length +
    document.getElementsByClassName("response").length;
}

/**
 * Take the specified choice for the current page.
 * @param {number} choice The index of the choice to take.
 */
function utChoose(choice) {
  var el;
  var hotspots = document.getElementsByClassName("hotspot");
  var responses = document.getElementsByClassName("response");
  // Sanity check.
  if (choice > -1 && choice < utGetChoiceCount()) {
    if ($.getCurrentPage() === "d1recap01") {
      choice = 1;
    }
    if (choice >= hotspots.length) {
      // Response-case.
      choice -= hotspots.length;
      el = responses[choice];
    } else {
      // Hotspot-case.
      el = hotspots[choice].getElementsByTagName("i")[0];
    }
    el.click();
  }
}

//#endregion

//#endregion

//#region #               Gameplay Functions                #

// Starts a new game.
function startNewGame() {
  // NOTE: This will be $.startNewGame() or similar.
  initVars();
  setFlag("ingame", true);
  $.history.clear();
  $.shared.fadeToEpisode("Base", "intro1");
}

//#endregion

//#region #              Achievement Functions              #

var achievementDataCache = {};
var achievementCanPopUp = true;

/**
 * Returns the data-cache entry for the specified achievement.
 * @param {string} achievement The achievement to be checked.
 * @returns {object} The data-cache entry.
 */
function getAchievementDataCacheEntry(achievement) {
  var result = achievementDataCache[achievement] || null;
  if (!result) {
    achievementDataCache[achievement] = {};
    result = achievementDataCache[achievement];
  }
  return result;
}

/**
 * Checks if the specified achievement is an activity.
 * @param {string} achievement The identifier of the achievement to be checked.
 * @returns {boolean} true if the specified achievement is an activity.
 */
function isActivity(achievement) {
  var data = getAchievementDataCacheEntry(achievement);
  if (typeof data.isActivity === "undefined") {
    data.isActivity = $.shared.isActivity(achievement);
  }
  return data.isActivity;
}

/**
 * Checks if the specified achievement is unobtainable.
 * @param {string} achievement The identifier of the achievement to be checked.
 * @returns {boolean} true if the achievement is hidden.
 */
function isAchievementHidden(achievement) {
  var data = getAchievementDataCacheEntry(achievement);
  if (typeof data.isAchievementHidden === "undefined") {
    data.isAchievementHidden = $.shared.isAchievementHidden(achievement);
  }
  return data.isAchievementHidden;
}

/**
 * Returns the specified achievement's dependencies.
 * @param {string} achievement The identifier of the achievement to be checked.
 * @returns {Array} The list of dependencies.
 */
function getAchievementDependecies(achievement) {
  var data = getAchievementDataCacheEntry(achievement);
  if (typeof data.dependencies === "undefined") {
    data.dependencies = $.shared.getAchievementDependencies(achievement);
    data.dependencies =
      data.dependencies === "" ? [] : data.dependencies.split("|");
  }
  return data.dependencies;
}

/**
 * Returns the list of all available achievements.
 * @returns {Array} The list of achievements.
 */
function getAchievementList() {
  return bewGetVariable("$achievements").elements;
}

/** Hides the achievement notification. */
function achievementHidePopup() {
  var container = document.getElementById("AchievementPopupContainer");
  if (container !== null) {
    container.className = "popup-hidden";
  }
}

/* Reveals the achievement notification. */
function achievementShowPopup() {
  var container = document.getElementById("AchievementPopupContainer");
  if (container !== null) {
    container.className = "popup-visible";
  }
}

/*
 * Adds the achievement unlocked notification.
 * @param {string} name - The name of the achievement that was unlocked.
 */
function achievementAddPopup(name) {
  // TODO: Implement better layout for the achievement pop-up.
  var container, imageDiv, icon, text, textTemplate;
  // Do not add pop-ups for hidden achievements.
  if (!achievementCanPopUp || isAchievementHidden(name)) {
    return;
  }
  // Create or get the container div for the achievement notification.
  container = document.getElementById("AchievementPopupContainer");
  if (!container) {
    container = document.createElement("div");
    container.id = "AchievementPopupContainer";
    container.addEventListener("click", achievementHidePopup);
    document.body.insertBefore(container, document.body.firstChild);
  } else {
    // Clear the container div.
    container.innerHTML = "";
  }
  // Build the achievement notification.
  container.className = "popup-hidden";
  imageDiv = document.createElement("div");
  imageDiv.className = "image";
  icon = document.createElement("div");
  icon.className = "icon";
  icon.style.backgroundImage = "url('" +
    assetManager.requestImage("icons." + name.toLowerCase()) + "')";
  imageDiv.appendChild(icon);
  textTemplate = "<@sAchievementUnlocked>\n<@achievements." + name + ">";
  text = document.createElement("div");
  text.className = "text";
  $.bindText(text, textTemplate);
  container.appendChild(imageDiv);
  container.appendChild(text);
  setTimeout(achievementShowPopup, 200);
  setTimeout(achievementHidePopup, 4200);
}

//#endregion

//#region #                 Scoring Section                 #

/** Displays the game score (in debug mode). */
function showScore() {
  var exportPath, scoreDiv, pageContent, p;
  var loggingState = bewDebugLoggingEnabled;
  var pointsArr = [];
  var str = "";
  bewDebugLoggingEnabled = false;
  exportPath = "<a onclick=\"replaySavePathData(); return false;\" href='#'>" +
    "Generate Debug File</a>";
  // In the output text, all instances of "{id}" will be
  // replaced with the value of "points".
  pointsArr.push({
    id: "score",
    points: $.shared.calcScore()
  });
  "efnslajm".split("").forEach(function (x) {
    pointsArr.push({
      id: x,
      points: $.shared.calcPoints(x)
    });
  });
  // Keep the score-line in a div for easier testing.
  str = "<div>Score: {score}, E: {e}, F: {f}, N: {n}, S: {s}, " +
    "L: {l}, A: {a}, J: {j}, M: {m}.</div>" +
    bewStoryPaths[$.get("path")] + " || ";
  // Perform formatting on str.
  pointsArr.forEach(function (x) {
    str = str.replace(new RegExp("\\{" + x.id + "\\}", "g"), x.points);
  });
  // Add the "Generate Debug File" option to the score-line.
  str += exportPath;
  // Display the score.
  scoreDiv = document.getElementById("Score");
  if (!scoreDiv) {
    scoreDiv = document.createElement("div");
    scoreDiv.id = "Score";
    scoreDiv.className = "debug-score";
    pageContent = document.getElementById("PageContent");
    if (pageContent) {
      pageContent.appendChild(scoreDiv);
    }
  }
  scoreDiv.innerHTML = "";
  p = document.createElement("p");
  p.innerHTML = str;
  scoreDiv.appendChild(p);
  bewDebugLoggingEnabled = loggingState;
}

//#endregion

//#region #                 Debug Functions                 #

var debugRecordTouches = false;
var debugRelevantVariables = [];
var debugGenerateRelevantList = null;
// TODO: Make custom bunch-modifier callbacks registerable.
var debugCustomBunchModifierCallbacks = [
  {
    name: "$achievements",
    increase: debugAchievementPlus,
    decrease: debugAchievementMinus
  }
];

var debugMode = (function () {
  var enabled = false;
  function updateTextFields(value) {
    var i;
    var list = document.querySelectorAll("p[spellcheck]");
    for (i = 0; i < list.length; i += 1) {
      list[i].spellcheck = value;
      list[i].contentEditable = value;
    }
  }
  function enable() {
    var mainText = template.getTextContainer();
    enabled = true;
    setFlag("debugMode", true);
    updateTextFields(checkFlag("spellChecking"));
    document.body.classList.add("debug");
    mainText.focus();
    mainText.blur();
  }
  function disable() {
    enabled = false;
    setFlag("debugMode", false);
    updateTextFields(false);
    document.body.classList.remove("debug");
  }
  function toggle() {
    if (enabled) {
      disable();
    } else {
      enable();
    }
  }
  function isEnabled() {
    return enabled;
  }
  return {
    isEnabled: isEnabled,
    enable: enable,
    disable: disable,
    toggle: toggle
  };
}());

// Adds a variable to the list of relevant variables.
function debugTouchVariable(name) {
  var variable = bewGetVariable(name);
  if (debugRecordTouches &&
      variable.group !== "$" &&
      variable.type !== "bunch" &&
      debugRelevantVariables.indexOf(name) < 0) {
    debugRelevantVariables.push(name);
  }
}

function debugAchievementPlus(e) {
  var element = e.currentTarget.parentNode.getAttribute("data-element") || "";
  if (element) {
    $.shared.setAchievement(element, true);
  }
}

function debugAchievementMinus(e) {
  var element = e.currentTarget.parentNode.getAttribute("data-element") || "";
  if (element) {
    $.shared.setAchievement(element, false);
  }
}

/** Adds human-friendly descriptions for all achievements on the inspector. */
function bewDebugInspectorAnnotateAchievements() {
  var i, el, name, textTemplate;
  var list = document.querySelectorAll("#DebugAchievements .value");
  // Returns the title of the specified achievement.
  function getAchievementTitleOrDescription(name) {
    if ($.requestString("achievements." + name, "")) {
      return "<@achievements." + name + ">";
    }
    if ($.requestString("achievements." + name + ".info", "")) {
      return "<@achievements." + name + ".info>";
    }
    return name;
  }
  for (i = 0; i < list.length; i += 1) {
    el = list[i];
    name = el.parentNode.getAttribute("data-element") || "";
    if (name) {
      textTemplate = getAchievementTitleOrDescription(name);
      $.bindHint(el, textTemplate);
    }
  }
}

/** Generates the debug navigation bar. */
function debugGenerateNavigationBar() {
  function addNavigationButton(caption, id, func) {
    var el = document.createElement("div");
    el.className = "debug-navigation-button";
    el.id = id;
    el.addEventListener("click", func);
    el.textContent = caption;
    wrapper.appendChild(el);
  }
  var wrapper = document.getElementById("DebugNavigationBar");
  if (!wrapper) {
    wrapper = document.createElement("div");
    wrapper.className = "debug-navigation-bar";
    wrapper.id = "DebugNavigationBar";
    document.body.appendChild(wrapper);
    // Create the sidebar toggle button.
    addNavigationButton("☰\uFE0E", "DebugToggleSidebar", function () {
      var filterInput = document.body.querySelector(".debug-page-list-filter");
      document.body.classList.toggle("show-sidebar");
      if (filterInput && document.body.classList.contains("show-sidebar")) {
        filterInput.focus();
      }
    });
    // Create the history button.
    //addNavigationButton("🕒\uFE0E", "DebugHistoryButton", function () {});
    // Create the back button.
    addNavigationButton("◀\uFE0E", "DebugBackButton", $.history.back);
    // Create the forward button.
    addNavigationButton("▶\uFE0E", "DebugForwardButton", $.history.forward);
  }
}

/** Generates the debug inspector. */
function debugGenerateInspector() {
  var i, name, variable;
  var varList = [];
  var bunchList = [];
  var sysVarList = [];
  var varNames = Object.keys(game.variables);
  var pageContent = document.getElementById("PageContent");
  function getInspectorContainer() {
    var el = document.getElementById("DebugInspector");
    if (!el) {
      el = document.createElement("div");
      el.id = "DebugInspector";
      el.className = "debug-inspector";
      pageContent.appendChild(el);
    }
    return el;
  }
  /** @param {Event} e - The event. */
  function toggleGroupVisibility(e) {
    e.currentTarget.parentNode.classList.toggle("minimized");
  }
  /** @param {Event} e - The event. */
  function pinVariableGroup(e) {
    var i;
    var el = e.currentTarget.parentNode;
    var list = document.querySelectorAll("#DebugVariables .group");
    for (i = 0; i < list.length; i += 1) {
      if (el !== list[i]) {
        list[i].classList.remove("pinned");
      }
    }
    el.classList.toggle("pinned");
  }
  /** @param {Array} list - The list of system variables. */
  function outputSystemVariables(list) {
    // TODO: Implement the inspector for system variables.
    void list;
  }
  /** @param {Array} list - The list variables. */
  function outputVariables(list) {
    var i, j, variable, container, header;
    var maxPerGroup = 0;
    var groupNames = [];
    var groupList = [];
    var group = [];
    /** @param {Event} e - The event. */
    function increaseOnClick(e) {
      var variable = e.currentTarget.parentNode.getAttribute("data-name") || "";
      if (variable) {
        $.add(variable, 1);
      }
    }
    /** @param {Event} e - The event. */
    function decreaseOnClick(e) {
      var variable = e.currentTarget.parentNode.getAttribute("data-name") || "";
      if (variable) {
        $.sub(variable, 1);
      }
    }
    function outputSingleLine(parent, name) {
      var variable = bewGetVariable(name);
      var container = document.createElement("div");
      var varDiv = document.createElement("div");
      var plusDiv = document.createElement("div");
      var minusDiv = document.createElement("div");
      container.className = "container";
      container.setAttribute("data-name", name);
      // Create the actual div for the variable.
      varDiv.id = "DebugVar-" + name;
      varDiv.className = "value" + (variable.value === 0 ? "" : " set");
      varDiv.innerHTML = name + " = " + variable.value;
      // Create the plus-div.
      plusDiv.className = "modifier plus";
      plusDiv.addEventListener("click", increaseOnClick);
      plusDiv.innerHTML = "+";
      // Create the minus-div.
      minusDiv.className = "modifier minus";
      minusDiv.addEventListener("click", decreaseOnClick);
      minusDiv.innerHTML = "-";
      // Add all div tags to the tree.
      container.appendChild(varDiv);
      container.appendChild(minusDiv);
      container.appendChild(plusDiv);
      parent.appendChild(container);
    }
    function generateRelevantGroup() {
      var i;
      var group = document.getElementById("DebugRelevantGroup");
      var header = document.createElement("div");
      header.className = "header";
      header.addEventListener("click", pinVariableGroup);
      header.innerHTML = "Relevant Variables";
      group.innerHTML = "";
      group.appendChild(header);
      group.style.display = debugRelevantVariables.length > 0 ? "" : "none";
      for (i = 0; i < debugRelevantVariables.length; i += 1) {
        outputSingleLine(group, debugRelevantVariables[i]);
      }
    }
    // Gather the list of named variable groups.
    for (i = 0; i < list.length; i += 1) {
      variable = bewGetVariable(list[i]);
      if (typeof variable.group !== "undefined" &&
          groupNames.indexOf(variable.group) < 0) {
        groupNames.push(variable.group);
      }
    }
    // Gather all named groups.
    for (j = 0; j < groupNames.length; j += 1) {
      for (i = 0; i < list.length; i += 1) {
        variable = bewGetVariable(list[i]);
        if (typeof variable.group !== "undefined" &&
            variable.group === groupNames[j]) {
          group.push(list[i]);
        }
      }
      maxPerGroup = Math.max(group.length, maxPerGroup);
      groupList.push({
        name: groupNames[j],
        list: group
      });
      group = [];
    }
    // Gather all unnamed groups.
    for (i = 0; i < list.length; i += 1) {
      variable = bewGetVariable(list[i]);
      if (typeof variable.group === "undefined") {
        group.push(list[i]);
      }
      if (i === list.length - 1 || group.length === maxPerGroup) {
        groupList.push({
          name: "",
          list: group
        });
        group = [];
      }
    }
    // Add all remaining variables as one final group.
    if (group.length > 0) {
      groupList.push({
        name: "",
        list: group
      });
    }
    // Output all groups.
    if (groupList.length > 0) {
      // Add the recently used group.
      groupList.push({
        name: "RelevantGroup",
        list: []
      });
      // Create the wrapper div.
      container = document.createElement("div");
      container.id = "DebugVariables";
      container.className = "group-container";
      header = document.createElement("div");
      header.className = "header";
      header.innerHTML = "Variables";
      header.addEventListener("click", toggleGroupVisibility);
      container.appendChild(header);
      for (j = 0; j < groupList.length; j += 1) {
        // Create a new group.
        group = document.createElement("div");
        group.className = "group";
        // Create the header div.
        header = document.createElement("div");
        header.className = "header";
        header.addEventListener("click", pinVariableGroup);
        if (groupList[j].name !== "") {
          header.innerHTML = groupList[j].name;
          group.id = "Debug" + groupList[j].name;
        } else {
          header.innerHTML = "Variables";
        }
        group.appendChild(header);
        for (i = 0; i < groupList[j].list.length; i += 1) {
          outputSingleLine(group, groupList[j].list[i]);
        }
        container.appendChild(group);
      }
      // Add the generated output to the page.
      getInspectorContainer().appendChild(container);
      // Generate the recently used list.
      debugGenerateRelevantList = generateRelevantGroup;
      generateRelevantGroup();
      document.getElementById("DebugRelevantGroup").classList.add("pinned");
    }
  }
  /** @param {Array} list - The list of bunches. */
  function outputBunches(list) {
    var MINIMUM_PER_GROUP = 10;
    var i, j, k, name, variable, elementList, container, group, header,
      maxPerGroup, groupCount, increaseFunc, decreaseFunc, modifier;
    /**
     * Cleans the name of the specified bunch.
     * @param {string} name The bunch-name to be cleaned up.
     * @returns {string} The sanitized name.
     */
    function cleanBunchName(name) {
      name = name.replace(/^[-_$]+/g, "");
      name = name.charAt(0).toUpperCase() + name.slice(1);
      name = name.replace(/[.\s]+/g, "-");
      return name;
    }
    /** @param {Event} e - The event. */
    function increaseOnClick(e) {
      var bunch = e.currentTarget.parentNode.getAttribute("data-bunch") || "";
      var elem = e.currentTarget.parentNode.getAttribute("data-element") || "";
      if (bunch && elem) {
        $.add(bunch, elem);
      }
    }
    /** @param {Event} e - The event. */
    function decreaseOnClick(e) {
      var bunch = e.currentTarget.parentNode.getAttribute("data-bunch") || "";
      var elem = e.currentTarget.parentNode.getAttribute("data-element") || "";
      if (bunch && elem) {
        $.sub(bunch, elem);
      }
    }
    function outputSingleLine(parent, bunch, element, increase, decrease) {
      var container = document.createElement("div");
      var varDiv = document.createElement("div");
      var plusDiv = document.createElement("div");
      var minusDiv = document.createElement("div");
      var value = $.check(bunch, element);
      container.className = "container";
      container.setAttribute("data-bunch", bunch);
      container.setAttribute("data-element", element);
      // Create the actual div for the variable.
      varDiv.id = "Debug" + cleanBunchName(bunch + "-" + element);
      varDiv.className = "value" + (value ? " set" : "");
      varDiv.innerHTML = element + " = " + (value ? 1 : 0);
      // Create the plus-div.
      plusDiv.className = "modifier plus";
      plusDiv.addEventListener("click", increase);
      plusDiv.innerHTML = "+";
      // Create the minus-div.
      minusDiv.className = "modifier minus";
      minusDiv.addEventListener("click", decrease);
      minusDiv.innerHTML = "-";
      // Add all div tags to the tree.
      container.appendChild(varDiv);
      container.appendChild(minusDiv);
      container.appendChild(plusDiv);
      parent.appendChild(container);
    }
    // Add a separate inspector group for all bunches.
    for (j = 0; j < list.length; j += 1) {
      name = list[j];
      variable = bewGetVariable(name);
      // Get a sorted list of the bunch's elements.
      elementList = variable.elements.slice().sort();
      if (elementList.length > 0) {
        // TODO: Determine a good value for groupCount.
        groupCount = 11;
        // Determine the increase/decrease functions for the modifiers.
        increaseFunc = increaseOnClick;
        decreaseFunc = decreaseOnClick;
        for (i = 0; i < debugCustomBunchModifierCallbacks.length; i += 1) {
          modifier = debugCustomBunchModifierCallbacks[i];
          if (modifier.name === name) {
            increaseFunc = modifier.increase;
            decreaseFunc = modifier.decrease;
            break;
          }
        }
        container = document.createElement("div");
        container.id = "Debug" + cleanBunchName(name);
        container.className = "group-container";
        // Create the header div.
        header = document.createElement("div");
        header.className = "header";
        header.innerHTML = cleanBunchName(name);
        header.addEventListener("click", toggleGroupVisibility);
        container.appendChild(header);
        // Determine the number of elements to group together but make sure
        // that at least N (or the number of elements) get grouped together.
        maxPerGroup = Math.ceil(elementList.length / groupCount);
        maxPerGroup = Math.max(maxPerGroup, MINIMUM_PER_GROUP);
        maxPerGroup = Math.min(maxPerGroup, elementList.length);
        for (i = 0; i < elementList.length / maxPerGroup; i += 1) {
          // Create a new group.
          group = document.createElement("div");
          group.className = "group";
          // Generate the divs for each element.
          for (
            k = i * maxPerGroup;
            k < elementList.length && k < (i + 1) * maxPerGroup;
            k += 1
          ) {
            outputSingleLine(group, name,
              elementList[k], increaseFunc, decreaseFunc);
          }
          container.appendChild(group);
        }
        // Add the generated output to the page.
        getInspectorContainer().appendChild(container);
      }
    }
  }
  // Only generate the debug inspector if its container element exists.
  if (pageContent) {
    // Gather separate lists for pure variables, system variables, and bunches.
    for (i = 0; i < varNames.length; i += 1) {
      name = varNames[i];
      variable = bewGetVariable(name);
      if (variable.group === "$") {
        sysVarList.push(name);
      } else if (variable.type === "bunch") {
        bunchList.push(name);
      } else {
        varList.push(name);
      }
    }
    outputSystemVariables(sysVarList);
    outputVariables(varList);
    outputBunches(bunchList);
  }
}

/** Generate the debug sidebar. */
function debugGenerateSidebar() {
  function onPageClick(e) {
    var pageId = e.currentTarget.getAttribute("data-page") || "";
    if (pageId) {
      $.jmp(pageId);
    }
  }
  function addDebugPage(parent, pageId) {
    function getAllMatches(regex, str) {
      var result = [];
      var match = regex.exec(str);
      while (match !== null) {
        result.push(match);
        match = regex.exec(str);
      }
      return result;
    }
    function resolveText(str) {
      var result = str.replace(/\\"/g, "\"");
      result = $.evaluateText(result, pageId, true);
      // Step 1, annotate condition- and text-portions.
      result = result.replace(/\{(?:([^{]*?)\|)?([^{]*?)\}/g,
        "<div><span class=\"condition\">$1</span><span>$2</span></div>");
      // Step 2, wrap whole inline conditional blocks in divs.
      result = result.replace(/\{([^{]*?)\}/g,
        "<div class=\"hl-inline-block\">$1</div>");
      return result;
    }
    function generateDetailsFromSource(source) {
      var rBlitText = /\$\.blit\("(.*)"\);/g;
      var rPrintText = /\$\.print\("(.*)"\);/g;
      var rHotspotText = /\$\.addHotspot\(".*",\s*"(.*)",\s*".*"\);/g;
      var rResponseText = /\$\.addResponse\("(.*)",\s*".*"\);/g;
      var blitList = getAllMatches(rBlitText, source) || [];
      var printList = getAllMatches(rPrintText, source) || [];
      var hotspotList = getAllMatches(rHotspotText, source) || [];
      var responseList = getAllMatches(rResponseText, source) || [];
      var el = document.createElement("div");
      var options = [];
      var result = "";
      var i, imageName;
      for (i = 0; i < blitList.length; i += 1) {
        imageName = resolveText(blitList[i][1]);
        imageName = imageName.split("|")[0].split(":")[0];
        result += "<div class=\"image-name\">" + imageName + "</div>";
      }
      for (i = 0; i < printList.length; i += 1) {
        result += "<div>" + resolveText(printList[i][1]) + "</div>";
      }
      for (i = 0; i < hotspotList.length; i += 1) {
        options.push("<li>" + resolveText(hotspotList[i][1]) + "</li>");
      }
      for (i = 0; i < responseList.length; i += 1) {
        options.push("<li>" + resolveText(responseList[i][1]) + "</li>");
      }
      if (options.length > 0) {
        result += "<ul>" + options.join("") + "</ul>";
      }
      el.className = "details";
      el.innerHTML = result;
      return el;
    }
    var wrapper = document.createElement("div");
    var titleDiv = document.createElement("div");
    var detailsDiv = document.createElement("div");
    var page = game.pages[pageId] || null;
    var rAnimated = /\.blit\(.+?\|.+?\)/;
    var source, title;
    if (page !== null) {
      source = page.toString();
      title = pageId;
      if (rAnimated.test(source)) {
        title += " 📹";
      }
      wrapper.className = "debug-page";
      wrapper.setAttribute("data-page", pageId);
      wrapper.addEventListener("click", onPageClick);
      titleDiv.className = "name";
      titleDiv.textContent = title;
      detailsDiv = generateDetailsFromSource(source);
      wrapper.appendChild(titleDiv);
      wrapper.appendChild(detailsDiv);
      parent.appendChild(wrapper);
      // Add the textContent to the cache for filtering.
      textCache.push(wrapper.textContent);
    }
  }
  function addHeader(parent) {
    function filterPages(e) {
      var searchText = e.currentTarget.value.trim();
      var list = document.querySelectorAll(".debug-page-list .debug-page");
      var r = new RegExp(searchText);
      var j, text;
      for (j = 0; j < list.length; j += 1) {
        text = textCache[j];
        if (text.match(r) !== null) {
          list[j].classList.remove("hidden");
        } else {
          list[j].classList.add("hidden");
        }
      }
      listWrapper.scrollTop = 0;
    }
    var el = document.createElement("div");
    var input = document.createElement("input");
    el.className = "debug-page-list-header";
    input.className = "debug-page-list-filter";
    input.addEventListener("input", filterPages);
    el.appendChild(input);
    parent.appendChild(el);
  }
  var i;
  var container = document.getElementById("DebugSideBar");
  var listWrapper = document.createElement("div");
  var padding = document.createElement("div");
  var pageList = Object.keys(game.pages);
  var textCache = [];
  listWrapper.className = "debug-page-list";
  for (i = 0; i < pageList.length; i += 1) {
    addDebugPage(listWrapper, pageList[i]);
  }
  padding.className = "padding";
  listWrapper.appendChild(padding);
  addHeader(container);
  container.appendChild(listWrapper);
  // Register the active page updater.
  $.onPageLoad("*", function () {
    var currentPage = $.getCurrentPage();
    if (currentPage) {
      // Clear the previously set "current" page.
      [].slice.call(document.querySelectorAll(".debug-page.current"))
        .forEach(function (el) {
          el.classList.remove("current");
        });
      // Set the new "current" page.
      [].slice.call(
        document.querySelectorAll(".debug-page[data-page=" + currentPage + "]")
      ).forEach(function (el) {
        el.classList.add("current");
        el.scrollIntoViewIfNeeded(true);
      });
    }
  });
}

/** Outputs various debug specific information. */
function debug() {
  // Add the custom debug navigation bar.
  debugGenerateNavigationBar();
  // Add the debug inspector.
  debugGenerateInspector();
  // Add the debug sidebar.
  debugGenerateSidebar();
}

//#endregion

//#region #          Menu & Save/Restore Functions          #

//#region -------- User object ---------

var user = (function () {
  var name = "";
  var avatar = "";
  var frame = null;
  var userCard = null;
  function updateUserInfo() {
    if (checkFlag("onlineEnabled") || bewOnlineVersion || bewIsInternal) {
      contactGatekeeper();
    }
  }
  function setCurrentUser() {
    var image, div;
    if (!!name && !userCard) {
      userCard = document.createElement("div");
      image = document.createElement("div");
      div = document.createElement("div");
      div.className = "user-name";
      div.innerHTML = name;
      image.className = "user-avatar";
      if (avatar) {
        image.style.backgroundImage = "url('" + avatar + "')";
      }
      userCard.appendChild(image);
      userCard.appendChild(div);
      userCard.id = "UserCard";
      userCard.className = "user-card";
      userCard.addEventListener("click", updateUserInfo);
      document.body.appendChild(userCard);
    }
  }
  function sendInfoToGatekeeper() {
    if (frame) {
      frame.contentWindow.postMessage({
        command: "info",
        saves: bewGetSaves(),
        achievements: $.get("$achievements")
      }, "*");
    }
  }
  function removeSave(id) {
    if (frame) {
      frame.contentWindow.postMessage({
        command: "remove",
        saves: id
      }, "*");
    }
  }
  function forceAchievements(value) {
    value = value || "";
    $.set("$achievements", value);
    if (frame) {
      frame.contentWindow.postMessage({
        command: "force-achievements",
        achievements: value
      }, "*");
    }
  }
  function contactGatekeeper() {
    frame = document.getElementById("BrokerFrame");
    if (frame) {
      sendInfoToGatekeeper();
    } else {
      frame = document.createElement("iframe");
      frame.id = "BrokerFrame";
      if (checkFlag("betaFeatures")) {
        frame.src = atob(
          "aHR0cHM6Ly9icmFkc2Vyb3RpY3dlZWsucm9ja3Mvc3luYy9icm9rZXIuaHRtbA==") +
          "?r=" + Math.floor(Math.random() * 500).toString();
      } else {
        frame.src =
          atob("aHR0cHM6Ly9icmFkc2Vyb3RpY3dlZWsucm9ja3Mvc3luYy9icm9rZXIucGhw");
      }
      frame.style.display = "none";
      document.body.appendChild(frame);
      window.addEventListener("message", function (e) {
        var i, saves;
        if (e.origin === atob("aHR0cHM6Ly9icmFkc2Vyb3RpY3dlZWsucm9ja3M=")) {
          if (typeof e.data.command !== "undefined") {
            if (e.data.command === "hello") {
              sendInfoToGatekeeper();
            } else if (e.data.command === "user") {
              name = e.data.user;
              avatar = e.data.avatar;
              setCurrentUser();
            } else if (e.data.command === "update-achievements") {
              achievementCanPopUp = false;
              $.set("$achievements", e.data.achievements);
              $.shared.updateAchievements();
              achievementCanPopUp = true;
            } else if (e.data.command === "start-sync") {
              // Syncing started.
              if (userCard) {
                document.body.classList.add("syncing");
              }
            } else if (e.data.command === "end-sync") {
              // Syncing completed.
              if (userCard) {
                document.body.classList.remove("syncing");
              }
            } else if (e.data.command === "saves") {
              saves = e.data.saves;
              for (i = 0; i < saves.length; i += 1) {
                saves[i] = bewUtilFixSave(saves[i]);
              }
              $.set("$saves", JSON.stringify(saves));
            }
          }
        }
      }, false);
    }
  }
  return {
    forceAchievements: forceAchievements,
    updateUserInfo: updateUserInfo,
    removeSave: removeSave
  };
}());

//#endregion

//#region ------ Save/Load menus -------

// TODO: Merge generateRestorePage() and generateSavePage() into one function.

// Returns the label for a saved game from the specified save-data.
function getSaveName(data) {
  var label = data[2];
  if (label === "") {
    label = data[1];
    label = bewCleanId(label);
  }
  return label;
}

function uiSetSavePreview(id, text) {
  var previewImage = document.getElementById("SavePreviewImage");
  var previewText = document.getElementById("SavePreviewText");
  var src = assetManager.requestImage(id);
  if (previewImage) {
    previewImage.style.backgroundImage = "url('" + src + "')";
  }
  if (previewText) {
    previewText.innerHTML = "<p>" + (text || "") + "</p>";
  }
}

/** @param {Event} e - The event. */
function uiSaveSlotOnMouseOver(e) {
  var text, prediction;
  var page = e.currentTarget.getAttribute("data-page") || "";
  var date = e.currentTarget.getAttribute("data-date") || "";
  var state = e.currentTarget.getAttribute("data-state") || "";
  function predictSave() {
    $.jmp(page);
  }
  if (page) {
    prediction = $.predict(predictSave, state);
    date = date ? new Date(parseInt(date, 16)).toLocaleString() : "";
    text = (date ? "<span class=\"date\">" + date + "</span>" : "");
    text += prediction.text;
    uiSetSavePreview(prediction.images[0] || "", text);
  }
}

// Generates the dynamic list of saved games for the restore page.
function generateRestorePage() {
  var i, el, page, data, wrapperDiv, previewImage, previewText,
    listDiv, itemDiv, deleteDiv, containerDiv;
  var saves = bewGetSaves();
  function setNothingtAsPreview() {
    uiSetSavePreview("");
  }
  /** @param {Event} e - The event. */
  function doDeleteSave(e) {
    var id = e.currentTarget.getAttribute("data-id") || "";
    if (id) {
      deleteSave(id);
      generateRestorePage();
    }
  }
  /** @param {Event} e - The event. */
  function doRestoreGameState(e) {
    var id = e.currentTarget.getAttribute("data-id") || "";
    if (id) {
      // Reset the replay system and restore the saved state.
      $.history.clear();
      restoreState(id);
      if (!checkFlag("noEffects")) {
        // TODO: Handle toggling the menu more elegantly while loading a save.
        $.fx.enableDiv("#000");
        uiToggleMenu();
      }
    }
  }
  if (document.body.classList.contains("syncing")) {
    // Do not let the player enter the load menu while syncing is in progress.
    return;
  }
  uiClearMenu("menu-wrapper saveload");
  wrapperDiv = document.getElementById("Menu");
  previewImage = document.createElement("div");
  previewImage.id = "SavePreviewImage";
  previewImage.className = "save-preview-image";
  previewText = document.createElement("div");
  previewText.id = "SavePreviewText";
  previewText.className = "save-preview-text";
  wrapperDiv.appendChild(previewImage);
  wrapperDiv.appendChild(previewText);
  setNothingtAsPreview();
  wrapperDiv.addEventListener("mouseover", setNothingtAsPreview, true);
  // Create the list div.
  listDiv = document.createElement("div");
  listDiv.id = "SaveLoadList";
  listDiv.className = "save-load-list";
  // Populate the list.
  // First add the quick save if present.
  data = $.get("$qsave").split("|");
  if (data.length > 3) {
    page = data[1];
    // Create the container div.
    containerDiv = document.createElement("div");
    containerDiv.className = "container";
    containerDiv.setAttribute("data-page", page);
    containerDiv.setAttribute("data-date", data[0] || "");
    containerDiv.setAttribute("data-state", data.slice(4).join("|") || "");
    containerDiv.addEventListener("mouseover", uiSaveSlotOnMouseOver);
    // Add the item.
    itemDiv = document.createElement("div");
    itemDiv.className = "item new";
    itemDiv.setAttribute("data-id", "q");
    itemDiv.addEventListener("click", doRestoreGameState);
    $.bindText(itemDiv, "<@sQuickSavedGame>");
    containerDiv.appendChild(itemDiv);
    listDiv.appendChild(containerDiv);
  }
  // Then add all the other saves.
  for (i = saves.length - 1; i >= 0; i -= 1) {
    data = saves[i].split("|");
    // Date | page | label | <empty> | vars ...
    page = data[1];
    // Create the container div.
    containerDiv = document.createElement("div");
    containerDiv.className = "container";
    containerDiv.setAttribute("data-page", page);
    containerDiv.setAttribute("data-date", data[0] || "");
    containerDiv.setAttribute("data-state", data.slice(4).join("|") || "");
    containerDiv.addEventListener("mouseover", uiSaveSlotOnMouseOver);
    // Add the delete button.
    deleteDiv = document.createElement("div");
    deleteDiv.className = "delete";
    deleteDiv.setAttribute("data-id", i);
    deleteDiv.addEventListener("click", doDeleteSave);
    deleteDiv.innerHTML = "×";
    containerDiv.appendChild(deleteDiv);
    // Add the item.
    itemDiv = document.createElement("div");
    itemDiv.className = "item";
    itemDiv.setAttribute("data-id", i);
    itemDiv.addEventListener("click", doRestoreGameState);
    itemDiv.innerHTML = "Save " + (i + 1) + " - " + getSaveName(data);
    containerDiv.appendChild(itemDiv);
    listDiv.appendChild(containerDiv);
  }
  // Create the back-button.
  el = document.createElement("div");
  el.className = "saveload-back-button";
  $.bindText(el, "<@sBack>");
  el.tabIndex = 0;
  el.role = "button";
  el.addEventListener("click", uiGenerateMainMenu);
  el.addEventListener("keydown", bewKeyDownOnClick);
  wrapperDiv.appendChild(el);
  // Build and append our tree to the document.
  wrapperDiv.appendChild(listDiv);
}

// This one generates the dynamic list of saved games for the save page.
function generateSavePage() {
  var i, el, page, data, wrapperDiv, previewImage, previewText,
    listDiv, itemDiv, deleteDiv, containerDiv;
  var saves = bewGetSaves();
  function setCurrentAsPreview() {
    uiSetSavePreview($.getCurrentImage());
  }
  /**
   * Performs the actual delete-save operation.
   * @param {Event} e The initiator event.
   */
  function doDeleteSave(e) {
    var id = e.currentTarget.getAttribute("data-id") || "";
    if (id) {
      deleteSave(id);
      generateSavePage();
    }
  }
  /**
   * Performs the actual save operation.
   * @param {Event} e The initiator event.
   */
  function doSaveGameState(e) {
    var id = e.currentTarget.getAttribute("data-id") || "";
    var saves = bewGetSaves();
    var prevLabel = "";
    var newLabel = "";
    var page = $.getCurrentPage();
    if (id) {
      id = parseInt(id, 10);
      // Get the previous label of the save-slot.
      if (id < saves.length) {
        prevLabel = saves[id].split("|")[2];
      }
      prevLabel = prevLabel === "" ? page : prevLabel;
      // Prompt the player for a name for the save.
      newLabel = window.prompt("Please enter a name for your save", prevLabel);
      if (newLabel === null) {
        return;
      }
      if (page === newLabel || newLabel === "") {
        page += "|";
      } else {
        page += "|" + newLabel;
      }
      if (id < saves.length) {
        deleteSave(id);
      }
      saveState(saves.length, page);
      uiGenerateMainMenu();
    }
  }
  if (document.body.classList.contains("syncing")) {
    // Do not let the player enter the save menu while syncing is in progress.
    return;
  }
  uiClearMenu("menu-wrapper saveload");
  wrapperDiv = document.getElementById("Menu");
  previewImage = document.createElement("div");
  previewImage.id = "SavePreviewImage";
  previewImage.className = "save-preview-image";
  previewText = document.createElement("div");
  previewText.id = "SavePreviewText";
  previewText.className = "save-preview-text";
  wrapperDiv.appendChild(previewImage);
  wrapperDiv.appendChild(previewText);
  setCurrentAsPreview();
  wrapperDiv.addEventListener("mouseover", setCurrentAsPreview, true);
  // Create the list div.
  listDiv = document.createElement("div");
  listDiv.id = "SaveLoadList";
  listDiv.className = "save-load-list";
  // Add the "create a new save" option to the top.
  containerDiv = document.createElement("div");
  containerDiv.className = "container";
  containerDiv.addEventListener("mouseover", setCurrentAsPreview, true);
  itemDiv = document.createElement("div");
  itemDiv.className = "item new";
  itemDiv.setAttribute("data-id", saves.length);
  itemDiv.addEventListener("click", doSaveGameState);
  $.bindText(itemDiv, "<@sCreateNewSave>");
  containerDiv.appendChild(itemDiv);
  listDiv.appendChild(containerDiv);
  // Populate the list.
  for (i = saves.length - 1; i >= 0; i -= 1) {
    data = saves[i].split("|");
    // Date | page | label | <empty> | vars ...
    page = data[1];
    // Create the container div.
    containerDiv = document.createElement("div");
    containerDiv.className = "container";
    containerDiv.setAttribute("data-page", page);
    containerDiv.setAttribute("data-date", data[0] || "");
    containerDiv.setAttribute("data-state", data.slice(4).join("|") || "");
    containerDiv.addEventListener("mouseover", uiSaveSlotOnMouseOver);
    // Add the delete button.
    deleteDiv = document.createElement("div");
    deleteDiv.className = "delete";
    deleteDiv.setAttribute("data-id", i);
    deleteDiv.addEventListener("click", doDeleteSave);
    deleteDiv.innerHTML = "×";
    containerDiv.appendChild(deleteDiv);
    // Add the item.
    itemDiv = document.createElement("div");
    itemDiv.className = "item";
    itemDiv.setAttribute("data-id", i);
    itemDiv.addEventListener("click", doSaveGameState);
    itemDiv.innerHTML = "Save " + (i + 1) + " - " + getSaveName(data);
    containerDiv.appendChild(itemDiv);
    listDiv.appendChild(containerDiv);
  }
  // Create the back-button.
  el = document.createElement("div");
  el.className = "saveload-back-button";
  $.bindText(el, "<@sBack>");
  el.tabIndex = 0;
  el.role = "button";
  el.addEventListener("click", uiGenerateMainMenu);
  el.addEventListener("keydown", bewKeyDownOnClick);
  wrapperDiv.appendChild(el);
  // Build and append our tree to the document.
  wrapperDiv.appendChild(listDiv);
  // Scroll the list to the top, making the "create new save" option visible.
  listDiv.scrollTop = 0;
}

//#endregion

//#region ------ Achievement page ------

// Generates the achievement-page.
function generateAchievementPage() {
  function createList(listId) {
    var clearDiv = document.createElement("div");
    clearDiv.style.clear = "both";
    // Create the list div.
    listDiv = document.createElement("div");
    listDiv.className = "achievement-list";
    listDiv.id = listId;
    listDiv.style.display = "none";
    // Create the groups for locked/unlocked achievements.
    unlockedGroup = document.createElement("div");
    unlockedGroup.className = "achievement-list-unlocked";
    lockedGroup = document.createElement("div");
    lockedGroup.className = "achievement-list-locked";
    // Build and append to our tree.
    listDiv.appendChild(unlockedGroup);
    listDiv.appendChild(clearDiv);
    listDiv.appendChild(lockedGroup);
  }
  function createTab(name, captionTemplate) {
    var tabDiv = document.createElement("div");
    tabDiv.className = "achievement-tab";
    $.bindText(tabDiv, captionTemplate);
    tabDiv.setAttribute("target-name", name);
    tabDiv.addEventListener("click", function (e) {
      setActiveTab(e.target);
    });
    tabs.appendChild(tabDiv);
  }
  // Handles switching between the achievement tabs.
  function setActiveTab(tabBtn) {
    var i;
    var tab = document.getElementById(tabBtn.getAttribute("target-name"));
    var btns = document.getElementsByClassName("achievement-tab");
    var lists = document.getElementsByClassName("achievement-list");
    if (tab !== null) {
      for (i = 0; i < btns.length; i += 1) {
        btns[i].className = "achievement-tab";
      }
      for (i = 0; i < lists.length; i += 1) {
        lists[i].style.display = "none";
      }
      tabBtn.className = "achievement-tab active";
      tab.style.display = "block";
    }
  }
  /**
   * Adds the specified achievement to the achievement page.
   * @param {string} achievement - The name of the achievement.
   * @param {Element} lockedGroup - The div for the locked-group.
   * @param {Element} unlockedGroup - The div for the unlocked-group.
   */
  function addAchievementToPage(achievement, lockedGroup, unlockedGroup) {
    var containerDiv, itemDiv, imageDiv, img, titleDiv, descriptionDiv,
      clearDiv, totalChildren, setChildren, progressDiv, progress, hintStr,
      titleTemplate;
    var achievementsVariable = bewGetVariable("$achievements");
    var dependencies = getAchievementDependecies(achievement);
    function countUnlockedChildren() {
      var i;
      var totalSet = 0;
      for (i = 0; i < dependencies.length; i += 1) {
        if ($.shared.isAchievementSet(dependencies[i])) {
          totalSet += 1;
        }
      }
      return totalSet;
    }
    function listChildrenStatus() {
      var i;
      var result = [];
      for (i = 0; i < dependencies.length; i += 1) {
        if (!$.shared.isAchievementSet(dependencies[i])) {
          result.push("☐ <@achievements." + dependencies[i] + ".info>");
        }
      }
      for (i = 0; i < dependencies.length; i += 1) {
        if ($.shared.isAchievementSet(dependencies[i])) {
          result.push("☑ <@achievements." + dependencies[i] + ".info>");
        }
      }
      return result.join("\n");
    }
    /*
    * Recursively adds all children of the specified achievement.
    * @param {string} parent The name of the parent achievement.
    * @param {Element} lockedGroup The div for the locked-group.
    * @param {Element} unlockedGroup The div for the unlocked-group.
    */
    function addAchievementChildren(parent, lockedGroup, unlockedGroup) {
      var i;
      var dependencies = getAchievementDependecies(parent);
      // Populate the list.
      for (i = 0; i < dependencies.length; i += 1) {
        // Only add achievements, do not add activities.
        if (!isActivity(dependencies[i])) {
          addAchievementToPage(dependencies[i], lockedGroup, unlockedGroup);
        }
      }
    }
    // Bail if the achievement does not exist.
    if (achievementsVariable.elements.indexOf(achievement) < 0) {
      return;
    }
    // First add add all children of this achievement.
    addAchievementChildren(achievement, lockedGroup, unlockedGroup);
    // Bail if the achievement itself is marked as hidden or if it has
    // been added already.
    if (isAchievementHidden(achievement) ||
        displayedAchievements[achievement]) {
      return;
    }
    displayedAchievements[achievement] = true;
    // Generate this achievement.
    containerDiv = document.createElement("div");
    containerDiv.className = "container";
    itemDiv = document.createElement("div");
    itemDiv.className = "item";
    imageDiv = document.createElement("div");
    imageDiv.className = "image";
    img = document.createElement("div");
    img.className = "icon";
    imageDiv.appendChild(img);
    titleDiv = document.createElement("div");
    titleDiv.className = "title";
    $.bindText(titleDiv, "<@achievements." + achievement + ">");
    descriptionDiv = document.createElement("div");
    descriptionDiv.className = "description";
    clearDiv = document.createElement("div");
    clearDiv.setAttribute("style", "clear: both;");
    itemDiv.appendChild(titleDiv);
    itemDiv.appendChild(clearDiv);
    itemDiv.appendChild(descriptionDiv);
    containerDiv.appendChild(imageDiv);
    containerDiv.appendChild(itemDiv);
    // Check if the achievement is unlocked.
    if ($.shared.isAchievementSet(achievement)) {
      // The achievement is unlocked.
      img.style.backgroundImage = "url('" + assetManager.requestImage(
        "icons." + achievement.toLocaleLowerCase()) + "')";
      imageDiv.className = "image set";
      itemDiv.className = "item set";
      $.bindText(descriptionDiv, "<@achievements." + achievement + ".info>");
      unlockedGroup.appendChild(containerDiv);
    } else {
      // The achievement is locked, add the hint text.
      hintStr = "<@achievements." + achievement +
        ($.requestString("achievements." + achievement + ".hint", "")
          ? ".hint>"
          : ".info>");
      $.bindText(descriptionDiv, hintStr);
      // Check if the achievement has any children.
      totalChildren = dependencies.length;
      setChildren = countUnlockedChildren();
      if (totalChildren > 0) {
        progressDiv = document.createElement("div");
        progressDiv.className = "progress-container";
        progress = document.createElement("div");
        progress.className = "progress";
        progress.style.width =
          Math.floor(setChildren * 100 / totalChildren) + "%";
        titleTemplate = setChildren + "/" + totalChildren + "\n" +
          listChildrenStatus();
        $.bindHint(progressDiv, titleTemplate);
        progressDiv.appendChild(progress);
        itemDiv.appendChild(progressDiv);
      }
      if (setChildren > 0) {
        // Add partially completed achievements to the front of the locked group.
        lockedGroup.insertBefore(containerDiv, lockedGroup.firstChild);
      } else {
        // Append locked achevements to the end of the locked group.
        lockedGroup.appendChild(containerDiv);
      }
    }
  }
  var i, menu, el, listDiv, unlockedGroup, lockedGroup;
  var lists = document.createElement("div");
  var tabs = document.createElement("div");
  var displayedAchievements = {};
  // Generate the list-wrapper and the tab-wrapper.
  uiClearMenu("menu-wrapper achievements");
  lists.className = "achievement-list-wrapper";
  tabs.className = "achievement-tab-wrapper";
  lists.appendChild(tabs);
  // Generate a separate list for each day.
  for (i = 1; i < 3; i += 1) {
    // Create the groups for locked/unlocked achievements.
    createList("achievement-list-day" + i);
    // Populate the list.
    addAchievementToPage("d" + i, lockedGroup, unlockedGroup);
    addAchievementToPage("donD" + i, lockedGroup, unlockedGroup);
    lists.appendChild(listDiv);
  }
  // Generate a list for the remaining (general) achievements.
  createList("achievement-list-general");
  listDiv.style.display = "block";
  addAchievementToPage("GM", lockedGroup, unlockedGroup);
  addAchievementToPage("DM", lockedGroup, unlockedGroup);
  lists.appendChild(listDiv);
  // Generate the tabs.
  // First add the general tab.
  createTab("achievement-list-general", "<@achievementTabs.sGeneral>");
  // Then add the tabs for each day.
  createTab("achievement-list-day1", "<@achievementTabs.sWednesday>");
  createTab("achievement-list-day2", "<@achievementTabs.sThursday>");
  // Finally append our tree to the document.
  menu = document.getElementById("Menu");
  menu.appendChild(lists);
  // Generate the fake response-container.
  el = document.createElement("div");
  el.className = "response-container";
  el.id = "FakeResponses";
  menu.appendChild(el);
  uiAddFakeResponse("<@sBack>", uiGenerateMainMenu);
  // Focus on the first tab.
  setActiveTab(document.getElementsByClassName("achievement-tab")[0]);
}

//#endregion

//#endregion

//#region #                Additional Pages                 #

function uiGenerateFakePage() {
  var el, wrapper, menu;
  var doSpellCheck = debugMode.isEnabled() && checkFlag("spellChecking");
  uiClearMenu("menu-wrapper fake-content");
  menu = document.getElementById("Menu");
  // Generate the fake top text container.
  el = document.createElement("p");
  el.className = "main-text";
  el.spellcheck = doSpellCheck;
  el.contentEditable = doSpellCheck;
  el.id = "FakeText";
  wrapper = document.createElement("div");
  wrapper.className = "main-text-container";
  wrapper.appendChild(el);
  menu.appendChild(wrapper);
  // Generate the fake image and hotspot-container.
  el = document.createElement("canvas");
  el.alt = "";
  el.id = "FakeImage";
  wrapper = document.createElement("div");
  wrapper.className = "hotspot-wrapper";
  wrapper.appendChild(el);
  el = document.createElement("div");
  el.className = "hotspot-container";
  el.id = "FakeHotspots";
  wrapper.appendChild(el);
  menu.appendChild(wrapper);
  // Generate the fake response-container.
  wrapper = document.createElement("div");
  wrapper.className = "response-container";
  wrapper.id = "FakeResponses";
  menu.appendChild(wrapper);
}

function uiFakeTopText(textTemplate) {
  $.bindText(document.getElementById("FakeText"), textTemplate);
}

function uiFakeImage(id, scrambled) {
  var imgSrc = assetManager.requestImage(bewCleanId(id));
  var img = document.createElement("img");
  var canvas = document.getElementById("FakeImage");
  var ctx = canvas.getContext("2d");
  function basicDraw(e) {
    canvas.width = e.target.naturalWidth;
    canvas.height = e.target.naturalHeight;
    ctx.drawImage(img, 0, 0);
  }
  scrambled = typeof scrambled === "undefined" ? false : scrambled;
  if (scrambled) {
    img.addEventListener("load", function (e) {
      basicDraw(e);
      bewDescrambleImage(e.target, canvas);
    });
  } else {
    img.addEventListener("load", basicDraw);
  }
  if (imgSrc) {
    img.src = imgSrc;
  } else {
    ctx.clearRect(0, 0, canvas.width, canvas.height);
  }
}

function uiAddFakeHotspot(x, y, w, h, textTemplate, func, extraClass) {
  var el;
  var group = document.createElement("div");
  var doSpellCheck = debugMode.isEnabled() && checkFlag("spellChecking");
  extraClass = extraClass || "";
  group.className = "hotspot" + (extraClass ? " " + extraClass : "");
  group.style.left = x + "%";
  group.style.top = y + "%";
  group.style.width = w + "%";
  group.style.height = h + "%";
  // First create the underlying <i>-element.
  el = document.createElement("i");
  //el.innerHTML = text;
  el.role = "button";
  el.tabIndex = 0;
  el.addEventListener("click", func);
  el.addEventListener("keydown", bewKeyDownOnClick);
  group.appendChild(el);
  // Add the blocking <b>-element.
  el = document.createElement("b");
  group.appendChild(el);
  // Add the "tool-tip" <p>-element.
  el = document.createElement("p");
  el.spellcheck = doSpellCheck;
  el.contentEditable = doSpellCheck;
  $.bindText(el, textTemplate);
  document.getElementById("FakeHotspots").appendChild(group);
  document.getElementById("FakeHotspots").appendChild(el);
}

function uiAddFakeResponse(textTemplate, func) {
  var el = document.createElement("div");
  var p = document.createElement("p");
  var doSpellCheck = debugMode.isEnabled() && checkFlag("spellChecking");
  el.className = "response";
  el.role = "button";
  el.tabIndex = 0;
  el.addEventListener("click", func);
  el.addEventListener("keydown", bewKeyDownOnClick);
  p.spellcheck = doSpellCheck;
  p.contentEditable = doSpellCheck;
  $.bindText(p, textTemplate);
  el.appendChild(p);
  document.getElementById("FakeResponses").appendChild(el);
}

//#region ------ Bug Report -------

// Generates the bug report page.
function uiGenerateBugReportPage() {
  var el, hotspots;
  uiGenerateFakePage();
  hotspots = document.getElementById("FakeHotspots");
  // Generate the main bug-report page.
  uiFakeTopText("<@bugReportPage.sTitle>");
  uiFakeImage("BugReport");
  uiAddFakeResponse("<@sBack>", function () {
    uiGenerateHelpMenu();
  });
  uiAddFakeResponse("<@bugReportPage.sStep1>", replaySavePathData);
  uiAddFakeResponse("<@bugReportPage.sStep2>", function () {
    window.open("http://bew.rocks/bugs");
  });
  // Add the first bubble.
  el = document.createElement("div");
  el.className = "bugreport-bubble";
  el.style.width = "20%";
  el.style.top = "10%";
  el.style.left = "8%";
  $.bindText(el, "<@bugReportPage.sBubble1>");
  hotspots.appendChild(el);
  // Add the second bubble.
  el = document.createElement("div");
  el.className = "bugreport-bubble";
  el.style.width = "22%";
  el.style.top = "25%";
  el.style.left = "11%";
  $.bindText(el, "<@bugReportPage.sBubble2>");
  hotspots.appendChild(el);
  // Add the third bubble.
  el = document.createElement("div");
  el.className = "bugreport-bubble";
  el.style.width = "10%";
  el.style.top = "13%";
  el.style.left = "63%";
  $.bindText(el, "<@bugReportPage.sBubble3>");
  hotspots.appendChild(el);
}

//#endregion

//#region ----- Instructions ------

// Generates the requested page of the instructions.
function uiGenerateInstructions(pageNumber) {
  var extraClass = "instructions";
  uiGenerateFakePage();
  if (pageNumber === 0) {
    uiFakeTopText("<span style=\"display: block; text-align: left;\">" +
      "<@instructionsPage.sTitle1></span>");
    uiFakeImage("CityScape");
    uiAddFakeResponse("<@instructionsPage.sPreviousPage>", uiGenerateHelpMenu);
    uiAddFakeResponse("<@instructionsPage.sNextPage>", function () {
      uiGenerateInstructions(1);
    });
  } else if (pageNumber === 1) {
    $.shared.setAchievement("intIns");
    uiFakeTopText("<span style=\"display: block; text-align: left;\">" +
      "<@instructionsPage.sTitle2></span>");
    uiFakeImage("Instructions2");
    uiAddFakeResponse("<@instructionsPage.sPreviousPage>", function () {
      uiGenerateInstructions(0);
    });
    uiAddFakeResponse("<@instructionsPage.sNextPage>", function () {
      uiGenerateInstructions(2);
    });
  } else if (pageNumber === 2) {
    uiFakeTopText("<span style=\"display: block; text-align: left;\">" +
      "<@instructionsPage.sTitle3><span/>");
    uiFakeImage("office4c");
    uiAddFakeHotspot(0, 0, 27.4, 80, "<@instructionsPage.sExampleHotspot1>",
      function () {
        uiGenerateInstructions(3);
      }, extraClass);
    uiAddFakeHotspot(0, 80, 100, 20, "<@instructionsPage.sExampleHotspot2>",
      function () {
        uiGenerateInstructions(3);
      }, extraClass);
    uiAddFakeHotspot(56, 40, 24, 40, "<@instructionsPage.sExampleHotspot3>",
      function () {
        uiGenerateInstructions(3);
      }, extraClass);
    uiAddFakeHotspot(63, 20, 10, 20, "<@instructionsPage.sExampleHotspot4>",
      function () {
        uiGenerateInstructions(3);
      }, extraClass);
    uiAddFakeHotspot(38, 18, 10, 32, "<@instructionsPage.sExampleHotspot5>",
      function () {
        uiGenerateInstructions(3);
      }, extraClass);
    uiAddFakeResponse("<@instructionsPage.sGotoMenuSay>",
      function () {
        uiGenerateInstructions(3);
      });
    uiAddFakeResponse("<@instructionsPage.sGotoMenuThink>", function () {
      uiGenerateInstructions(3);
    });
    uiAddFakeResponse("<@instructionsPage.sGotoMenu>", function () {
      uiGenerateInstructions(3);
    });
  } else {
    uiGenerateHelpMenu();
  }
}

//#endregion

//#region --- Demo Bonus Scene ----

function uiGenerateDemoBonusScene(pageNumber) {
  var imgSrc = [
    "DemoBonus01",
    "DemoBonus02",
    "DemoBonus03",
    "DemoBonus04",
    "DemoBonus05",
    "DemoBonus07",
    "DemoBonus08",
    "DemoBonus07",
    "DemoBonus08",
    "DemoBonus06",
    "DemoBonus09",
    "DemoBonus10",
    "DemoBonus11"
  ];
  pageNumber = typeof pageNumber === "undefined" ? 0 : pageNumber;
  if (pageNumber >= imgSrc.length) {
    uiGenerateMainMenu();
    return;
  }
  uiGenerateFakePage();
  uiFakeTopText("<@bonusSceneDM.topText" + pageNumber + ">");
  uiFakeImage(imgSrc[pageNumber], true);
  uiAddFakeResponse(
    "<@bonusSceneDM.optionText" + pageNumber + ">",
    function () {
      uiGenerateDemoBonusScene(pageNumber + 1);
    }
  );
}

//#endregion

//#region --- Day 1 Bonus Scene ---

function uiGenerateDay1BonusScene(pageNumber) {
  var imgSrc = [
    "d1b01",
    "d1b02",
    "d1b04",
    "d1b03",
    "d1b05",
    "D1B06",
    "d1b07",
    "D1B08",
    "D1B09",
    "D1B10",
    "D1B11",
    "D1B13",
    "D1B12",
    "D1B14",
    "D1B15",
    "D1B15-A",
    "D1B16",
    "D1B17",
    "D1B18",
    "D1B19",
    "D1B20",
    "D1B21",
    "D1B22",
    "D1B23",
    "D1B24",
    "D1B25",
    "D1B26",
    "D1B27",
    "D1B28",
    "D1B29",
    "D1B30"
  ];
  pageNumber = typeof pageNumber === "undefined" ? 0 : pageNumber;
  if (pageNumber >= imgSrc.length) {
    uiGenerateMainMenu();
    return;
  }
  uiGenerateFakePage();
  uiFakeTopText("<@bonusSceneDay1.topText" + pageNumber + ">");
  uiFakeImage(imgSrc[pageNumber], true);
  uiAddFakeResponse(
    "<@bonusSceneDay1.optionText" + pageNumber + ">",
    function () {
      uiGenerateDay1BonusScene(pageNumber + 1);
    }
  );
}

//#endregion

//#endregion

//#region #                      VM                         #

/**
 * Returns a JS-safe string with escaping.
 * @param {string} text The input text.
 * @returns {string} The JS-safe escaped text.
 */
function safeString(text) {
  return text.replace(/"/g, "\\\"").replace(/\n/g, "\\n");
}

function BiskCompiler(game) {
  var inlineId = 0;
  var commands = [];
  var definitions = {};
  var publicList = [];
  /**
   * Tokenizes the specified input string.
   * @param {string} source The input string to tokenize.
   * @returns {Array} The list of tokens.
   */
  function tokenize(source) {
    var i, symbol, next;
    var token = "";
    var tokens = [];
    var keywords = [
      "as", "return", "if", "then", "elif", "else", "end"
    ];
    var qualifiers = [
      "persistent", "public"
    ];
    var blocks = [
      "function", "page", "locale", "hotspot", "response"
    ];
    var operators = [
      "not", "and", "or", "has", "<", "<=", "<>", ">=", ">", "=="
    ];
    var booleans = [
      "true", "false"
    ];
    var types = [
      "boolean", "number", "string", "bunch"
    ];
    var separators = [
      "(", ")", "!", ",", ";"
    ];
    var whitespaces = [
      " ", "\t", "\n", "\r"
    ];
    var line = 1;
    var column = 1;
    function classifyToken(token) {
      function isNumber(token) {
        return /^-?\d+(\.\d+)?$/.test(token);
      }
      if (whitespaces.indexOf(token) > -1) {
        return "whitespace";
      }
      if (keywords.indexOf(token) > -1) {
        return "keyword";
      }
      if (qualifiers.indexOf(token) > -1) {
        return "qualifier";
      }
      if (blocks.indexOf(token) > -1) {
        return "block";
      }
      if (operators.indexOf(token) > -1) {
        return "operator";
      }
      if (booleans.indexOf(token) > -1) {
        return "boolean";
      }
      if (types.indexOf(token) > -1) {
        return "type";
      }
      if (separators.indexOf(token) > -1) {
        if (token === "!") {
          return "call";
        }
        return "separator";
      }
      if (commands.indexOf(token) > -1) {
        return "command";
      }
      if (isNumber(token)) {
        return "number";
      }
      return "identifier";
    }
    function addToken(classification) {
      var startPos = column - token.length;
      classification = classification || classifyToken(token);
      if (classification === "string") {
        startPos -= 1;
      }
      if ((classification !== "whitespace") &&
          (token.length > 0 || classification === "string")) {
        tokens.push({
          token: token,
          type: classification,
          line: line,
          column: startPos
        });
        token = "";
      }
    }
    /** Consumes a single line comment. */
    function consumeComment() {
      token = "";
      while (i < source.length) {
        i += 1;
        column += 1;
        symbol = source[i];
        if (symbol === "\n") {
          line += 1;
          column = 0;
          break;
        }
      }
    }
    /** Consumes a block comment. */
    function consumeBlockComment() {
      token = "";
      while (i < source.length) {
        i += 1;
        column += 1;
        symbol = source[i];
        next = i < source.length - 2 ? source[i + 1] : "";
        if (symbol === "-" && next === "}") {
          i += 1;
          break;
        } else if (symbol === "\n") {
          line += 1;
          column = 1;
        }
      }
    }
    /** Consumes a string. */
    function consumeString() {
      token = "";
      while (i < source.length - 1) {
        i += 1;
        column += 1;
        symbol = source[i];
        if (symbol === "\"") {
          break;
        } else if (symbol === "\n") {
          line += 1;
          column = 1;
        } else if (symbol === "\\") {
          i += 1;
          symbol = i < source.length ? source[i] : "";
        }
        token += symbol;
      }
      addToken("string");
    }
    source = source.split("");
    for (i = 0; i < source.length; i += 1, column += 1) {
      symbol = source[i];
      next = i < source.length - 2 ? source[i + 1] : "";
      if ((whitespaces.indexOf(symbol) > -1)) {
        addToken();
        if (symbol === "\n") {
          // Only increase the line number on line-breaks.
          line += 1;
          column = 0;
        }
      } else if (separators.indexOf(symbol) > -1) {
        addToken();
        token = symbol;
        column += 1;
        addToken();
        column -= 1;
      } else if (symbol === "-" && next === "-") {
        consumeComment();
      } else if (symbol === "{" && next === "-") {
        consumeBlockComment();
      } else if (symbol === "\"") {
        consumeString();
      } else {
        token += symbol;
      }
    }
    addToken();
    return tokens;
  }
  /**
   * Compiles the specified Bisk source.
   * @param {string} source The source code to compile.
   * @param {string} sourceFile The name of the source file.
   */
  function compile(source, sourceFile) {
    var tokenIndex = 0;
    var tokens = tokenize(source);
    var token, context, qualifier;
    sourceFile = sourceFile || "";
    source = source.replace(/\t/g, "  ").split("\n");
    function logIssue(token, message) {
      var i;
      var line = token.line;
      var column = token.column;
      var padLen = (source.length + "").length;
      var str = "%c" + sourceFile + "[Line " + line + ":" + column + "] ";
      function pad(s, len) {
        var pattern = "";
        while (pattern.length < len) {
          pattern += " ";
        }
        return (pattern + s).slice(-len);
      }
      function printLine(line) {
        str += "\n" + pad(line + 1, padLen) + "|" + source[line];
      }
      function markIssue() {
        var markLen = token.token.length + (token.type === "string" ? 2 : 0);
        str += "\n";
        for (i = 0; i < padLen + column; i += 1) {
          str += " ";
        }
        str += "%c";
        for (i = 0; i < markLen; i += 1) {
          str += "^";
        }
        str += "%c";
      }
      str += message;
      str += "\n%cFrom " + context[context.length - 1].name + ":%c";
      context[context.length - 1].issues += 1;
      if (line > 1) {
        printLine(line - 2);
      }
      printLine(line - 1);
      markIssue();
      if (line < source.length) {
        printLine(line);
      }
      terminal.error(str, "",
        "text-decoration: underline", "", "color: red", "");
    }
    function peekToken() {
      var implicitEnd = {
        token: "end",
        type: "keyword",
        line: 0,
        column: 0
      };
      return tokenIndex < tokens.length ? tokens[tokenIndex] : implicitEnd;
    }
    function nextToken() {
      token = peekToken();
      tokenIndex += 1;
    }
    function expectToken(type) {
      type = typeof type === "string" ? [type] : type;
      nextToken();
      if (type.indexOf(token.type) > -1) {
        return token.token;
      }
      logIssue(token,
        "Unexpected " + token.type + " - " + type.join(" or ") + " expected.");
      return null;
    }
    /**
     * Parses and adds a function to the specified collection.
     * @param {Array} collection - The collection to hold the function.
     * @param {string} collectionType - The type of the collection. (ex: "page")
     * @param {Array} [parent] - The parent context.
     */
    function addFunction(collection, collectionType, parent) {
      var func = {
        name: "",
        body: [],
        returnType: "",
        variables: {},
        exported: qualifier === "public"
      };
      collectionType = collectionType || "function";
      parent = parent || [];
      context.push({
        name: context[context.length - 1].name,
        issues: 0
      });
      function addVariable(name, type) {
        func.variables[name] = { type: type };
      }
      function verifyVariable(name) {
        function checkVariables(variables) {
          var i;
          var result = false;
          var names = Object.keys(variables);
          for (i = 0; i < names.length; i += 1) {
            result = result || names[i] === name;
          }
          return result;
        }
        if (!checkVariables(func.variables) &&
            !checkVariables(game.variables)) {
          logIssue(token, "Undefined variable `" + name + "` accessed.");
        }
      }
      function processCommand(output) {
        var i;
        var params = [];
        var cmd = token.token;
        var definition = definitions[cmd];
        for (i = 0; i < definition.length; i += 1) {
          expectToken(definition[i]);
          if (token.type === "identifier" &&
              definition[i].length === 1 &&
              definition[i][0] === "identifier") {
            // Special case for passing the identifier by name instead of value.
            verifyVariable(token.token);
            params.push("\"" + token.token + "\"");
          } else if (token.type === "call") {
            processCall(params);
          } else if (token.type === "identifier") {
            verifyVariable(token.token);
            params.push("$.get(\"" + token.token + "\")");
          } else if (token.type === "string") {
            params.push("\"" + safeString(token.token) + "\"");
          } else {
            params.push(token.token);
          }
        }
        output.push(
          (cmd === "jmp" ? "return " : "") +
          "$." + cmd + "(" + params.join(", ") + ")"
        );
      }
      function processCall(output) {
        function getParamType(variables, name) {
          var i;
          var names = Object.keys(variables);
          for (i = 0; i < names.length; i += 1) {
            if (names[i] === name) {
              return variables[name].type;
            }
          }
          return null;
        }
        var paramType;
        var params = [];
        var funcName = expectToken("identifier") + "!";
        expectToken("separator");
        while (peekToken().token !== ")") {
          expectToken(["identifier", "keyword",
            "string", "number", "boolean", "call"]);
          if (token.type === "call") {
            paramType = processCall(params);
          } else if (token.type === "identifier") {
            verifyVariable(token.token);
            params.push("$.get(\"" + token.token + "\")");
            // Try to determine the type of the parameter, default to "number".
            paramType = getParamType(func.variables, token.token) ||
              getParamType(game.variables, token.token) || "number";
          } else if (token.type === "string") {
            params.push("\"" + safeString(token.token) + "\"");
            paramType = "string";
          } else {
            params.push(token.token);
            if (token.token === "true" || token.token === "false") {
              paramType = "boolean";
            } else {
              paramType = "number";
            }
          }
          if (paramType === "string") {
            funcName += "s";
          } else if (paramType === "boolean") {
            funcName += "b";
          } else {
            funcName += "n";
          }
        }
        nextToken();
        output.push(
          "$.call(\"" + funcName + "\", [" + params.join(", ") + "])");
        return func.returnType;
      }
      function extractCondition(output) {
        var previousToken;
        var needOperator = false;
        var opMap = {
          "not": "!",
          "and": "&&",
          "or": "||",
          "<": "<",
          "<=": "<=",
          "<>": "!==",
          ">=": ">=",
          ">": ">",
          "==": "==="
        };
        while (peekToken().token !== "then") {
          expectToken(["identifier", "operator", "separator", "call",
            "string", "number", "boolean"]);
          if (!needOperator && token.type === "call") {
            processCall(output);
            needOperator = true;
          } else if (!needOperator && peekToken().token === "has") {
            needOperator = true;
          } else if (!needOperator && token.type === "identifier") {
            verifyVariable(token.token);
            output.push("$.get(\"" + token.token + "\")");
            needOperator = true;
          } else if (!needOperator && token.type === "string") {
            output.push("\"" + safeString(token.token) + "\"");
            needOperator = true;
          } else if (token.type === "operator" && token.token !== "has") {
            output.push(opMap[token.token]);
            needOperator = false;
          } else {
            if (!needOperator || ["(", ")"].indexOf(token.token) > -1) {
              output.push(token.token);
            } else {
              logIssue(token, "Unexpected " + token.type + " `" +
                token.token + "` - operator expected.");
            }
          }
          if (needOperator && peekToken().token !== "then") {
            if (["identifier"].indexOf(token.type) > -1 &&
                peekToken().token === "has") {
              // Special case for `bunch has item` statements.
              previousToken = token;
              expectToken("operator");
              expectToken(["identifier", "string"]);
              if (token.type === "identifier") {
                output.push("$.check(\"" + previousToken.token +
                  "\", $.get(\"" + token.token + "\"))");
              } else {
                output.push("$.check(\"" + previousToken.token +
                  "\"), \"" + token.token + "\")");
              }
              needOperator = false;
            } else {
              expectToken(["operator", "separator"]);
              if (token.type === "operator") {
                output.push(opMap[token.token]);
                needOperator = false;
              } else if (["(", ")"].indexOf(token.token) > -1) {
                output.push(token.token);
              } else {
                logIssue(token, "Unexpected " + token.type + " `" +
                  token.token + "` - operator expected.");
              }
            }
          }
        }
        nextToken();
      }
      function processIf(output) {
        var condition = [];
        extractCondition(condition);
        condition = condition.join(" ");
        output.push("if ($._if(\"if " + condition.replace(/"/g, "\\\"") +
          " then\", " + condition + ")) {");
        while (peekToken().token !== "end") {
          processExpressions(output, ["elif", "else", "end"]);
          if (token.token === "elif") {
            condition = [];
            extractCondition(condition);
            condition = condition.join(" ");
            output.push("} else if ($._if(\"elif " +
              condition.replace(/"/g, "\\\"") +
              " then\", " + condition + ")) {");
          } else if (token.token === "else") {
            output.push("} else {");
            if (peekToken().token === "if" && peekToken().line === token.line) {
              // If an "else" follows an "if" on the same line then
              // it's most probably by error, so log it as an issue.
              logIssue(peekToken(),
                "Unexpected `if` after `else` - use `elif` instead.");
            }
          } else if (token.token === "end") {
            break;
          }
        }
        output.push("}");
      }
      function processExpressions(output, terminators) {
        var varType, varName;
        var returnContext = false;
        terminators = typeof terminators === "undefined"
          ? ["end"]
          : typeof terminators === "string"
            ? [terminators]
            : terminators;
        while (terminators.indexOf(peekToken().token) < 0) {
          expectToken(["block", "type", "keyword", "identifier",
            "command", "call", "string", "number", "boolean"]);
          if (token.token === "if") {
            processIf(output);
          } else if (collectionType === "page" &&
              (token.token === "hotspot" || token.token === "response")) {
            addFunction(game.functions, token.token, output);
          } else if (collectionType === "page" && token.token === "locale") {
            addLocaleStrings("pages." + func.name.replace(/\s/g, ""));
          } else if (token.token === "return") {
            output.push("return");
            returnContext = true;
          } else if (returnContext &&
              ["boolean", "number"].indexOf(token.type) > -1) {
            output.push(token.token + ";");
            returnContext = false;
          } else if (returnContext && token.type === "string") {
            output.push("\"" + safeString(token.token) + "\";");
            returnContext = false;
          } else if (returnContext && token.type === "identifier") {
            verifyVariable(token.token);
            output.push("$.get(\"" + token.token + "\")");
            returnContext = false;
          } else if (token.type === "type") {
            varType = token.token;
            varName = expectToken("identifier");
            addVariable(varName, varType);
            output.push(
              "$.declare(\"" + varType + "\", \"" + varName + "\")");
          } else if (token.type === "command") {
            processCommand(output);
          } else if (token.type === "call") {
            processCall(output);
          } else {
            logIssue(token,
              "Unexpected " + token.type + " `" + token.token + "`.");
          }
        }
        nextToken();
      }
      function processFunctionDeclaration(output) {
        var varType, varName, typeToken;
        func.name = expectToken("identifier");
        context[context.length - 1].name = collectionType + " " + func.name;
        func.name += "!";
        if (peekToken().token === "(") {
          nextToken();
          while (peekToken().token !== ")") {
            varType = expectToken("type");
            typeToken = token;
            varName = expectToken("identifier");
            addVariable(varName, varType);
            if (varType === "string") {
              func.name += "s";
              output.push(
                "$.declare(\"string\", \"" + varName + "\", $.pop())");
            } else if (varType === "number") {
              func.name += "n";
              output.push(
                "$.declare(\"number\", \"" + varName + "\", $.pop())");
            } else if (varType === "boolean") {
              func.name += "b";
              output.push(
                "$.declare(\"boolean\", \"" + varName + "\", $.pop())");
            } else {
              logIssue(typeToken, "Unexpected " + typeToken.type + " `" +
                typeToken.token + "` specified as a function parameter.");
            }
          }
          nextToken();
        }
        if (peekToken().token === "as") {
          nextToken();
          func.returnType = expectToken("type");
        }
      }
      function processPageDeclaration() {
        func.name = expectToken("string");
        context[context.length - 1].name =
          collectionType + " \"" + func.name + "\"";
      }
      function processHotspotDeclaration() {
        var descriptor = expectToken("string");
        var text = expectToken("string");
        inlineId += 1;
        func.name = "anonymous!" + inlineId;
        context[context.length - 1].name =
          collectionType + " \"" + func.name + "\"";
        parent.push("$.addHotspot(\"" + descriptor + "\", \"" +
          safeString(text) + "\", \"" + func.name + "\")");
      }
      function processResponseDeclaration() {
        var text = expectToken("string");
        inlineId += 1;
        func.name = "anonymous!" + inlineId;
        context[context.length - 1].name =
          collectionType + " \"" + func.name + "\"";
        parent.push("$.addResponse(\"" + safeString(text) +
          "\", \"" + func.name + "\")");
      }
      function processDeclaration() {
        if (collectionType === "function") {
          processFunctionDeclaration(func.body);
        } else if (collectionType === "page") {
          processPageDeclaration();
        } else if (collectionType === "hotspot") {
          processHotspotDeclaration();
        } else if (collectionType === "response") {
          processResponseDeclaration();
        }
      }
      function prettifyOutput() {
        var i, k;
        var level = 1;
        for (i = 0; i < func.body.length; i += 1) {
          if (func.body[i].split("").pop() === ")") {
            func.body[i] += ";";
          }
          if (func.body[i].indexOf("}") > -1) {
            level -= 1;
          }
          for (k = 0; k < level; k += 1) {
            func.body[i] = "  " + func.body[i];
          }
          if (func.body[i].indexOf("{") > -1) {
            level += 1;
          }
        }
        func.body = func.body.join("\n");
        func.body = func.body.replace(/return\n\s+/g, "return ");
      }
      processDeclaration();
      processExpressions(func.body, ["end"]);
      // Format the generated function body.
      prettifyOutput();
      // Generate and add the function to the specified function collection.
      if (context[context.length - 1].issues === 0 && func.body.length > 0) {
        try {
          collection[func.name] = new Function(func.body);
          if (func.exported) {
            publicList.push(func.name);
          }
        } catch (e) {
          terminal.error(e);
          terminal.error(func.body);
        }
      }
      context.pop();
    }
    /**
     * Adds a new global variable to the specified list.
     * @param {Array} variables - The list to hold the new variable.
     */
    function addGlobalVariable(variables) {
      var variable = {};
      var name;
      function addGlobalNumber() {
        variable.defaultValue = 0;
        if (peekToken().type === "number") {
          variable.defaultValue = parseFloat(expectToken("number"));
          if (peekToken().type === "string") {
            variable.group = expectToken("string");
          }
        }
      }
      function addGlobalString() {
        variable.defaultValue = "";
        if (peekToken().type === "string") {
          variable.defaultValue = expectToken("string");
          if (peekToken().type === "string") {
            variable.group = expectToken("string");
          }
        }
      }
      function addGlobalBoolean() {
        // NOTE(kexter): Internally, Booleans are stored as numbers.
        variable.defaultValue = 0;
        if (peekToken().type === "boolean") {
          variable.defaultValue = expectToken("boolean") === "true" ? 1 : 0;
          if (peekToken().type === "string") {
            variable.group = expectToken("string");
          }
        }
      }
      function addGlobalBunch() {
        var i, group, defaultValue;
        var states = [];
        variable.defaultValue = "";
        variable.elements = [];
        while (peekToken().token !== "end") {
          variable.elements.push(expectToken("string"));
          if (peekToken().type === "boolean") {
            states.push(expectToken("boolean") === "true" ? 1 : 0);
          } else {
            states.push(0);
          }
        }
        nextToken();
        // Generate the defaultValue from the states array.
        states.push(0, 0, 0);
        defaultValue = "";
        for (i = 0; i < (variable.elements.length + 3) >> 2; i += 1) {
          group = states.splice(0, 4);
          group = group[0] + group[1] * 2 + group[2] * 4 + group[3] * 8;
          defaultValue += group.toString(16);
        }
        defaultValue = defaultValue.replace(/0+$/g, "");
        variable.defaultValue = defaultValue;
      }
      variable.type = token.token;
      name = expectToken("identifier");
      variable.persistent = qualifier === "persistent";
      if (variable.type === "number") {
        addGlobalNumber();
      } else if (variable.type === "string") {
        addGlobalString();
      } else if (variable.type === "boolean") {
        addGlobalBoolean();
      } else if (variable.type === "bunch") {
        addGlobalBunch();
      } else {
        logIssue(token, "Unexpected type `" + variable.type + "`.");
        return;
      }
      variable.value = variable.defaultValue;
      variables[name] = variable;
    }
    function addLocaleStrings(ctx) {
      var id, value, locale;
      var localeId = expectToken("string");
      ctx = ctx || "";
      ctx += ctx === "" ? "" : ".";
      locale = game.locales[localeId];
      if (typeof locale === "undefined") {
        game.locales[localeId] = {};
        locale = game.locales[localeId];
      }
      while (peekToken().token !== "end") {
        id = ctx + expectToken("identifier");
        value = expectToken("string");
        locale[id] = value;
      }
      nextToken();
    }
    function publishFunctions() {
      function publishGroup(group) {
        var name = group.name.split("!")[0];
        var PREAMBLE = "var args = [].slice.call(arguments);\n" +
          "var len = args.length;\n" +
          "var types = args.map(function (x) {\n" +
          "  return (typeof x).split(\"\").shift();\n" +
          "});\n";
        var EPILOGUE =
          "terminal.error(\"Function `" + name + "!\" + " +
          "types.join(\"\") + \"` is not public.\");\n";
        var routes = group.list
          .sort(function (a, b) {
            var lenA = a.length;
            var lenB = b.length;
            return lenA > lenB ? -1 : lenA < lenB ? 1 : 0;
          })
          .map(function (name) {
            var types = name.split("!")[1].split("");
            var conditions = types.map(function (value, index) {
              return "types[" + index + "] === \"" + value + "\"";
            });
            conditions.unshift("len >= " + types.length);
            conditions = conditions.join(" && ");
            return "if (" + conditions + ") {\n" +
              "  return $.call(\"" + name + "\", args);\n" +
              "}\n";
          });
        $.shared[name] = new Function(PREAMBLE + routes.join("") + EPILOGUE);
      }
      var groups = [];
      var i, group, name;
      if (publicList.length > 0) {
        publicList = publicList.sort();
        group = {
          name: "",
          list: []
        };
        for (i = 0; i < publicList.length; i += 1) {
          name = publicList[i].split("!")[0];
          if (group.name !== name) {
            group = {
              name: name,
              list: []
            };
            groups.push(group);
          }
          group.list.push(publicList[i]);
        }
        for (i = 0; i < groups.length; i += 1) {
          publishGroup(groups[i]);
        }
      }
    }
    // { token, type, line, column }
    qualifier = "";
    context = [{
      name: sourceFile,
      issues: 0
    }];
    while (tokenIndex < tokens.length) {
      expectToken(["block", "type", "qualifier"]);
      if (token.type === "qualifier") {
        qualifier = token.token;
        continue;
      } else if (token.type === "type") {
        addGlobalVariable(game.variables);
      } else if (token.token === "function") {
        addFunction(game.functions, token.token);
      } else if (token.token === "page") {
        addFunction(game.pages, token.token);
      } else if (token.token === "locale") {
        addLocaleStrings();
      }
      qualifier = "";
    }
    publishFunctions();
  }
  /**
   * Defines a command.
   * @param {string} command - The name of the command.
   * @param {string[]} definition - The definition for the command.
   */
  function define(command, definition) {
    commands.push(command);
    definitions[command] = definition;
  }
  return {
    compile: compile,
    define: define
  };
}

function BiskStack() {
  var data = [];
  var frames = [];
  function push(input) {
    data.push(input);
  }
  function pop() {
    return data.pop();
  }
  function declare(type, name, value) {
    var frame;
    if (frames.length > 0) {
      if (typeof value === "undefined") {
        if (type === "string") {
          value = "";
        } else if (type === "number") {
          value = 0;
        } else if (type === "boolean") {
          value = false;
        }
      }
      frame = frames[frames.length - 1];
      name = frame.name + "_" + name;
      frame.variables.push({
        name: name,
        type: type,
        value: value
      });
    }
  }
  function resolve(name) {
    var i, frame, mangledName;
    if (frames.length > 0) {
      frame = frames[frames.length - 1];
      mangledName = frame.name + "_" + name;
      for (i = 0; i < frame.variables.length; i += 1) {
        if (frame.variables[i].name === mangledName) {
          return frame.variables[i];
        }
      }
    }
    return null;
  }
  function isEmpty() {
    return frames.length === 0;
  }
  function frameOpen(name) {
    frames.push({
      name: name,
      variables: []
    });
  }
  function frameClose() {
    frames.pop();
  }
  function depth() {
    return frames.length;
  }
  return {
    push: push,
    pop: pop,
    resolve: resolve,
    isEmpty: isEmpty,
    frameOpen: frameOpen,
    frameClose: frameClose,
    declare: declare,
    depth: depth
  };
}

var $ = (function (game) {
  var vm;
  var predictorMode = false;
  var engine = new BrewEngine();
  var stack = new BiskStack();
  var compiler = new BiskCompiler(game);
  var variableChangeEvents = [];
  var publicFunctions = {};
  /**
   * Exposes the specified function to the VM as the specified command with
   * the specified definition.
   * @param {function} func - The function to expose.
   * @param {string} command - The name of the exposed command.
   * @param {Array} definition - The definition of the command.
   */
  function expose(func, command, definition) {
    vm[command] = func;
    compiler.define(command, definition);
  }
  function publishEngineFunctions() {
    Object.keys(engine).forEach(function (key) {
      vm[key] = engine[key];
    });
  }
  function logBuiltInCall(call) {
    debugLogAddLine(call, { level: stack.depth() });
  }
  function logBeginFunctionCall(funcName, params, level) {
    var i, text;
    var result = null;
    var options = {
      level: level,
      className: ""
    };
    text = funcName.slice(0, funcName.lastIndexOf("!"));
    for (i = 0; i < params.length; i += 1) {
      if (typeof params[i] === "string") {
        params[i] = "\"" + safeString(params[i]) + "\"";
      }
    }
    params = params.join(" ");
    text += "(" + params + ")";
    result = debugLogAddLine("!" + text, options);
    return result;
  }
  function logEndFunctionCall(entry, result) {
    if (entry) {
      if (typeof result !== "undefined") {
        entry.text += "{=" + result + "}";
        if (typeof result === "boolean") {
          entry.className = result ? "pass" : "fail";
        }
      }
    }
  }
  function predict(func, initialState) {
    var savedVariables, backup;
    var loggingState = bewDebugLoggingEnabled;
    var result = {
      page: "",
      state: "",
      text: "",
      images: []
    };
    function predictJump(name) {
      var startPage = $.get("$startpage");
      var endPage = $.get("$endpage");
      if (name.indexOf("|") > -1) {
        // Remove all parameters.
        name = name.split("|").shift();
      }
      // If the page doesn't exist then redirect to the end page.
      name = typeof game.pages[name] === "function" ? name : endPage;
      result.page = name;
      if (name === startPage) {
        setState("");
      }
      game.pages[name]();
      result.state = getState();
    }
    function predictBlit(descriptor/*, layerId*/) {
      var i, id;
      // Descriptor samples:
      //   "frame01::500|frame02::500|frame03::500"
      //   "sample@25,25,50,50"
      descriptor = descriptor || "";
      descriptor = descriptor.split("@")[0].split("|");
      for (i = 0; i < descriptor.length; i += 1) {
        id = descriptor[i].split(":")[0];
        if (result.images.indexOf(id) < 0) {
          result.images.push(id);
        }
      }
    }
    function predictPrint(text) {
      result.text = $.evaluateText(text, result.page);
    }
    function predictAddHotspot(/*descriptor, text, callback*/) {
      // NOP.
    }
    function predictAddResponse(/*text, callback*/) {
      // NOP.
    }
    if (!predictorMode && typeof func === "function") {
      // Disable logging and enable predictor mode.
      bewDebugLoggingEnabled = false;
      predictorMode = true;
      // Save all variables and the current page id.
      savedVariables = JSON.parse(JSON.stringify(game.variables));
      // Backup the original VM functions.
      backup = {
        jmp: vm.jmp,
        blit: vm.blit,
        print: vm.print,
        addHotspot: vm.addHotspot,
        addResponse: vm.addResponse
      };
      // Replace VM functions.
      vm.jmp = predictJump;
      vm.blit = predictBlit;
      vm.print = predictPrint;
      vm.addHotspot = predictAddHotspot;
      vm.addResponse = predictAddResponse;
      // Set the initial state, if specified.
      if (typeof initialState === "string") {
        setState(initialState);
      }
      // Execute the specified function.
      func();
      // Restore VM functions.
      Object.keys(backup).forEach(function (key) {
        vm[key] = backup[key];
      });
      // Restore all variables and the current page id.
      game.variables = savedVariables;
      // Disable predictor mode and reset the state of logging.
      predictorMode = false;
      bewDebugLoggingEnabled = loggingState;
    }
    return result;
  }
  function isPredicting() {
    return predictorMode;
  }
  /*
   * Registers a callback function which gets called every time the specified
   * variable is modified.
   * @param {string} name - The name of the variable or "*" for all variables.
   * @param {function} callback - The callback function.
   */
  function onVariableChange(name, callback) {
    if (typeof name === "string" && typeof callback === "function") {
      variableChangeEvents.push({
        name: name,
        callback: callback
      });
    }
  }
  function _if(condition, value) {
    var match, varValue;
    var formatted = condition;
    var rGet = /(\$\.get\("([^"]*)"\))/g;
    if (bewDebugLoggingEnabled) {
      // Annotate $.get() calls.
      match = rGet.exec(condition);
      while (match !== null) {
        varValue = $.get(match[2]);
        varValue = typeof varValue === "string"
          ? "\"" + varValue + "\""
          : varValue;
        varValue = match[2] + "{=" + varValue + "}";
        formatted = formatted.replace(match[1], varValue);
        match = rGet.exec(condition);
      }
      formatted = bewUtilJsToBisk(formatted);
      debugLogAddLine(formatted, {
        level: stack.depth(),
        className: !value ? "fail" : "pass"
      });
    }
    return value;
  }
  function call(funcName, params) {
    var i, entry, result;
    var func = game.functions[funcName];
    var excludeList = [
      "frac!n",
      "floor!n",
      "round!n",
      "ceil!n",
      "isActivity!s",
      "isAchievementSet!s",
      "isAchievementHidden!s",
      "isEventSet!s",
      "updateAchievements!",
      "setAchievement!sb",
      "setEvent!sb",
      "getAchievementDependencies!s",
      "checkScore!snnn",
      "checkScore!sns",
      "checkScore!sn"
    ];
    var loggingState = bewDebugLoggingEnabled;
    var doLogCall = excludeList.indexOf(funcName) < 0;
    // TODO: Remove the exclude list once proper tracing is implemented.
    if (typeof func === "function") {
      if (typeof params === "object" &&
          typeof params.length !== "undefined") {
        for (i = params.length - 1; i >= 0; i -= 1) {
          stack.push(params[i]);
        }
      }
      if (doLogCall) {
        entry = logBeginFunctionCall(funcName, params, stack.depth());
      } else {
        bewDebugLoggingEnabled = false;
      }
      stack.frameOpen(funcName);
      result = func();
      stack.frameClose();
      if (doLogCall) {
        if (entry) {
          logEndFunctionCall(entry, result);
        }
      } else {
        bewDebugLoggingEnabled = loggingState;
      }
    } else {
      terminal.error("ReferenceError: " + funcName + " is not defined.");
    }
    return result;
  }
  function jmp(page) {
    var i, params, param, value;
    var zoomToRectData = null;
    var doFade = false;
    var devZoomEnabled = false;
    function performJump() {
      function doJump() {
        var startPage = $.get("$startpage");
        var endPage = $.get("$endpage");
        if (typeof game.pages === "undefined") {
          return;
        }
        // Clean up the page-identifier.
        page = bewCleanId(page);
        // If the page doesn't exist then redirect to the end-page.
        page = typeof game.pages[page] === "function" ? page : endPage;
        if (page === endPage) {
          saveState("X");
        }
        // We are not in the menu.
        document.body.classList.remove("in-menu");
        // Set the current page.
        location.hash = page;
        if (page === startPage) {
          initVars();
          $.history.clear();
        }
        $.history.record(page, getState());
        stack.frameOpen(page);
        $.generate();
        stack.frameClose();
        // Check if we are on the end-page and we have data in "$xsave".
        if (page === endPage) {
          bewSpecialSave = $.get("$xsave");
          if (bewSpecialSave) {
            bewSpecialSave = bewSpecialSave.split("|");
            bewSpecialSave.splice(0, 1);
            deleteSave("X");
          }
        } else {
          bewSpecialSave = "";
        }
        // Generate the debug log.
        debugGenerateLog();
        bewDebugLoggingEnabled = false;
      }
      bewDebugLoggingEnabled = true;
      logBuiltInCall("jmp \"" + page + "\"");
      doFade = checkFlag("noEffects") ? false : doFade;
      if (doFade) {
        $.fx.fadeOut();
        setTimeout(function () {
          doJump();
          $.fx.fadeIn();
        }, 500);
      } else {
        doJump();
      }
    }
    $.fx.zoomFromRectData = null;
    if (page.indexOf("|") > -1) {
      // Parameters are in the form "param=value", separated by vertical bars.
      params = page.split("|");
      page = params.shift();
      for (i = 0; i < params.length; i += 1) {
        value = params[i].toLowerCase().split("=");
        param = value.shift().trim();
        value = value.join("=");
        // TODO: Re-enable zoom-transition effects. (devZoomEnabled = true)
        if (param === "transition" && value === "fade") {
          doFade = true;
        } else if (devZoomEnabled && param === "zoomtorect") {
          value = value.split(",");
          if (value.length === 4) {
            doFade = false;
            zoomToRectData = {
              x: parseFloat(value[0].trim()),
              y: parseFloat(value[1].trim()),
              w: parseFloat(value[2].trim()),
              h: parseFloat(value[3].trim())
            };
          }
        } else if (devZoomEnabled && param === "zoomfromrect") {
          value = value.split(",");
          if (value.length === 4) {
            doFade = false;
            $.fx.zoomFromRectData = {
              x: parseFloat(value[0].trim()),
              y: parseFloat(value[1].trim()),
              w: parseFloat(value[2].trim()),
              h: parseFloat(value[3].trim())
            };
          }
        }
      }
    }
    if (zoomToRectData !== null) {
      $.fx.zoomToRect(zoomToRectData.x, zoomToRectData.y,
        zoomToRectData.w, zoomToRectData.h, performJump);
    } else {
      performJump();
    }
  }
  function check(bunch, id) {
    var variable = bewGetVariable(bunch);
    if (typeof variable === "object" && variable.type === "bunch") {
      return bewGetBitfieldValue(bunch, variable.elements, id) === 1;
    }
    return false;
  }
  function get(name) {
    // First search for the variable amongst the local variables.
    var variable = stack.resolve(name);
    if (variable !== null) {
      return variable.value;
    }
    // Then check for it amongst the global variables.
    variable = bewGetVariable(name);
    if (variable !== null) {
      if (!predictorMode) {
        debugTouchVariable(name);
      }
      if (variable.type === "boolean") {
        return !!variable.value;
      }
      return variable.value;
    }
    // Return 0 if the variable doesn't exist.
    return 0;
  }
  function set(name, value) {
    var i, event, previousValue;
    var loggingState = bewDebugLoggingEnabled;
    var variable = stack.resolve(name);
    if (variable !== null) {
      variable.value = value;
    } else {
      variable = bewGetVariable(name);
      if (variable !== null) {
        // Only log calls affecting global variables,
        // but not system variables or bunches.
        if (variable.type !== "bunch" && variable.group !== "$") {
          logBuiltInCall("set " + name + " " +
            (typeof value === "string" ? "\"" + value + "\"" : value));
        }
        previousValue = variable.value;
        bewSetVariable(name, value);
        if (game.running && !predictorMode && previousValue !== value) {
          bewSerializeData();
          bewDebugLoggingEnabled = false;
          // Run any registered callbacks.
          for (i = 0; i < variableChangeEvents.length; i += 1) {
            event = variableChangeEvents[i];
            if (event.name === "*" || event.name === name) {
              event.callback(name, previousValue);
            }
          }
          bewDebugLoggingEnabled = loggingState;
        }
      }
    }
  }
  function add(name, value) {
    var variable = bewGetVariable(name);
    if (variable !== null && variable.type === "bunch") {
      bewSetBitfieldValue(name, variable.elements, value, 1);
    } else {
      set(name, get(name) + value);
    }
  }
  function sub(name, value) {
    var variable = bewGetVariable(name);
    if (variable !== null && variable.type === "bunch") {
      bewSetBitfieldValue(name, variable.elements, value, 0);
    } else {
      set(name, get(name) - value);
    }
  }
  function mul(name, value) {
    set(name, get(name) * value);
  }
  function div(name, value) {
    set(name, get(name) / value);
  }
  function mod(name, value) {
    set(name, get(name) % value);
  }
  function trunc(name) {
    var value = get(name);
    value = value < 0 ? Math.ceil(value) : Math.floor(value);
    set(name, value);
  }
  vm = {
    shared: publicFunctions,
    compile: compiler.compile,
    predict: predict,
    isPredicting: isPredicting,
    onVariableChange: onVariableChange,
    requestString: assetManager.requestString,
    declare: stack.declare,
    pop: stack.pop,
    _if: _if,
    call: call,
    get: get,
    check: check
  };
  // Make all functions from the engine object public.
  publishEngineFunctions();
  // Expose functions as commands to the VM.
  expose(set, "set",
    [["identifier"], ["identifier", "string", "number", "boolean", "call"]]);
  expose(add, "add",
    [["identifier"], ["identifier", "string", "number", "call"]]);
  expose(sub, "sub",
    [["identifier"], ["identifier", "string", "number", "call"]]);
  expose(mul, "mul", [["identifier"], ["identifier", "number", "call"]]);
  expose(div, "div", [["identifier"], ["identifier", "number", "call"]]);
  expose(mod, "mod", [["identifier"], ["identifier", "number", "call"]]);
  expose(trunc, "trunc", [["identifier"]]);
  expose(jmp, "jmp", [["identifier", "string"]]);
  expose(vm.print, "print", [["identifier", "string"]]);
  expose(vm.blit, "blit", [["identifier", "string"]]);
  return vm;
}(game));

//#endregion
